@@ -18,6 +18,8 @@ import argparse
 
 # Packaging variables
 PACKAGE_NAME = "influxdb"
+USER = "influxdb"
+GROUP = "influxdb"
 INSTALL_ROOT_DIR = "/usr/bin"
 LOG_DIR = "/var/log/influxdb"
 DATA_DIR = "/var/lib/influxdb"
@@ -28,6 +30,7 @@ MAN_DIR = "/usr/share/man"
 
 INIT_SCRIPT = "scripts/init.sh"
 SYSTEMD_SCRIPT = "scripts/influxdb.service"
+SYSTEMD_START_SCRIPT = "scripts/influxd-systemd-start.sh"
 PREINST_SCRIPT = "scripts/pre-install.sh"
 POSTINST_SCRIPT = "scripts/post-install.sh"
 POSTUNINST_SCRIPT = "scripts/post-uninstall.sh"
@@ -49,7 +52,7 @@ VENDOR = "InfluxData"
 DESCRIPTION = "Distributed time-series database."
 
 prereqs = [ 'git', 'go' ]
-go_vet_command = "go tool vet ./"
+go_vet_command = "go vet ./..."
 optional_prereqs = [ 'fpm', 'rpmbuild', 'gpg' ]
 
 fpm_common_args = "-f -s dir --log error \
@@ -63,6 +66,8 @@ fpm_common_args = "-f -s dir --log error \
 --directories {} \
 --directories {} \
 --directories {} \
+--rpm-attr 755,{},{}:{} \
+--rpm-attr 755,{},{}:{} \
 --description \"{}\"".format(
      VENDOR,
      PACKAGE_URL,
@@ -74,6 +79,8 @@ fpm_common_args = "-f -s dir --log error \
      LOG_DIR,
      DATA_DIR,
      MAN_DIR,
+     USER, GROUP, LOG_DIR,
+     USER, GROUP, DATA_DIR,
      DESCRIPTION)
 
 for f in CONFIGURATION_FILES:
@@ -84,7 +91,7 @@ targets = {
     'influxd' : './cmd/influxd',
     'influx_stress' : './cmd/influx_stress',
     'influx_inspect' : './cmd/influx_inspect',
-    'influx_tsm' : './cmd/influx_tsm',
+    'influx_tools': './cmd/influx_tools',
 }
 
 supported_builds = {
@@ -143,6 +150,8 @@ def package_scripts(build_root, config_only=False, windows=False):
         os.chmod(os.path.join(build_root, SCRIPT_DIR[1:], INIT_SCRIPT.split('/')[1]), 0o644)
         shutil.copyfile(SYSTEMD_SCRIPT, os.path.join(build_root, SCRIPT_DIR[1:], SYSTEMD_SCRIPT.split('/')[1]))
         os.chmod(os.path.join(build_root, SCRIPT_DIR[1:], SYSTEMD_SCRIPT.split('/')[1]), 0o644)
+        shutil.copyfile(SYSTEMD_START_SCRIPT, os.path.join(build_root, SCRIPT_DIR[1:], SYSTEMD_START_SCRIPT.split('/')[1]))
+        os.chmod(os.path.join(build_root, SCRIPT_DIR[1:], SYSTEMD_START_SCRIPT.split('/')[1]), 0o755)
         shutil.copyfile(LOGROTATE_SCRIPT, os.path.join(build_root, LOGROTATE_DIR[1:], "influxdb"))
         os.chmod(os.path.join(build_root, LOGROTATE_DIR[1:], "influxdb"), 0o644)
         shutil.copyfile(DEFAULT_CONFIG, os.path.join(build_root, CONFIG_DIR[1:], "influxdb.conf"))
@@ -162,16 +171,9 @@ def go_get(branch, update=False, no_uncommitted=False):
     if local_changes() and no_uncommitted:
         logging.error("There are uncommitted changes in the current directory.")
         return False
-    if not check_path_for("gdm"):
-        logging.info("Downloading `gdm`...")
-        get_command = "go get github.com/sparrc/gdm"
-        run(get_command)
-    logging.info("Retrieving dependencies with `gdm`...")
-    sys.stdout.flush()
-    run("{}/bin/gdm restore -v".format(os.environ.get("GOPATH")))
     return True
 
-def run_tests(race, parallel, timeout, no_vet, junit=False):
+def run_tests(race, parallel, timeout, no_vet, run_gen, junit=False):
     """Run the Go test suite on binary output.
     """
     logging.info("Starting tests...")
@@ -181,11 +183,27 @@ def run_tests(race, parallel, timeout, no_vet, junit=False):
         logging.info("Using parallel: {}".format(parallel))
     if timeout is not None:
         logging.info("Using timeout: {}".format(timeout))
+
+    logging.info("Fetching module dependencies...")
+    run("go mod download")
+
+    logging.info("Ensuring code is properly formatted with go fmt...")
     out = run("go fmt ./...")
     if len(out) > 0:
         logging.error("Code not formatted. Please use 'go fmt ./...' to fix formatting errors.")
         logging.error("{}".format(out))
         return False
+
+    if run_gen:
+        logging.info("Ensuring codegen is up to date ...")
+        out = run("./generate.sh")
+        if len(out) > 0:
+            logging.error("Please use 'go generate ./...' to regenerate generated code.")
+            logging.error("{}".format(out))
+            return False
+    else:
+      logging.info("Skipping codegen check")
+
     if not no_vet:
         logging.info("Running 'go vet'...")
         out = run(go_vet_command)
@@ -385,8 +403,8 @@ def check_environ(build_dir = None):
         logging.debug("Using '{}' for {}".format(os.environ.get(v), v))
 
     cwd = os.getcwd()
-    if build_dir is None and os.environ.get("GOPATH") and os.environ.get("GOPATH") not in cwd:
-        logging.warn("Your current directory is not under your GOPATH. This may lead to build failures.")
+    if build_dir is None and os.environ.get("GOPATH") and os.environ.get("GOPATH") in cwd:
+        logging.warn("Your current directory is under your GOPATH. This may lead to build failures when using modules.")
     return True
 
 def check_prereqs():
@@ -704,7 +722,12 @@ def package(build_output, pkg_name, version, nightly=False, iteration=1, static=
                     elif package_type not in ['zip', 'tar'] and static or "static_" in arch:
                         logging.info("Skipping package type '{}' for static builds.".format(package_type))
                     else:
-                        fpm_command = "fpm {} --name {} -a {} -t {} --version {} --iteration {} -C {} -p {} ".format(
+                        if package_type == "rpm":
+                            if package_arch == "armhf":
+                                package_arch = "armv7hl"
+                            elif package_arch == "arm64":
+                                package_arch = "aarch64"
+                        fpm_command = "fpm {} --name {} -a {} -t {} --version {} --iteration {} -C {} -p {} --depends curl ".format(
                             fpm_common_args,
                             name,
                             package_arch,
@@ -788,7 +811,7 @@ def main(args):
             return 1
 
     if args.test:
-        if not run_tests(args.race, args.parallel, args.timeout, args.no_vet, args.junit_report):
+        if not run_tests(args.race, args.parallel, args.timeout, args.no_vet, args.arch == "amd64", args.junit_report):
             return 1
 
     platforms = []
