@@ -31,6 +31,7 @@ const (
 
 	// DefaultTimeout is the default connection timeout used to connect to an InfluxDB instance
 	DefaultTimeout = 0
+	DefaultPath    = ""
 )
 
 // Query is used to send a command to the server. Both Command and Database are required.
@@ -38,6 +39,10 @@ type Query struct {
 	Command  string
 	Database string
 
+	// RetentionPolicy tells the server which retention policy to use by default.
+	// This option is only effective when querying a server of version 1.6.0 or later.
+	RetentionPolicy string
+
 	// Chunked tells the server to send back chunked responses. This places
 	// less load on the server by sending back chunks of the response rather
 	// than waiting for the entire response all at once.
@@ -60,23 +65,36 @@ type Query struct {
 	NodeID int
 }
 
+// SplitPath gets the path of a url
+func SplitPath(v string) (string, string) {
+	i := strings.Index(v, "/")
+	if i == -1 {
+		return v, ""
+	}
+	return v[:i] /* first */, v[i+1:] /* rest */
+}
+
 // ParseConnectionString will parse a string to create a valid connection URL
 func ParseConnectionString(path string, ssl bool) (url.URL, error) {
 	var host string
 	var port int
+	var pth string = ""
 
 	h, p, err := net.SplitHostPort(path)
 	if err != nil {
 		if path == "" {
 			host = DefaultHost
 		} else {
-			host = path
+			host, pth = SplitPath(path)
 		}
 		// If they didn't specify a port, always use the default port
 		port = DefaultPort
 	} else {
 		host = h
-		port, err = strconv.Atoi(p)
+		prt, pt := SplitPath(p)
+		pth = pt
+
+		port, err = strconv.Atoi(prt)
 		if err != nil {
 			return url.URL{}, fmt.Errorf("invalid port number %q: %s\n", path, err)
 		}
@@ -85,6 +103,7 @@ func ParseConnectionString(path string, ssl bool) (url.URL, error) {
 	u := url.URL{
 		Scheme: "http",
 		Host:   host,
+		Path:   pth,
 	}
 	if ssl {
 		u.Scheme = "https"
@@ -113,6 +132,8 @@ type Config struct {
 	Precision        string
 	WriteConsistency string
 	UnsafeSsl        bool
+	Proxy            func(req *http.Request) (*url.URL, error)
+	TLS              *tls.Config
 }
 
 // NewConfig will create a config to be used in connecting to the client
@@ -149,11 +170,14 @@ const (
 
 // NewClient will instantiate and return a connected client to issue commands to the server.
 func NewClient(c Config) (*Client, error) {
-	tlsConfig := &tls.Config{
-		InsecureSkipVerify: c.UnsafeSsl,
+	tlsConfig := new(tls.Config)
+	if c.TLS != nil {
+		tlsConfig = c.TLS.Clone()
 	}
+	tlsConfig.InsecureSkipVerify = c.UnsafeSsl
 
 	tr := &http.Transport{
+		Proxy:           c.Proxy,
 		TLSClientConfig: tlsConfig,
 	}
 
@@ -206,6 +230,9 @@ func (c *Client) QueryContext(ctx context.Context, q Query) (*Response, error) {
 	values := u.Query()
 	values.Set("q", q.Command)
 	values.Set("db", q.Database)
+	if q.RetentionPolicy != "" {
+		values.Set("rp", q.RetentionPolicy)
+	}
 	if q.Chunked {
 		values.Set("chunked", "true")
 		if q.ChunkSize > 0 {
