@@ -102,11 +102,16 @@ func NewTest(t *testing.T) *Test {
 			tagsSlice = append(tagsSlice, nil)
 		}
 
-		_, err := seriesFile.CreateSeriesListIfNotExists(names, tagsSlice)
+		ids, err := seriesFile.CreateSeriesListIfNotExists(names, tagsSlice)
 		if err != nil {
 			return err
 		}
 
+		// delete one series
+		if err := seriesFile.DeleteSeriesID(ids[0]); err != nil {
+			return err
+		}
+
 		// wait for compaction to make sure we detect issues with the index
 		partitions := seriesFile.Partitions()
 	wait:
