@@ -2,7 +2,6 @@
 package backup
 
 import (
-	"compress/gzip"
 	"encoding/binary"
 	"encoding/json"
 	"errors"
@@ -20,6 +19,7 @@ import (
 	"github.com/influxdata/influxdb/cmd/influxd/backup_util"
 	"github.com/influxdata/influxdb/services/snapshotter"
 	"github.com/influxdata/influxdb/tcp"
+	gzip "github.com/klauspost/pgzip"
 )
 
 const (
@@ -58,6 +58,7 @@ type Command struct {
 	portable         bool
 	manifest         backup_util.Manifest
 	portableFileBase string
+	continueOnError  bool
 
 	BackupFiles []string
 }
@@ -146,7 +147,9 @@ func (cmd *Command) parseFlags(args []string) (err error) {
 
 	fs.StringVar(&cmd.host, "host", "localhost:8088", "")
 	fs.StringVar(&cmd.database, "database", "", "")
+	fs.StringVar(&cmd.database, "db", "", "")
 	fs.StringVar(&cmd.retentionPolicy, "retention", "", "")
+	fs.StringVar(&cmd.retentionPolicy, "rp", "", "")
 	fs.StringVar(&cmd.shardID, "shard", "", "")
 	var sinceArg string
 	var startArg string
@@ -155,6 +158,7 @@ func (cmd *Command) parseFlags(args []string) (err error) {
 	fs.StringVar(&startArg, "start", "", "")
 	fs.StringVar(&endArg, "end", "", "")
 	fs.BoolVar(&cmd.portable, "portable", false, "")
+	fs.BoolVar(&cmd.continueOnError, "skip-errors", false, "")
 
 	fs.SetOutput(cmd.Stderr)
 	fs.Usage = cmd.printUsage
@@ -249,14 +253,14 @@ func (cmd *Command) backupShard(db, rp, sid string) error {
 
 	// TODO: verify shard backup data
 	err = cmd.downloadAndVerify(req, shardArchivePath, nil)
+	if err != nil {
+		os.Remove(shardArchivePath)
+		return err
+	}
 	if !cmd.portable {
 		cmd.BackupFiles = append(cmd.BackupFiles, shardArchivePath)
 	}
 
-	if err != nil {
-		return err
-	}
-
 	if cmd.portable {
 		f, err := os.Open(shardArchivePath)
 		if err != nil {
@@ -377,7 +381,8 @@ func (cmd *Command) backupResponsePaths(response *snapshotter.Response) error {
 
 		err = cmd.backupShard(db, rp, id)
 
-		if err != nil {
+		if err != nil && !cmd.continueOnError {
+			cmd.StderrLogger.Printf("error (%s) when backing up db: %s, rp %s, shard %s. continuing backup on remaining shards", err, db, rp, id)
 			return err
 		}
 	}
@@ -536,9 +541,15 @@ func (cmd *Command) download(req *snapshotter.Request, path string) error {
 			}
 
 			// Read snapshot from the connection
-			if n, err := io.Copy(f, conn); err != nil || n == 0 {
+			n, err := io.Copy(f, conn)
+			if err != nil {
 				return fmt.Errorf("copy backup to file: err=%v, n=%d", err, n)
 			}
+			if n == 0 {
+				// Unfortunately there is no out-of-band channel to actually return errors from the snapshot service, just
+				// 'data' or 'no data'.
+				return fmt.Errorf("copy backup to file: no data returned, check server logs for snapshot errors")
+			}
 			return nil
 		}(); err == nil {
 			break
@@ -586,28 +597,35 @@ func (cmd *Command) requestInfo(request *snapshotter.Request) (*snapshotter.Resp
 // printUsage prints the usage message to STDERR.
 func (cmd *Command) printUsage() {
 	fmt.Fprintf(cmd.Stdout, `
-Downloads a file level age-based snapshot of a data node and saves it to disk.
+Creates a backup copy of specified InfluxDB OSS database(s) and saves the files in an Enterprise-compatible
+format to PATH (directory where backups are saved). 
 
-Usage: influxd backup [flags] PATH
+Usage: influxd backup [options] PATH
 
-    -host <host:port>
-            The host to connect to snapshot. Defaults to 127.0.0.1:8088.
-    -database <name>
-            The database to backup.
-    -retention <name>
-            Optional. The retention policy to backup.
-    -shard <id>
-            Optional. The shard id to backup. If specified, retention is required.
-    -since <2015-12-24T08:12:23Z>
-            Optional. Do an incremental backup since the passed in RFC3339
-            formatted time.  Not compatible with -start or -end.
-    -start <2015-12-24T08:12:23Z>
-            All points earlier than this time stamp will be excluded from the export. Not compatible with -since.
-    -end <2015-12-24T08:12:23Z>
-            All points later than this time stamp will be excluded from the export. Not compatible with -since.
     -portable
-            Generate backup files in a format that is portable between different influxdb products.
-
+            Required to generate backup files in a portable format that can be restored to InfluxDB OSS or InfluxDB 
+            Enterprise. Use unless the legacy backup is required.
+    -host <host:port>
+            InfluxDB OSS host to back up from. Optional. Defaults to 127.0.0.1:8088.
+    -db <name>
+            InfluxDB OSS database name to back up. Optional. If not specified, all databases are backed up when 
+            using '-portable'.
+    -rp <name>
+            Retention policy to use for the backup. Optional. If not specified, all retention policies are used by 
+            default.
+    -shard <id>
+            The identifier of the shard to back up. Optional. If specified, '-rp <rp_name>' is required.
+    -start <2015-12-24T08:12:23Z>
+            Include all points starting with specified timestamp (RFC3339 format). 
+            Not compatible with '-since <timestamp>'.
+    -end <2015-12-24T08:12:23Z>
+            Exclude all points after timestamp (RFC3339 format). 
+            Not compatible with '-since <timestamp>'.
+    -since <2015-12-24T08:12:23Z>
+            Create an incremental backup of all points after the timestamp (RFC3339 format). Optional. 
+            Recommend using '-start <timestamp>' instead.
+    -skip-errors 
+            Optional flag to continue backing up the remaining shards when the current shard fails to backup. 
 `)
 
 }
