@@ -1,6 +1,8 @@
 package run
 
 import (
+	"context"
+	"crypto/tls"
 	"fmt"
 	"io"
 	"log"
@@ -13,6 +15,7 @@ import (
 
 	"github.com/influxdata/influxdb"
 	"github.com/influxdata/influxdb/coordinator"
+	"github.com/influxdata/influxdb/flux/control"
 	"github.com/influxdata/influxdb/logger"
 	"github.com/influxdata/influxdb/models"
 	"github.com/influxdata/influxdb/monitor"
@@ -26,16 +29,18 @@ import (
 	"github.com/influxdata/influxdb/services/precreator"
 	"github.com/influxdata/influxdb/services/retention"
 	"github.com/influxdata/influxdb/services/snapshotter"
+	"github.com/influxdata/influxdb/services/storage"
 	"github.com/influxdata/influxdb/services/subscriber"
 	"github.com/influxdata/influxdb/services/udp"
+	"github.com/influxdata/influxdb/storage/reads"
 	"github.com/influxdata/influxdb/tcp"
 	"github.com/influxdata/influxdb/tsdb"
 	client "github.com/influxdata/usage-client/v1"
 	"go.uber.org/zap"
 
-	// Initialize the engine & index packages
-	"github.com/influxdata/influxdb/services/storage"
+	// Initialize the engine package
 	_ "github.com/influxdata/influxdb/tsdb/engine"
+	// Initialize the index package
 	_ "github.com/influxdata/influxdb/tsdb/index"
 )
 
@@ -70,7 +75,7 @@ type Server struct {
 	MetaClient *meta.Client
 
 	TSDBStore     *tsdb.Store
-	QueryExecutor *query.QueryExecutor
+	QueryExecutor *query.Executor
 	PointsWriter  *coordinator.PointsWriter
 	Subscriber    *subscriber.Service
 
@@ -85,8 +90,10 @@ type Server struct {
 	reportingDisabled bool
 
 	// Profiling
-	CPUProfile string
-	MemProfile string
+	CPUProfile            string
+	CPUProfileWriteCloser io.WriteCloser
+	MemProfile            string
+	MemProfileWriteCloser io.WriteCloser
 
 	// httpAPIAddr is the host:port combination for the main HTTP API for querying and writing data
 	httpAPIAddr string
@@ -100,8 +107,30 @@ type Server struct {
 	config *Config
 }
 
+// updateTLSConfig stores with into the tls config pointed at by into but only if with is not nil
+// and into is nil. Think of it as setting the default value.
+func updateTLSConfig(into **tls.Config, with *tls.Config) {
+	if with != nil && into != nil && *into == nil {
+		*into = with
+	}
+}
+
 // NewServer returns a new instance of Server built from a config.
 func NewServer(c *Config, buildInfo *BuildInfo) (*Server, error) {
+	// First grab the base tls config we will use for all clients and servers
+	tlsConfig, err := c.TLS.Parse()
+	if err != nil {
+		return nil, fmt.Errorf("tls configuration: %v", err)
+	}
+
+	// Update the TLS values on each of the configs to be the parsed one if
+	// not already specified (set the default).
+	updateTLSConfig(&c.HTTPD.TLS, tlsConfig)
+	updateTLSConfig(&c.Subscriber.TLS, tlsConfig)
+	for i := range c.OpenTSDBInputs {
+		updateTLSConfig(&c.OpenTSDBInputs[i].TLS, tlsConfig)
+	}
+
 	// We need to ensure that a meta directory always exists even if
 	// we don't start the meta store.  node.json is always stored under
 	// the meta directory.
@@ -121,7 +150,7 @@ func NewServer(c *Config, buildInfo *BuildInfo) (*Server, error) {
 		}
 	}
 
-	_, err := influxdb.LoadNode(c.Meta.Dir)
+	_, err = influxdb.LoadNode(c.Meta.Dir)
 	if err != nil {
 		if !os.IsNotExist(err) {
 			return nil, err
@@ -178,20 +207,21 @@ func NewServer(c *Config, buildInfo *BuildInfo) (*Server, error) {
 	s.PointsWriter.TSDBStore = s.TSDBStore
 
 	// Initialize query executor.
-	s.QueryExecutor = query.NewQueryExecutor()
+	s.QueryExecutor = query.NewExecutor()
 	s.QueryExecutor.StatementExecutor = &coordinator.StatementExecutor{
 		MetaClient:  s.MetaClient,
 		TaskManager: s.QueryExecutor.TaskManager,
-		TSDBStore:   coordinator.LocalTSDBStore{Store: s.TSDBStore},
+		TSDBStore:   s.TSDBStore,
 		ShardMapper: &coordinator.LocalShardMapper{
 			MetaClient: s.MetaClient,
 			TSDBStore:  coordinator.LocalTSDBStore{Store: s.TSDBStore},
 		},
-		Monitor:           s.Monitor,
-		PointsWriter:      s.PointsWriter,
-		MaxSelectPointN:   c.Coordinator.MaxSelectPointN,
-		MaxSelectSeriesN:  c.Coordinator.MaxSelectSeriesN,
-		MaxSelectBucketsN: c.Coordinator.MaxSelectBucketsN,
+		StrictErrorHandling: s.TSDBStore.EngineOptions.Config.StrictErrorHandling,
+		Monitor:             s.Monitor,
+		PointsWriter:        s.PointsWriter,
+		MaxSelectPointN:     c.Coordinator.MaxSelectPointN,
+		MaxSelectSeriesN:    c.Coordinator.MaxSelectSeriesN,
+		MaxSelectBucketsN:   c.Coordinator.MaxSelectBucketsN,
 	}
 	s.QueryExecutor.TaskManager.QueryTimeout = time.Duration(c.Coordinator.QueryTimeout)
 	s.QueryExecutor.TaskManager.LogQueriesAfter = time.Duration(c.Coordinator.LogQueriesAfter)
@@ -255,24 +285,19 @@ func (s *Server) appendHTTPDService(c httpd.Config) {
 	}
 	srv := httpd.NewService(c)
 	srv.Handler.MetaClient = s.MetaClient
-	srv.Handler.QueryAuthorizer = meta.NewQueryAuthorizer(s.MetaClient)
+	authorizer := meta.NewQueryAuthorizer(s.MetaClient)
+	srv.Handler.QueryAuthorizer = authorizer
 	srv.Handler.WriteAuthorizer = meta.NewWriteAuthorizer(s.MetaClient)
 	srv.Handler.QueryExecutor = s.QueryExecutor
 	srv.Handler.Monitor = s.Monitor
 	srv.Handler.PointsWriter = s.PointsWriter
 	srv.Handler.Version = s.buildInfo.Version
 	srv.Handler.BuildType = "OSS"
-
-	s.Services = append(s.Services, srv)
-}
-
-func (s *Server) appendStorageService(c storage.Config) {
-	if !c.Enabled {
-		return
+	ss := storage.NewStore(s.TSDBStore, s.MetaClient)
+	srv.Handler.Store = ss
+	if s.config.HTTPD.FluxEnabled {
+		srv.Handler.Controller = control.NewController(s.MetaClient, reads.NewReader(ss), authorizer, c.AuthEnabled, s.Logger)
 	}
-	srv := storage.NewService(c)
-	srv.MetaClient = s.MetaClient
-	srv.TSDBStore = s.TSDBStore
 
 	s.Services = append(s.Services, srv)
 }
@@ -353,8 +378,10 @@ func (s *Server) Err() <-chan error { return s.err }
 
 // Open opens the meta and data store and all services.
 func (s *Server) Open() error {
-	// Start profiling, if set.
-	startProfile(s.CPUProfile, s.MemProfile)
+	// Start profiling if requested.
+	if err := s.startProfile(); err != nil {
+		return err
+	}
 
 	// Open shared TCP connection.
 	ln, err := net.Listen("tcp", s.BindAddress)
@@ -373,7 +400,6 @@ func (s *Server) Open() error {
 	s.appendSnapshotterService()
 	s.appendContinuousQueryService(s.config.ContinuousQuery)
 	s.appendHTTPDService(s.config.HTTPD)
-	s.appendStorageService(s.config.Storage)
 	s.appendRetentionPolicyService(s.config.Retention)
 	for _, i := range s.config.GraphiteInputs {
 		if err := s.appendGraphiteService(i); err != nil {
@@ -419,7 +445,7 @@ func (s *Server) Open() error {
 		return fmt.Errorf("open tsdb store: %s", err)
 	}
 
-	// Open the subcriber service
+	// Open the subscriber service
 	if err := s.Subscriber.Open(); err != nil {
 		return fmt.Errorf("open subscriber: %s", err)
 	}
@@ -447,7 +473,7 @@ func (s *Server) Open() error {
 
 // Close shuts down the meta and data stores and all services.
 func (s *Server) Close() error {
-	stopProfile()
+	s.stopProfile()
 
 	// Close the listener first to stop any new connections
 	if s.Listener != nil {
@@ -515,14 +541,16 @@ func (s *Server) reportServer() {
 
 	for _, db := range dbs {
 		name := db.Name
-		n, err := s.TSDBStore.SeriesCardinality(name)
+		// Use the context.Background() to avoid timing out on this.
+		n, err := s.TSDBStore.SeriesCardinality(context.Background(), name)
 		if err != nil {
 			s.Logger.Error(fmt.Sprintf("Unable to get series cardinality for database %s: %v", name, err))
 		} else {
 			numSeries += n
 		}
 
-		n, err = s.TSDBStore.MeasurementsCardinality(name)
+		// Use the context.Background() to avoid timing out on this.
+		n, err = s.TSDBStore.MeasurementsCardinality(context.Background(), name)
 		if err != nil {
 			s.Logger.Error(fmt.Sprintf("Unable to get measurement cardinality for database %s: %v", name, err))
 		} else {
@@ -563,47 +591,56 @@ type Service interface {
 }
 
 // prof stores the file locations of active profiles.
-var prof struct {
-	cpu *os.File
-	mem *os.File
-}
-
 // StartProfile initializes the cpu and memory profile, if specified.
-func startProfile(cpuprofile, memprofile string) {
-	if cpuprofile != "" {
-		f, err := os.Create(cpuprofile)
+func (s *Server) startProfile() error {
+	if s.CPUProfile != "" {
+		f, err := os.Create(s.CPUProfile)
 		if err != nil {
-			log.Fatalf("cpuprofile: %v", err)
+			return fmt.Errorf("cpuprofile: %v", err)
 		}
-		log.Printf("writing CPU profile to: %s\n", cpuprofile)
-		prof.cpu = f
-		pprof.StartCPUProfile(prof.cpu)
+
+		s.CPUProfileWriteCloser = f
+		if err := pprof.StartCPUProfile(s.CPUProfileWriteCloser); err != nil {
+			return err
+		}
+
+		log.Printf("Writing CPU profile to: %s\n", s.CPUProfile)
 	}
 
-	if memprofile != "" {
-		f, err := os.Create(memprofile)
+	if s.MemProfile != "" {
+		f, err := os.Create(s.MemProfile)
 		if err != nil {
-			log.Fatalf("memprofile: %v", err)
+			return fmt.Errorf("memprofile: %v", err)
 		}
-		log.Printf("writing mem profile to: %s\n", memprofile)
-		prof.mem = f
+
+		s.MemProfileWriteCloser = f
 		runtime.MemProfileRate = 4096
+
+		log.Printf("Writing mem profile to: %s\n", s.MemProfile)
 	}
 
+	return nil
 }
 
 // StopProfile closes the cpu and memory profiles if they are running.
-func stopProfile() {
-	if prof.cpu != nil {
+func (s *Server) stopProfile() error {
+	if s.CPUProfileWriteCloser != nil {
 		pprof.StopCPUProfile()
-		prof.cpu.Close()
+		if err := s.CPUProfileWriteCloser.Close(); err != nil {
+			return err
+		}
 		log.Println("CPU profile stopped")
 	}
-	if prof.mem != nil {
-		pprof.Lookup("heap").WriteTo(prof.mem, 0)
-		prof.mem.Close()
+
+	if s.MemProfileWriteCloser != nil {
+		pprof.Lookup("heap").WriteTo(s.MemProfileWriteCloser, 0)
+		if err := s.MemProfileWriteCloser.Close(); err != nil {
+			return err
+		}
 		log.Println("mem profile stopped")
 	}
+
+	return nil
 }
 
 // monitorPointsWriter is a wrapper around `coordinator.PointsWriter` that helps
