@@ -784,6 +784,7 @@ func (ms *MetaClient) CreateDatabase(name, defaultRetentionPolicy string) error
 	// Create database.
 	ms.DatabaseInfos = append(ms.DatabaseInfos, meta.DatabaseInfo{
 		Name: name,
+
 		DefaultRetentionPolicy: defaultRetentionPolicy,
 	})
 
@@ -824,7 +825,7 @@ type StatementExecutor struct {
 	ExecuteStatementFn func(stmt influxql.Statement, ctx *query.ExecutionContext) error
 }
 
-func (e *StatementExecutor) ExecuteStatement(stmt influxql.Statement, ctx *query.ExecutionContext) error {
+func (e *StatementExecutor) ExecuteStatement(ctx *query.ExecutionContext, stmt influxql.Statement) error {
 	return e.ExecuteStatementFn(stmt, ctx)
 }
 
