@@ -101,21 +101,56 @@ func buildLogLine(l *responseLogger, r *http.Request, start time.Time) string {
 
 	userAgent := r.UserAgent()
 
-	return fmt.Sprintf(`%s - %s [%s] "%s %s %s" %s %s "%s" "%s" %s %d`,
-		host,
-		detect(username, "-"),
-		start.Format("02/Jan/2006:15:04:05 -0700"),
-		r.Method,
-		uri,
-		r.Proto,
-		detect(strconv.Itoa(l.Status()), "-"),
-		strconv.Itoa(l.Size()),
-		detect(referer, "-"),
-		detect(userAgent, "-"),
-		r.Header.Get("Request-Id"),
-		// response time, report in microseconds because this is consistent
-		// with apache's %D parameter in mod_log_config
-		int64(time.Since(start)/time.Microsecond))
+	allKeyValues := make([]string, 0, len(r.PostForm))
+	if r.Method == "POST" {
+		for k, values := range r.PostForm {
+			if k == "p" || k == "P" {
+				// Note: if there are multiple "p" values, they are all replaced by a single "[REDACTED]".
+				r.PostForm.Set(k, "[REDACTED]")
+				values = r.PostForm[k]
+			}
+			valuesSlice := make([]string, 0, len(values))
+			for _, v := range values {
+				valuesSlice = append(valuesSlice, fmt.Sprintf("'%s'", v))
+			}
+			joined := strings.Join(valuesSlice, ", ")
+			allKeyValues = append(allKeyValues, fmt.Sprintf("{'%s': %s}", k, joined))
+		}
+
+		return fmt.Sprintf(`%s - %s [%s] "%s %s %s %s" %s %s "%s" "%s" %s %d`,
+			host,
+			detect(username, "-"),
+			start.Format("02/Jan/2006:15:04:05 -0700"),
+			r.Method,
+			uri,
+			r.Proto,
+			strings.Join(allKeyValues, ", "),
+			detect(strconv.Itoa(l.Status()), "-"),
+			strconv.Itoa(l.Size()),
+			detect(referer, "-"),
+			detect(userAgent, "-"),
+			r.Header.Get("Request-Id"),
+			// response time, report in microseconds because this is consistent
+			// with apache's %D parameter in mod_log_config
+			int64(time.Since(start)/time.Microsecond))
+
+	} else {
+		return fmt.Sprintf(`%s - %s [%s] "%s %s %s" %s %s "%s" "%s" %s %d`,
+			host,
+			detect(username, "-"),
+			start.Format("02/Jan/2006:15:04:05 -0700"),
+			r.Method,
+			uri,
+			r.Proto,
+			detect(strconv.Itoa(l.Status()), "-"),
+			strconv.Itoa(l.Size()),
+			detect(referer, "-"),
+			detect(userAgent, "-"),
+			r.Header.Get("Request-Id"),
+			// response time, report in microseconds because this is consistent
+			// with apache's %D parameter in mod_log_config
+			int64(time.Since(start)/time.Microsecond))
+	}
 }
 
 // detect detects the first presence of a non blank string and returns it
