@@ -3,6 +3,7 @@ package httpd
 import (
 	"encoding/csv"
 	"encoding/json"
+	"errors"
 	"io"
 	"net/http"
 	"strconv"
@@ -20,27 +21,57 @@ type ResponseWriter interface {
 	http.ResponseWriter
 }
 
+type formatter interface {
+	WriteResponse(w io.Writer, resp Response) error
+}
+
+type supportedContentType struct {
+	full          string
+	acceptType    string
+	acceptSubType string
+	formatter     func(pretty bool) formatter
+}
+
+var (
+	csvFormatFactory     = func(pretty bool) formatter { return &csvFormatter{statementID: -1} }
+	msgpackFormatFactory = func(pretty bool) formatter { return &msgpackFormatter{} }
+	jsonFormatFactory    = func(pretty bool) formatter { return &jsonFormatter{Pretty: pretty} }
+
+	contentTypes = []supportedContentType{
+		{full: "application/json", acceptType: "application", acceptSubType: "json", formatter: jsonFormatFactory},
+		{full: "application/csv", acceptType: "application", acceptSubType: "csv", formatter: csvFormatFactory},
+		{full: "text/csv", acceptType: "text", acceptSubType: "csv", formatter: csvFormatFactory},
+		{full: "application/x-msgpack", acceptType: "application", acceptSubType: "x-msgpack", formatter: msgpackFormatFactory},
+	}
+	defaultContentType = contentTypes[0]
+)
+
 // NewResponseWriter creates a new ResponseWriter based on the Accept header
 // in the request that wraps the ResponseWriter.
 func NewResponseWriter(w http.ResponseWriter, r *http.Request) ResponseWriter {
 	pretty := r.URL.Query().Get("pretty") == "true"
 	rw := &responseWriter{ResponseWriter: w}
-	switch r.Header.Get("Accept") {
-	case "application/csv", "text/csv":
-		w.Header().Add("Content-Type", "text/csv")
-		rw.formatter = &csvFormatter{statementID: -1, Writer: w}
-	case "application/x-msgpack":
-		w.Header().Add("Content-Type", "application/x-msgpack")
-		rw.formatter = &msgpackFormatter{Writer: w}
-	case "application/json":
-		fallthrough
-	default:
-		w.Header().Add("Content-Type", "application/json")
-		rw.formatter = &jsonFormatter{Pretty: pretty, Writer: w}
+
+	acceptHeaders := parseAccept(r.Header["Accept"])
+	for _, accept := range acceptHeaders {
+		for _, ct := range contentTypes {
+			if match(accept, ct) {
+				w.Header().Add("Content-Type", ct.full)
+				rw.formatter = ct.formatter(pretty)
+				return rw
+			}
+		}
 	}
+	w.Header().Add("Content-Type", defaultContentType.full)
+	rw.formatter = defaultContentType.formatter(pretty)
 	return rw
 }
 
+func match(ah accept, sct supportedContentType) bool {
+	return (ah.Type == "*" || ah.Type == sct.acceptType) &&
+		(ah.SubType == "*" || ah.SubType == sct.acceptSubType)
+}
+
 // WriteError is a convenience function for writing an error response to the ResponseWriter.
 func WriteError(w ResponseWriter, err error) (int, error) {
 	return w.WriteResponse(Response{Err: err})
@@ -49,14 +80,27 @@ func WriteError(w ResponseWriter, err error) (int, error) {
 // responseWriter is an implementation of ResponseWriter.
 type responseWriter struct {
 	formatter interface {
-		WriteResponse(resp Response) (int, error)
+		WriteResponse(w io.Writer, resp Response) error
 	}
 	http.ResponseWriter
 }
 
+type bytesCountWriter struct {
+	w io.Writer
+	n int
+}
+
+func (w *bytesCountWriter) Write(data []byte) (int, error) {
+	n, err := w.w.Write(data)
+	w.n += n
+	return n, err
+}
+
 // WriteResponse writes the response using the formatter.
 func (w *responseWriter) WriteResponse(resp Response) (int, error) {
-	return w.formatter.WriteResponse(resp)
+	writer := bytesCountWriter{w: w.ResponseWriter}
+	err := w.formatter.WriteResponse(&writer, resp)
+	return writer.n, err
 }
 
 // Flush flushes the ResponseWriter if it has a Flush() method.
@@ -76,137 +120,172 @@ func (w *responseWriter) CloseNotify() <-chan bool {
 }
 
 type jsonFormatter struct {
-	io.Writer
 	Pretty bool
 }
 
-func (w *jsonFormatter) WriteResponse(resp Response) (n int, err error) {
+func (f *jsonFormatter) WriteResponse(w io.Writer, resp Response) error {
 	var b []byte
-	if w.Pretty {
+	var err error
+	if f.Pretty {
 		b, err = json.MarshalIndent(resp, "", "    ")
 	} else {
 		b, err = json.Marshal(resp)
 	}
 
 	if err != nil {
-		n, err = io.WriteString(w, err.Error())
-	} else {
-		n, err = w.Write(b)
+		unnestedErr := unnestError(err)
+		// ignore any errors in this section, we already have a 'real' error to return
+		resp := Response{Err: unnestedErr}
+		if f.Pretty {
+			b, _ = json.MarshalIndent(resp, "", "    ")
+		} else {
+			b, _ = json.Marshal(resp)
+		}
+		w.Write(b)
+		w.Write([]byte("\n"))
+		return err
 	}
 
-	w.Write([]byte("\n"))
-	n++
-	return n, err
+	_, err = w.Write(b)
+	if err != nil {
+		return err
+	}
+	_, err = w.Write([]byte("\n"))
+	return err
+}
+
+func unnestError(err error) error {
+	for errNested := err; errNested != nil; errNested = errors.Unwrap(err) {
+		err = errNested
+	}
+	return err
 }
 
 type csvFormatter struct {
-	io.Writer
 	statementID int
 	columns     []string
 }
 
-func (w *csvFormatter) WriteResponse(resp Response) (n int, err error) {
+func (f *csvFormatter) WriteResponse(w io.Writer, resp Response) (err error) {
 	csv := csv.NewWriter(w)
 	if resp.Err != nil {
 		csv.Write([]string{"error"})
 		csv.Write([]string{resp.Err.Error()})
 		csv.Flush()
-		return n, csv.Error()
+		return csv.Error()
 	}
 
 	for _, result := range resp.Results {
-		if result.StatementID != w.statementID {
+		if result.StatementID != f.statementID {
 			// If there are no series in the result, skip past this result.
 			if len(result.Series) == 0 {
 				continue
 			}
 
 			// Set the statement id and print out a newline if this is not the first statement.
-			if w.statementID >= 0 {
+			if f.statementID >= 0 {
 				// Flush the csv writer and write a newline.
 				csv.Flush()
 				if err := csv.Error(); err != nil {
-					return n, err
+					return err
 				}
 
-				out, err := io.WriteString(w, "\n")
-				if err != nil {
-					return n, err
+				if _, err := io.WriteString(w, "\n"); err != nil {
+					return err
 				}
-				n += out
 			}
-			w.statementID = result.StatementID
+			f.statementID = result.StatementID
 
 			// Print out the column headers from the first series.
-			w.columns = make([]string, 2+len(result.Series[0].Columns))
-			w.columns[0] = "name"
-			w.columns[1] = "tags"
-			copy(w.columns[2:], result.Series[0].Columns)
-			if err := csv.Write(w.columns); err != nil {
-				return n, err
+			f.columns = make([]string, 2+len(result.Series[0].Columns))
+			f.columns[0] = "name"
+			f.columns[1] = "tags"
+			copy(f.columns[2:], result.Series[0].Columns)
+			if err := csv.Write(f.columns); err != nil {
+				return err
 			}
 		}
 
-		for _, row := range result.Series {
-			w.columns[0] = row.Name
+		for i, row := range result.Series {
+			if i > 0 && !stringsEqual(result.Series[i-1].Columns, row.Columns) {
+				// The columns have changed. Print a newline and reprint the header.
+				csv.Flush()
+				if err := csv.Error(); err != nil {
+					return err
+				}
+
+				if _, err := io.WriteString(w, "\n"); err != nil {
+					return err
+				}
+
+				f.columns = make([]string, 2+len(row.Columns))
+				f.columns[0] = "name"
+				f.columns[1] = "tags"
+				copy(f.columns[2:], row.Columns)
+				if err := csv.Write(f.columns); err != nil {
+					return err
+				}
+			}
+
+			f.columns[0] = row.Name
+			f.columns[1] = ""
 			if len(row.Tags) > 0 {
-				w.columns[1] = string(models.NewTags(row.Tags).HashKey()[1:])
-			} else {
-				w.columns[1] = ""
+				hashKey := models.NewTags(row.Tags).HashKey()
+				if len(hashKey) > 0 {
+					f.columns[1] = string(hashKey[1:])
+				}
 			}
 			for _, values := range row.Values {
 				for i, value := range values {
 					if value == nil {
-						w.columns[i+2] = ""
+						f.columns[i+2] = ""
 						continue
 					}
 
 					switch v := value.(type) {
 					case float64:
-						w.columns[i+2] = strconv.FormatFloat(v, 'f', -1, 64)
+						f.columns[i+2] = strconv.FormatFloat(v, 'f', -1, 64)
 					case int64:
-						w.columns[i+2] = strconv.FormatInt(v, 10)
+						f.columns[i+2] = strconv.FormatInt(v, 10)
 					case uint64:
-						w.columns[i+2] = strconv.FormatUint(v, 10)
+						f.columns[i+2] = strconv.FormatUint(v, 10)
 					case string:
-						w.columns[i+2] = v
+						f.columns[i+2] = v
 					case bool:
 						if v {
-							w.columns[i+2] = "true"
+							f.columns[i+2] = "true"
 						} else {
-							w.columns[i+2] = "false"
+							f.columns[i+2] = "false"
 						}
 					case time.Time:
-						w.columns[i+2] = strconv.FormatInt(v.UnixNano(), 10)
+						f.columns[i+2] = strconv.FormatInt(v.UnixNano(), 10)
 					case *float64, *int64, *string, *bool:
-						w.columns[i+2] = ""
+						f.columns[i+2] = ""
 					}
 				}
-				csv.Write(w.columns)
+				csv.Write(f.columns)
 			}
 		}
 	}
 	csv.Flush()
-	return n, csv.Error()
+	return csv.Error()
 }
 
-type msgpackFormatter struct {
-	io.Writer
-}
+type msgpackFormatter struct{}
 
 func (f *msgpackFormatter) ContentType() string {
 	return "application/x-msgpack"
 }
 
-func (f *msgpackFormatter) WriteResponse(resp Response) (n int, err error) {
-	enc := msgp.NewWriter(f.Writer)
+func (f *msgpackFormatter) WriteResponse(w io.Writer, resp Response) (err error) {
+	enc := msgp.NewWriter(w)
 	defer enc.Flush()
 
 	enc.WriteMapHeader(1)
 	if resp.Err != nil {
 		enc.WriteString("error")
 		enc.WriteString(resp.Err.Error())
-		return 0, nil
+		return nil
 	} else {
 		enc.WriteString("results")
 		enc.WriteArrayHeader(uint32(len(resp.Results)))
@@ -289,5 +368,17 @@ func (f *msgpackFormatter) WriteResponse(resp Response) (n int, err error) {
 			}
 		}
 	}
-	return 0, nil
+	return nil
+}
+
+func stringsEqual(a, b []string) bool {
+	if len(a) != len(b) {
+		return false
+	}
+	for i := range a {
+		if a[i] != b[i] {
+			return false
+		}
+	}
+	return true
 }
