@@ -2,6 +2,7 @@ package meta
 
 import (
 	"errors"
+	"fmt"
 	"net"
 	"net/url"
 	"sort"
@@ -10,8 +11,6 @@ import (
 	"time"
 	"unicode"
 
-	"fmt"
-
 	"github.com/gogo/protobuf/proto"
 	"github.com/influxdata/influxdb"
 	"github.com/influxdata/influxdb/models"
@@ -34,6 +33,10 @@ const (
 
 	// MinRetentionPolicyDuration represents the minimum duration for a policy.
 	MinRetentionPolicyDuration = time.Hour
+
+	// MaxNameLen is the maximum length of a database or retention policy name.
+	// InfluxDB uses the name for the directory name on disk.
+	MaxNameLen = 255
 )
 
 // Data represents the top level collection of all metadata.
@@ -79,6 +82,8 @@ func (data *Data) CloneDatabases() []DatabaseInfo {
 func (data *Data) CreateDatabase(name string) error {
 	if name == "" {
 		return ErrDatabaseNameRequired
+	} else if len(name) > MaxNameLen {
+		return ErrNameTooLong
 	} else if data.Database(name) != nil {
 		return nil
 	}
@@ -129,6 +134,8 @@ func (data *Data) CreateRetentionPolicy(database string, rpi *RetentionPolicyInf
 		return ErrRetentionPolicyRequired
 	} else if rpi.Name == "" {
 		return ErrRetentionPolicyNameRequired
+	} else if len(rpi.Name) > MaxNameLen {
+		return ErrNameTooLong
 	} else if rpi.ReplicaN < 1 {
 		return ErrReplicationFactorTooLow
 	}
@@ -871,7 +878,7 @@ func (data *Data) importOneDB(other Data, backupDBName, restoreDBName, backupRPN
 		for j, sgImport := range rpImport.ShardGroups {
 			data.MaxShardGroupID++
 			rpImport.ShardGroups[j].ID = data.MaxShardGroupID
-			for k, _ := range sgImport.Shards {
+			for k := range sgImport.Shards {
 				data.MaxShardID++
 				shardIDMap[sgImport.Shards[k].ID] = data.MaxShardID
 				sgImport.Shards[k].ID = data.MaxShardID
@@ -1355,9 +1362,16 @@ func (sgi ShardGroupInfo) clone() ShardGroupInfo {
 	return other
 }
 
-// ShardFor returns the ShardInfo for a Point hash.
-func (sgi *ShardGroupInfo) ShardFor(hash uint64) ShardInfo {
-	return sgi.Shards[hash%uint64(len(sgi.Shards))]
+type hashIDer interface {
+	HashID() uint64
+}
+
+// ShardFor returns the ShardInfo for a Point or other hashIDer.
+func (sgi *ShardGroupInfo) ShardFor(p hashIDer) ShardInfo {
+	if len(sgi.Shards) == 1 {
+		return sgi.Shards[0]
+	}
+	return sgi.Shards[p.HashID()%uint64(len(sgi.Shards))]
 }
 
 // marshal serializes to a protobuf representation.
@@ -1559,7 +1573,7 @@ func (cqi *ContinuousQueryInfo) unmarshal(pb *internal.ContinuousQueryInfo) {
 	cqi.Query = pb.GetQuery()
 }
 
-var _ query.Authorizer = (*UserInfo)(nil)
+var _ query.FineAuthorizer = (*UserInfo)(nil)
 
 // UserInfo represents metadata about a user in the system.
 type UserInfo struct {
@@ -1577,7 +1591,7 @@ type UserInfo struct {
 }
 
 type User interface {
-	query.Authorizer
+	query.FineAuthorizer
 	ID() string
 	AuthorizeUnrestricted() bool
 }
@@ -1605,7 +1619,14 @@ func (u *UserInfo) AuthorizeSeriesWrite(database string, measurement []byte, tag
 	return true
 }
 
-// AuthorizeUnrestricted allows admins to shortcut access checks.
+// IsOpen is a method on FineAuthorizer to indicate all fine auth is permitted and short circuit some checks.
+func (u *UserInfo) IsOpen() bool {
+	return true
+}
+
+// AuthorizeUnrestricted identifies the admin user
+//
+// Only the pprof endpoint uses this, we should prefer to have explicit permissioning instead.
 func (u *UserInfo) AuthorizeUnrestricted() bool {
 	return u.Admin
 }
