@@ -6,7 +6,7 @@ import (
 	"time"
 
 	"github.com/influxdata/influxdb/models"
-	"github.com/influxdata/influxdb/stress/v2/stress_client"
+	stressClient "github.com/influxdata/influxdb/stress/v2/stress_client"
 )
 
 // QueryStatement is a Statement Implementation to run queries on the target InfluxDB instance
@@ -142,8 +142,12 @@ func getRandomFieldKey(m map[string]interface{}) string {
 }
 
 func setMapValues(m map[string]interface{}, p models.Point) {
-	m["%f"] = getRandomFieldKey(p.Fields())
-	m["%m"] = p.Name()
+	fields, err := p.Fields()
+	if err != nil {
+		panic(err)
+	}
+	m["%f"] = getRandomFieldKey(fields)
+	m["%m"] = string(p.Name())
 	m["%t"] = getRandomTagPair(p.Tags())
 	m["%a"] = p.UnixNano()
 }
