@@ -4,7 +4,7 @@ import (
 	"log"
 	"strconv"
 
-	"github.com/influxdata/influxdb/stress/v2/stress_client"
+	stressClient "github.com/influxdata/influxdb/stress/v2/stress_client"
 )
 
 // Statement is the common interface to shape the testing environment and prepare database requests
