@@ -21,6 +21,7 @@ import (
 	"path/filepath"
 	"runtime"
 	"sort"
+	"strings"
 	"sync"
 	"sync/atomic"
 	"time"
@@ -68,6 +69,18 @@ func (e errCompactionAborted) Error() string {
 	return "compaction aborted"
 }
 
+type errBlockRead struct {
+	file string
+	err  error
+}
+
+func (e errBlockRead) Error() string {
+	if e.err != nil {
+		return fmt.Sprintf("block read error on %s: %s", e.file, e.err)
+	}
+	return fmt.Sprintf("block read error on %s", e.file)
+}
+
 // CompactionGroup represents a list of files eligible to be compacted together.
 type CompactionGroup []string
 
@@ -83,6 +96,8 @@ type CompactionPlanner interface {
 	// ForceFull causes the planner to return a full compaction plan the next
 	// time Plan() is called if there are files that could be compacted.
 	ForceFull()
+
+	SetFileStore(fs *FileStore)
 }
 
 // DefaultPlanner implements CompactionPlanner using a strategy to roll up
@@ -122,6 +137,7 @@ type fileStore interface {
 	Stats() []FileStat
 	LastModified() time.Time
 	BlockCount(path string, idx int) int
+	ParseFileName(path string) (int, int, error)
 }
 
 func NewDefaultPlanner(fs fileStore, writeColdDuration time.Duration) *DefaultPlanner {
@@ -136,8 +152,16 @@ func NewDefaultPlanner(fs fileStore, writeColdDuration time.Duration) *DefaultPl
 // 000001-01.tsm, 000001-02.tsm would be in the same generation
 // 000001 each with different sequence numbers.
 type tsmGeneration struct {
-	id    int
-	files []FileStat
+	id            int
+	files         []FileStat
+	parseFileName ParseFileNameFunc
+}
+
+func newTsmGeneration(id int, parseFileNameFunc ParseFileNameFunc) *tsmGeneration {
+	return &tsmGeneration{
+		id:            id,
+		parseFileName: parseFileNameFunc,
+	}
 }
 
 // size returns the total size of the files in the generation.
@@ -155,7 +179,7 @@ func (t *tsmGeneration) level() int {
 	// 1 file with a sequence num of 1.  Level 2 is generated by compacting multiple
 	// level 1 files.  Level 3 is generate by compacting multiple level 2 files.  Level
 	// 4 is for anything else.
-	_, seq, _ := ParseTSMFileName(t.files[0].Path)
+	_, seq, _ := t.parseFileName(t.files[0].Path)
 	if seq < 4 {
 		return seq
 	}
@@ -178,6 +202,14 @@ func (t *tsmGeneration) hasTombstones() bool {
 	return false
 }
 
+func (c *DefaultPlanner) SetFileStore(fs *FileStore) {
+	c.FileStore = fs
+}
+
+func (c *DefaultPlanner) ParseFileName(path string) (int, int, error) {
+	return c.FileStore.ParseFileName(path)
+}
+
 // FullyCompacted returns true if the shard is fully compacted.
 func (c *DefaultPlanner) FullyCompacted() bool {
 	gens := c.findGenerations(false)
@@ -550,7 +582,7 @@ func (c *DefaultPlanner) Plan(lastWrite time.Time) []CompactionGroup {
 	// With the groups, we need to evaluate whether the group as a whole can be compacted
 	compactable := []tsmGenerations{}
 	for _, group := range groups {
-		//if we don't have enough generations to compact, skip it
+		// if we don't have enough generations to compact, skip it
 		if len(group) < 4 && !group.hasTombstones() {
 			continue
 		}
@@ -596,7 +628,7 @@ func (c *DefaultPlanner) findGenerations(skipInUse bool) tsmGenerations {
 	tsmStats := c.FileStore.Stats()
 	generations := make(map[int]*tsmGeneration, len(tsmStats))
 	for _, f := range tsmStats {
-		gen, _, _ := ParseTSMFileName(f.Path)
+		gen, _, _ := c.ParseFileName(f.Path)
 
 		// Skip any files that are assigned to a current compaction plan
 		if _, ok := c.filesInUse[f.Path]; skipInUse && ok {
@@ -605,9 +637,7 @@ func (c *DefaultPlanner) findGenerations(skipInUse bool) tsmGenerations {
 
 		group := generations[gen]
 		if group == nil {
-			group = &tsmGeneration{
-				id: gen,
-			}
+			group = newTsmGeneration(gen, c.ParseFileName)
 			generations[gen] = group
 		}
 		group.files = append(group.files, f)
@@ -675,6 +705,9 @@ type Compactor struct {
 	// RateLimit is the limit for disk writes for all concurrent compactions.
 	RateLimit limiter.Rate
 
+	formatFileName FormatFileNameFunc
+	parseFileName  ParseFileNameFunc
+
 	mu                 sync.RWMutex
 	snapshotsEnabled   bool
 	compactionsEnabled bool
@@ -692,6 +725,22 @@ type Compactor struct {
 	files map[string]struct{}
 }
 
+// NewCompactor returns a new instance of Compactor.
+func NewCompactor() *Compactor {
+	return &Compactor{
+		formatFileName: DefaultFormatFileName,
+		parseFileName:  DefaultParseFileName,
+	}
+}
+
+func (c *Compactor) WithFormatFileNameFunc(formatFileNameFunc FormatFileNameFunc) {
+	c.formatFileName = formatFileNameFunc
+}
+
+func (c *Compactor) WithParseFileNameFunc(parseFileNameFunc ParseFileNameFunc) {
+	c.parseFileName = parseFileNameFunc
+}
+
 // Open initializes the Compactor.
 func (c *Compactor) Open() {
 	c.mu.Lock()
@@ -808,7 +857,7 @@ func (c *Compactor) WriteSnapshot(cache *Cache) ([]string, error) {
 	for i := 0; i < concurrency; i++ {
 		go func(sp *Cache) {
 			iter := NewCacheKeyIterator(sp, tsdb.DefaultMaxPointsPerBlock, intC)
-			files, err := c.writeNewFiles(c.FileStore.NextGeneration(), 0, iter, throttle)
+			files, err := c.writeNewFiles(c.FileStore.NextGeneration(), 0, nil, iter, throttle)
 			resC <- res{files: files, err: err}
 
 		}(splits[i])
@@ -857,7 +906,7 @@ func (c *Compactor) compact(fast bool, tsmFiles []string) ([]string, error) {
 	// number to ensure we write to the next unique location.
 	var maxGeneration, maxSequence int
 	for _, f := range tsmFiles {
-		gen, seq, err := ParseTSMFileName(f)
+		gen, seq, err := c.parseFileName(f)
 		if err != nil {
 			return nil, err
 		}
@@ -888,6 +937,7 @@ func (c *Compactor) compact(fast bool, tsmFiles []string) ([]string, error) {
 			// doesn't exist.
 			return nil, errCompactionAborted{fmt.Errorf("bad plan: %s", file)}
 		}
+		defer tr.Unref() // inform that we're done with this reader when this method returns.
 		trs = append(trs, tr)
 	}
 
@@ -895,12 +945,12 @@ func (c *Compactor) compact(fast bool, tsmFiles []string) ([]string, error) {
 		return nil, nil
 	}
 
-	tsm, err := NewTSMKeyIterator(size, fast, intC, trs...)
+	tsm, err := NewTSMBatchKeyIterator(size, fast, intC, tsmFiles, trs...)
 	if err != nil {
 		return nil, err
 	}
 
-	return c.writeNewFiles(maxGeneration, maxSequence, tsm, true)
+	return c.writeNewFiles(maxGeneration, maxSequence, tsmFiles, tsm, true)
 }
 
 // CompactFull writes multiple smaller TSM files into 1 or more larger files.
@@ -981,14 +1031,15 @@ func (c *Compactor) removeTmpFiles(files []string) error {
 
 // writeNewFiles writes from the iterator into new TSM files, rotating
 // to a new file once it has reached the max TSM file size.
-func (c *Compactor) writeNewFiles(generation, sequence int, iter KeyIterator, throttle bool) ([]string, error) {
+func (c *Compactor) writeNewFiles(generation, sequence int, src []string, iter KeyIterator, throttle bool) ([]string, error) {
 	// These are the new TSM files written
 	var files []string
 
 	for {
 		sequence++
+
 		// New TSM files are written to a temp file and renamed when fully completed.
-		fileName := filepath.Join(c.Dir, fmt.Sprintf("%09d-%09d.%s.%s", generation, sequence, TSMFileExtension, TmpTSMFileExtension))
+		fileName := filepath.Join(c.Dir, c.formatFileName(generation, sequence)+"."+TSMFileExtension+"."+TmpTSMFileExtension)
 
 		// Write as much as possible to this file
 		err := c.write(fileName, iter, throttle)
@@ -1036,10 +1087,18 @@ func (c *Compactor) write(path string, iter KeyIterator, throttle bool) (err err
 		return errCompactionInProgress{err: err}
 	}
 
+	// syncingWriter ensures that whatever we wrap the above file descriptor in
+	// it will always be able to be synced by the tsm writer, since it does
+	// type assertions to attempt to sync.
+	type syncingWriter interface {
+		io.Writer
+		Sync() error
+	}
+
 	// Create the write for the new TSM file.
 	var (
 		w           TSMWriter
-		limitWriter io.Writer = fd
+		limitWriter syncingWriter = fd
 	)
 
 	if c.RateLimit != nil && throttle {
@@ -1095,6 +1154,10 @@ func (c *Compactor) write(path string, iter KeyIterator, throttle bool) (err err
 			return err
 		}
 
+		if minTime > maxTime {
+			return fmt.Errorf("invalid index entry for block. min=%d, max=%d", minTime, maxTime)
+		}
+
 		// Write the key and value
 		if err := w.WriteBlock(key, minTime, maxTime, block); err == ErrMaxBlocksExceeded {
 			if err := w.WriteIndex(); err != nil {
@@ -1173,6 +1236,15 @@ type KeyIterator interface {
 	// be required to store all the series and entries in the KeyIterator.
 	EstimatedIndexSize() int
 }
+type TSMErrors []error
+
+func (t TSMErrors) Error() string {
+	e := []string{}
+	for _, v := range t {
+		e = append(e, v.Error())
+	}
+	return strings.Join(e, ", ")
+}
 
 // tsmKeyIterator implements the KeyIterator for set of TSMReaders.  Iteration produces
 // keys in sorted order and the values between the keys sorted and deduped.  If any of
@@ -1184,12 +1256,12 @@ type tsmKeyIterator struct {
 	// values is the temporary buffers for each key that is returned by a reader
 	values map[string][]Value
 
-	// pos is the current key postion within the corresponding readers slice.  A value of
+	// pos is the current key position within the corresponding readers slice.  A value of
 	// pos[0] = 1, means the reader[0] is currently at key 1 in its ordered index.
 	pos []int
 
-	// err is any error we received while iterating values.
-	err error
+	// TSMError wraps any error we received while iterating values.
+	errs TSMErrors
 
 	// indicates whether the iterator should choose a faster merging strategy over a more
 	// optimally compressed one.  If fast is true, multiple blocks will just be added as is
@@ -1225,6 +1297,10 @@ type tsmKeyIterator struct {
 	interrupt chan struct{}
 }
 
+func (t *tsmKeyIterator) AppendError(err error) {
+	t.errs = append(t.errs, err)
+}
+
 type block struct {
 	key              []byte
 	minTime, maxTime int64
@@ -1347,7 +1423,7 @@ RETRY:
 			if iter.Next() {
 				key, minTime, maxTime, typ, _, b, err := iter.Read()
 				if err != nil {
-					k.err = err
+					k.AppendError(err)
 				}
 
 				// This block may have ranges of time removed from it that would
@@ -1380,7 +1456,7 @@ RETRY:
 					iter.Next()
 					key, minTime, maxTime, typ, _, b, err := iter.Read()
 					if err != nil {
-						k.err = err
+						k.AppendError(err)
 					}
 
 					tombstones := iter.r.TombstoneRange(key)
@@ -1410,7 +1486,7 @@ RETRY:
 			}
 
 			if iter.Err() != nil {
-				k.err = iter.Err()
+				k.AppendError(iter.Err())
 			}
 		}
 	}
@@ -1473,7 +1549,7 @@ func (k *tsmKeyIterator) merge() {
 	case BlockString:
 		k.mergeString()
 	default:
-		k.err = fmt.Errorf("unknown block type: %v", k.typ)
+		k.AppendError(fmt.Errorf("unknown block type: %v", k.typ))
 	}
 }
 
@@ -1486,11 +1562,11 @@ func (k *tsmKeyIterator) Read() ([]byte, int64, int64, []byte, error) {
 	}
 
 	if len(k.merged) == 0 {
-		return nil, 0, 0, nil, k.err
+		return nil, 0, 0, nil, k.Err()
 	}
 
 	block := k.merged[0]
-	return block.key, block.minTime, block.maxTime, block.b, k.err
+	return block.key, block.minTime, block.maxTime, block.b, k.Err()
 }
 
 func (k *tsmKeyIterator) Close() error {
@@ -1507,7 +1583,324 @@ func (k *tsmKeyIterator) Close() error {
 
 // Error returns any errors encountered during iteration.
 func (k *tsmKeyIterator) Err() error {
-	return k.err
+	if len(k.errs) == 0 {
+		return nil
+	}
+	return k.errs
+}
+
+// tsmBatchKeyIterator implements the KeyIterator for set of TSMReaders.  Iteration produces
+// keys in sorted order and the values between the keys sorted and deduped.  If any of
+// the readers have associated tombstone entries, they are returned as part of iteration.
+type tsmBatchKeyIterator struct {
+	// readers is the set of readers it produce a sorted key run with
+	readers []*TSMReader
+
+	// values is the temporary buffers for each key that is returned by a reader
+	values map[string][]Value
+
+	// pos is the current key position within the corresponding readers slice.  A value of
+	// pos[0] = 1, means the reader[0] is currently at key 1 in its ordered index.
+	pos []int
+
+	// errs is any error we received while iterating values.
+	errs TSMErrors
+
+	// indicates whether the iterator should choose a faster merging strategy over a more
+	// optimally compressed one.  If fast is true, multiple blocks will just be added as is
+	// and not combined.  In some cases, a slower path will need to be utilized even when
+	// fast is true to prevent overlapping blocks of time for the same key.
+	// If false, the blocks will be decoded and duplicated (if needed) and
+	// then chunked into the maximally sized blocks.
+	fast bool
+
+	// size is the maximum number of values to encode in a single block
+	size int
+
+	// key is the current key lowest key across all readers that has not be fully exhausted
+	// of values.
+	key []byte
+	typ byte
+
+	// tsmFiles are the string names of the files for use in tracking errors, ordered the same
+	// as iterators and buf
+	tsmFiles []string
+	// currentTsm is the current TSM file being iterated over
+	currentTsm string
+
+	iterators []*BlockIterator
+	blocks    blocks
+
+	buf []blocks
+
+	// mergeValues are decoded blocks that have been combined
+	mergedFloatValues    *tsdb.FloatArray
+	mergedIntegerValues  *tsdb.IntegerArray
+	mergedUnsignedValues *tsdb.UnsignedArray
+	mergedBooleanValues  *tsdb.BooleanArray
+	mergedStringValues   *tsdb.StringArray
+
+	// merged are encoded blocks that have been combined or used as is
+	// without decode
+	merged    blocks
+	interrupt chan struct{}
+}
+
+func (t *tsmBatchKeyIterator) AppendError(err error) {
+	t.errs = append(t.errs, err)
+}
+
+// NewTSMBatchKeyIterator returns a new TSM key iterator from readers.
+// size indicates the maximum number of values to encode in a single block.
+func NewTSMBatchKeyIterator(size int, fast bool, interrupt chan struct{}, tsmFiles []string, readers ...*TSMReader) (KeyIterator, error) {
+	var iter []*BlockIterator
+	for _, r := range readers {
+		iter = append(iter, r.BlockIterator())
+	}
+
+	return &tsmBatchKeyIterator{
+		readers:              readers,
+		values:               map[string][]Value{},
+		pos:                  make([]int, len(readers)),
+		size:                 size,
+		iterators:            iter,
+		fast:                 fast,
+		tsmFiles:             tsmFiles,
+		buf:                  make([]blocks, len(iter)),
+		mergedFloatValues:    &tsdb.FloatArray{},
+		mergedIntegerValues:  &tsdb.IntegerArray{},
+		mergedUnsignedValues: &tsdb.UnsignedArray{},
+		mergedBooleanValues:  &tsdb.BooleanArray{},
+		mergedStringValues:   &tsdb.StringArray{},
+		interrupt:            interrupt,
+	}, nil
+}
+
+func (k *tsmBatchKeyIterator) hasMergedValues() bool {
+	return k.mergedFloatValues.Len() > 0 ||
+		k.mergedIntegerValues.Len() > 0 ||
+		k.mergedUnsignedValues.Len() > 0 ||
+		k.mergedStringValues.Len() > 0 ||
+		k.mergedBooleanValues.Len() > 0
+}
+
+func (k *tsmBatchKeyIterator) EstimatedIndexSize() int {
+	var size uint32
+	for _, r := range k.readers {
+		size += r.IndexSize()
+	}
+	return int(size) / len(k.readers)
+}
+
+// Next returns true if there are any values remaining in the iterator.
+func (k *tsmBatchKeyIterator) Next() bool {
+RETRY:
+	// Any merged blocks pending?
+	if len(k.merged) > 0 {
+		k.merged = k.merged[1:]
+		if len(k.merged) > 0 {
+			return true
+		}
+	}
+
+	// Any merged values pending?
+	if k.hasMergedValues() {
+		k.merge()
+		if len(k.merged) > 0 || k.hasMergedValues() {
+			return true
+		}
+	}
+
+	// If we still have blocks from the last read, merge them
+	if len(k.blocks) > 0 {
+		k.merge()
+		if len(k.merged) > 0 || k.hasMergedValues() {
+			return true
+		}
+	}
+
+	// Read the next block from each TSM iterator
+	for i, v := range k.buf {
+		if len(v) != 0 {
+			continue
+		}
+
+		iter := k.iterators[i]
+		k.currentTsm = k.tsmFiles[i]
+		if iter.Next() {
+			key, minTime, maxTime, typ, _, b, err := iter.Read()
+			if err != nil {
+				k.AppendError(errBlockRead{k.currentTsm, err})
+			}
+
+			// This block may have ranges of time removed from it that would
+			// reduce the block min and max time.
+			tombstones := iter.r.TombstoneRange(key)
+
+			var blk *block
+			if cap(k.buf[i]) > len(k.buf[i]) {
+				k.buf[i] = k.buf[i][:len(k.buf[i])+1]
+				blk = k.buf[i][len(k.buf[i])-1]
+				if blk == nil {
+					blk = &block{}
+					k.buf[i][len(k.buf[i])-1] = blk
+				}
+			} else {
+				blk = &block{}
+				k.buf[i] = append(k.buf[i], blk)
+			}
+			blk.minTime = minTime
+			blk.maxTime = maxTime
+			blk.key = key
+			blk.typ = typ
+			blk.b = b
+			blk.tombstones = tombstones
+			blk.readMin = math.MaxInt64
+			blk.readMax = math.MinInt64
+
+			blockKey := key
+			for bytes.Equal(iter.PeekNext(), blockKey) {
+				iter.Next()
+				key, minTime, maxTime, typ, _, b, err := iter.Read()
+				if err != nil {
+					k.AppendError(errBlockRead{k.currentTsm, err})
+				}
+
+				tombstones := iter.r.TombstoneRange(key)
+
+				var blk *block
+				if cap(k.buf[i]) > len(k.buf[i]) {
+					k.buf[i] = k.buf[i][:len(k.buf[i])+1]
+					blk = k.buf[i][len(k.buf[i])-1]
+					if blk == nil {
+						blk = &block{}
+						k.buf[i][len(k.buf[i])-1] = blk
+					}
+				} else {
+					blk = &block{}
+					k.buf[i] = append(k.buf[i], blk)
+				}
+
+				blk.minTime = minTime
+				blk.maxTime = maxTime
+				blk.key = key
+				blk.typ = typ
+				blk.b = b
+				blk.tombstones = tombstones
+				blk.readMin = math.MaxInt64
+				blk.readMax = math.MinInt64
+			}
+		}
+
+		if iter.Err() != nil {
+			k.AppendError(errBlockRead{k.currentTsm, iter.Err()})
+		}
+	}
+
+	// Each reader could have a different key that it's currently at, need to find
+	// the next smallest one to keep the sort ordering.
+	var minKey []byte
+	var minType byte
+	for _, b := range k.buf {
+		// block could be nil if the iterator has been exhausted for that file
+		if len(b) == 0 {
+			continue
+		}
+		if len(minKey) == 0 || bytes.Compare(b[0].key, minKey) < 0 {
+			minKey = b[0].key
+			minType = b[0].typ
+		}
+	}
+	k.key = minKey
+	k.typ = minType
+
+	// Now we need to find all blocks that match the min key so we can combine and dedupe
+	// the blocks if necessary
+	for i, b := range k.buf {
+		if len(b) == 0 {
+			continue
+		}
+		if bytes.Equal(b[0].key, k.key) {
+			k.blocks = append(k.blocks, b...)
+			k.buf[i] = k.buf[i][:0]
+		}
+	}
+
+	if len(k.blocks) == 0 {
+		return false
+	}
+
+	k.merge()
+
+	// After merging all the values for this key, we might not have any.  (e.g. they were all deleted
+	// through many tombstones).  In this case, move on to the next key instead of ending iteration.
+	if len(k.merged) == 0 {
+		goto RETRY
+	}
+
+	return len(k.merged) > 0
+}
+
+// merge combines the next set of blocks into merged blocks.
+func (k *tsmBatchKeyIterator) merge() {
+	switch k.typ {
+	case BlockFloat64:
+		k.mergeFloat()
+	case BlockInteger:
+		k.mergeInteger()
+	case BlockUnsigned:
+		k.mergeUnsigned()
+	case BlockBoolean:
+		k.mergeBoolean()
+	case BlockString:
+		k.mergeString()
+	default:
+		k.AppendError(errBlockRead{k.currentTsm, fmt.Errorf("unknown block type: %v", k.typ)})
+	}
+}
+
+func (k *tsmBatchKeyIterator) handleEncodeError(err error, typ string) {
+	k.AppendError(errBlockRead{k.currentTsm, fmt.Errorf("encode error: unable to compress block type %s for key '%s': %v", typ, k.key, err)})
+}
+
+func (k *tsmBatchKeyIterator) handleDecodeError(err error, typ string) {
+	k.AppendError(errBlockRead{k.currentTsm, fmt.Errorf("decode error: unable to decompress block type %s for key '%s': %v", typ, k.key, err)})
+}
+
+func (k *tsmBatchKeyIterator) Read() ([]byte, int64, int64, []byte, error) {
+	// See if compactions were disabled while we were running.
+	select {
+	case <-k.interrupt:
+		return nil, 0, 0, nil, errCompactionAborted{}
+	default:
+	}
+
+	if len(k.merged) == 0 {
+		return nil, 0, 0, nil, k.Err()
+	}
+
+	block := k.merged[0]
+	return block.key, block.minTime, block.maxTime, block.b, k.Err()
+}
+
+func (k *tsmBatchKeyIterator) Close() error {
+	k.values = nil
+	k.pos = nil
+	k.iterators = nil
+	for _, r := range k.readers {
+		if err := r.Close(); err != nil {
+			return err
+		}
+	}
+	return nil
+}
+
+// Error returns any errors encountered during iteration.
+func (k *tsmBatchKeyIterator) Err() error {
+	if len(k.errs) == 0 {
+		return nil
+	}
+	return k.errs
 }
 
 type cacheKeyIterator struct {
