@@ -6,6 +6,7 @@ import (
 	"os"
 	"path/filepath"
 	"sort"
+	"strings"
 	"testing"
 	"time"
 
@@ -38,7 +39,7 @@ func TestCompactor_Snapshot(t *testing.T) {
 	compactor.Dir = dir
 	compactor.FileStore = &fakeFileStore{}
 
-	files, err := compactor.WriteSnapshot(c, tsm1.DefaultFormatFileName)
+	files, err := compactor.WriteSnapshot(c)
 	if err == nil {
 		t.Fatalf("expected error writing snapshot: %v", err)
 	}
@@ -49,7 +50,7 @@ func TestCompactor_Snapshot(t *testing.T) {
 
 	compactor.Open()
 
-	files, err = compactor.WriteSnapshot(c, tsm1.DefaultFormatFileName)
+	files, err = compactor.WriteSnapshot(c)
 	if err != nil {
 		t.Fatalf("unexpected error writing snapshot: %v", err)
 	}
@@ -88,6 +89,58 @@ func TestCompactor_Snapshot(t *testing.T) {
 	}
 }
 
+func TestCompactor_CompactFullLastTimestamp(t *testing.T) {
+	dir := MustTempDir()
+	defer os.RemoveAll(dir)
+
+	var vals tsm1.Values
+	ts := int64(1e9)
+	for i := 0; i < 120; i++ {
+		vals = append(vals, tsm1.NewIntegerValue(ts, 1))
+		ts += 1e9
+	}
+	// 121st timestamp skips a second
+	ts += 1e9
+	vals = append(vals, tsm1.NewIntegerValue(ts, 1))
+	writes := map[string][]tsm1.Value{
+		"cpu,host=A#!~#value": vals[:100],
+	}
+	f1 := MustWriteTSM(dir, 1, writes)
+
+	writes = map[string][]tsm1.Value{
+		"cpu,host=A#!~#value": vals[100:],
+	}
+	f2 := MustWriteTSM(dir, 2, writes)
+
+	fs := &fakeFileStore{}
+	defer fs.Close()
+	compactor := tsm1.NewCompactor()
+	compactor.Dir = dir
+	compactor.FileStore = fs
+	compactor.Open()
+
+	files, err := compactor.CompactFull([]string{f1, f2})
+	if err != nil {
+		t.Fatalf("unexpected error writing snapshot: %#v", err)
+	}
+
+	r := MustOpenTSMReader(files[0])
+	entries := r.Entries([]byte("cpu,host=A#!~#value"))
+	_, b, err := r.ReadBytes(&entries[0], nil)
+	if err != nil {
+		t.Fatalf("ReadBytes: unexpected error %v", err)
+	}
+	var a tsdb.IntegerArray
+	err = tsm1.DecodeIntegerArrayBlock(b, &a)
+	if err != nil {
+		t.Fatalf("DecodeIntegerArrayBlock: unexpected error %v", err)
+	}
+
+	if a.MaxTime() != entries[0].MaxTime {
+		t.Fatalf("expected MaxTime == a.MaxTime()")
+	}
+}
+
 // Ensures that a compaction will properly merge multiple TSM files
 func TestCompactor_CompactFull(t *testing.T) {
 	dir := MustTempDir()
@@ -142,13 +195,13 @@ func TestCompactor_CompactFull(t *testing.T) {
 		t.Fatalf("files length mismatch: got %v, exp %v", got, exp)
 	}
 
-	expGen, expSeq, err := tsm1.ParseFileName(f3)
+	expGen, expSeq, err := tsm1.DefaultParseFileName(f3)
 	if err != nil {
 		t.Fatalf("unexpected error parsing file name: %v", err)
 	}
 	expSeq = expSeq + 1
 
-	gotGen, gotSeq, err := tsm1.ParseFileName(files[0])
+	gotGen, gotSeq, err := tsm1.DefaultParseFileName(files[0])
 	if err != nil {
 		t.Fatalf("unexpected error parsing file name: %v", err)
 	}
@@ -192,6 +245,64 @@ func TestCompactor_CompactFull(t *testing.T) {
 	}
 }
 
+// Ensures that a compaction will properly merge multiple TSM files
+func TestCompactor_DecodeError(t *testing.T) {
+	dir := MustTempDir()
+	defer os.RemoveAll(dir)
+
+	// write 3 TSM files with different data and one new point
+	a1 := tsm1.NewValue(1, 1.1)
+	writes := map[string][]tsm1.Value{
+		"cpu,host=A#!~#value": {a1},
+	}
+	f1 := MustWriteTSM(dir, 1, writes)
+
+	a2 := tsm1.NewValue(2, 1.2)
+	b1 := tsm1.NewValue(1, 2.1)
+	writes = map[string][]tsm1.Value{
+		"cpu,host=A#!~#value": {a2},
+		"cpu,host=B#!~#value": {b1},
+	}
+	f2 := MustWriteTSM(dir, 2, writes)
+
+	a3 := tsm1.NewValue(1, 1.3)
+	c1 := tsm1.NewValue(1, 3.1)
+	writes = map[string][]tsm1.Value{
+		"cpu,host=A#!~#value": {a3},
+		"cpu,host=C#!~#value": {c1},
+	}
+	f3 := MustWriteTSM(dir, 3, writes)
+	f, err := os.OpenFile(f3, os.O_RDWR, os.ModePerm)
+	if err != nil {
+		panic(err)
+	}
+	f.WriteAt([]byte("ffff"), 10) // skip over header
+	f.Close()
+
+	fs := &fakeFileStore{}
+	defer fs.Close()
+	compactor := tsm1.NewCompactor()
+	compactor.Dir = dir
+	compactor.FileStore = fs
+
+	files, err := compactor.CompactFull([]string{f1, f2, f3})
+	if err == nil {
+		t.Fatalf("expected error writing snapshot: %v", err)
+	}
+	if len(files) > 0 {
+		t.Fatalf("no files should be compacted: got %v", len(files))
+
+	}
+
+	compactor.Open()
+
+	files, err = compactor.CompactFull([]string{f1, f2, f3})
+	if err == nil ||
+		!strings.Contains(err.Error(), "decode error: unable to decompress block type float for key 'cpu,host=A#!~#value': unpackBlock: not enough data for timestamp") {
+		t.Fatalf("expected error writing snapshot: %v", err)
+	}
+}
+
 // Ensures that a compaction will properly merge multiple TSM files
 func TestCompactor_Compact_OverlappingBlocks(t *testing.T) {
 	dir := MustTempDir()
@@ -527,13 +638,13 @@ func TestCompactor_CompactFull_SkipFullBlocks(t *testing.T) {
 		t.Fatalf("files length mismatch: got %v, exp %v", got, exp)
 	}
 
-	expGen, expSeq, err := tsm1.ParseFileName(f3)
+	expGen, expSeq, err := tsm1.DefaultParseFileName(f3)
 	if err != nil {
 		t.Fatalf("unexpected error parsing file name: %v", err)
 	}
 	expSeq = expSeq + 1
 
-	gotGen, gotSeq, err := tsm1.ParseFileName(files[0])
+	gotGen, gotSeq, err := tsm1.DefaultParseFileName(files[0])
 	if err != nil {
 		t.Fatalf("unexpected error parsing file name: %v", err)
 	}
@@ -629,13 +740,13 @@ func TestCompactor_CompactFull_TombstonedSkipBlock(t *testing.T) {
 		t.Fatalf("files length mismatch: got %v, exp %v", got, exp)
 	}
 
-	expGen, expSeq, err := tsm1.ParseFileName(f3)
+	expGen, expSeq, err := tsm1.DefaultParseFileName(f3)
 	if err != nil {
 		t.Fatalf("unexpected error parsing file name: %v", err)
 	}
 	expSeq = expSeq + 1
 
-	gotGen, gotSeq, err := tsm1.ParseFileName(files[0])
+	gotGen, gotSeq, err := tsm1.DefaultParseFileName(files[0])
 	if err != nil {
 		t.Fatalf("unexpected error parsing file name: %v", err)
 	}
@@ -732,13 +843,13 @@ func TestCompactor_CompactFull_TombstonedPartialBlock(t *testing.T) {
 		t.Fatalf("files length mismatch: got %v, exp %v", got, exp)
 	}
 
-	expGen, expSeq, err := tsm1.ParseFileName(f3)
+	expGen, expSeq, err := tsm1.DefaultParseFileName(f3)
 	if err != nil {
 		t.Fatalf("unexpected error parsing file name: %v", err)
 	}
 	expSeq = expSeq + 1
 
-	gotGen, gotSeq, err := tsm1.ParseFileName(files[0])
+	gotGen, gotSeq, err := tsm1.DefaultParseFileName(files[0])
 	if err != nil {
 		t.Fatalf("unexpected error parsing file name: %v", err)
 	}
@@ -840,13 +951,13 @@ func TestCompactor_CompactFull_TombstonedMultipleRanges(t *testing.T) {
 		t.Fatalf("files length mismatch: got %v, exp %v", got, exp)
 	}
 
-	expGen, expSeq, err := tsm1.ParseFileName(f3)
+	expGen, expSeq, err := tsm1.DefaultParseFileName(f3)
 	if err != nil {
 		t.Fatalf("unexpected error parsing file name: %v", err)
 	}
 	expSeq = expSeq + 1
 
-	gotGen, gotSeq, err := tsm1.ParseFileName(files[0])
+	gotGen, gotSeq, err := tsm1.DefaultParseFileName(files[0])
 	if err != nil {
 		t.Fatalf("unexpected error parsing file name: %v", err)
 	}
@@ -956,13 +1067,13 @@ func TestCompactor_CompactFull_MaxKeys(t *testing.T) {
 		t.Fatalf("files length mismatch: got %v, exp %v", got, exp)
 	}
 
-	expGen, expSeq, err := tsm1.ParseFileName(f2Name)
+	expGen, expSeq, err := tsm1.DefaultParseFileName(f2Name)
 	if err != nil {
 		t.Fatalf("unexpected error parsing file name: %v", err)
 	}
 	expSeq = expSeq + 1
 
-	gotGen, gotSeq, err := tsm1.ParseFileName(files[0])
+	gotGen, gotSeq, err := tsm1.DefaultParseFileName(files[0])
 	if err != nil {
 		t.Fatalf("unexpected error parsing file name: %v", err)
 	}
@@ -2885,3 +2996,7 @@ func (w *fakeFileStore) Close() {
 	}
 	w.readers = nil
 }
+
+func (w *fakeFileStore) ParseFileName(path string) (int, int, error) {
+	return tsm1.DefaultParseFileName(path)
+}
