@@ -6,31 +6,34 @@ import (
 	"runtime"
 	"time"
 
-	"github.com/influxdata/influxdb/influxql"
 	"github.com/influxdata/influxdb/pkg/pool"
 	"github.com/influxdata/influxdb/tsdb"
+	"github.com/influxdata/influxql"
 )
 
 const (
-	// BlockFloat64 designates a block encodes float64 values
+	// BlockFloat64 designates a block encodes float64 values.
 	BlockFloat64 = byte(0)
 
-	// BlockInteger designates a block encodes int64 values
+	// BlockInteger designates a block encodes int64 values.
 	BlockInteger = byte(1)
 
-	// BlockBoolean designates a block encodes boolean values
+	// BlockBoolean designates a block encodes boolean values.
 	BlockBoolean = byte(2)
 
-	// BlockString designates a block encodes string values
+	// BlockString designates a block encodes string values.
 	BlockString = byte(3)
 
+	// BlockUnsigned designates a block encodes uint64 values.
+	BlockUnsigned = byte(4)
+
 	// encodedBlockHeaderSize is the size of the header for an encoded block.  There is one
 	// byte encoding the type of the block.
 	encodedBlockHeaderSize = 1
 )
 
 func init() {
-	// Prime the pools with with at one encoder/decoder for each available CPU
+	// Prime the pools with one encoder/decoder for each available CPU.
 	vals := make([]interface{}, 0, runtime.NumCPU())
 	for _, p := range []*pool.Generic{
 		timeEncoderPool, timeDecoderPool,
@@ -54,6 +57,7 @@ func init() {
 
 var (
 	// encoder pools
+
 	timeEncoderPool = pool.NewGeneric(runtime.NumCPU(), func(sz int) interface{} {
 		return NewTimeEncoder(sz)
 	})
@@ -71,6 +75,7 @@ var (
 	})
 
 	// decoder pools
+
 	timeDecoderPool = pool.NewGeneric(runtime.NumCPU(), func(sz int) interface{} {
 		return &TimeDecoder{}
 	})
@@ -88,57 +93,88 @@ var (
 	})
 )
 
+// Value represents a TSM-encoded value.
 type Value interface {
+	// UnixNano returns the timestamp of the value in nanoseconds since unix epoch.
 	UnixNano() int64
+
+	// Value returns the underlying value.
 	Value() interface{}
+
+	// Size returns the number of bytes necessary to represent the value and its timestamp.
 	Size() int
+
+	// String returns the string representation of the value and its timestamp.
 	String() string
 
+	// internalOnly is unexported to ensure implementations of Value
+	// can only originate in this package.
 	internalOnly()
 }
 
+// NewValue returns a new Value with the underlying type dependent on value.
 func NewValue(t int64, value interface{}) Value {
 	switch v := value.(type) {
 	case int64:
-		return &IntegerValue{unixnano: t, value: v}
+		return IntegerValue{unixnano: t, value: v}
+	case uint64:
+		return UnsignedValue{unixnano: t, value: v}
 	case float64:
-		return &FloatValue{unixnano: t, value: v}
+		return FloatValue{unixnano: t, value: v}
 	case bool:
-		return &BooleanValue{unixnano: t, value: v}
+		return BooleanValue{unixnano: t, value: v}
 	case string:
-		return &StringValue{unixnano: t, value: v}
+		return StringValue{unixnano: t, value: v}
 	}
 	return EmptyValue{}
 }
 
+// NewIntegerValue returns a new integer value.
 func NewIntegerValue(t int64, v int64) Value {
-	return &IntegerValue{unixnano: t, value: v}
+	return IntegerValue{unixnano: t, value: v}
 }
 
+// NewUnsignedValue returns a new unsigned integer value.
+func NewUnsignedValue(t int64, v uint64) Value {
+	return UnsignedValue{unixnano: t, value: v}
+}
+
+// NewFloatValue returns a new float value.
 func NewFloatValue(t int64, v float64) Value {
-	return &FloatValue{unixnano: t, value: v}
+	return FloatValue{unixnano: t, value: v}
 }
 
+// NewBooleanValue returns a new boolean value.
 func NewBooleanValue(t int64, v bool) Value {
-	return &BooleanValue{unixnano: t, value: v}
+	return BooleanValue{unixnano: t, value: v}
 }
 
+// NewStringValue returns a new string value.
 func NewStringValue(t int64, v string) Value {
-	return &StringValue{unixnano: t, value: v}
+	return StringValue{unixnano: t, value: v}
 }
 
+// EmptyValue is used when there is no appropriate other value.
 type EmptyValue struct{}
 
-func (e EmptyValue) UnixNano() int64    { return tsdb.EOF }
+// UnixNano returns tsdb.EOF.
+func (e EmptyValue) UnixNano() int64 { return tsdb.EOF }
+
+// Value returns nil.
 func (e EmptyValue) Value() interface{} { return nil }
-func (e EmptyValue) Size() int          { return 0 }
-func (e EmptyValue) String() string     { return "" }
+
+// Size returns 0.
+func (e EmptyValue) Size() int { return 0 }
+
+// String returns the empty string.
+func (e EmptyValue) String() string { return "" }
 
 func (_ EmptyValue) internalOnly()    {}
-func (_ *StringValue) internalOnly()  {}
-func (_ *IntegerValue) internalOnly() {}
-func (_ *BooleanValue) internalOnly() {}
-func (_ *FloatValue) internalOnly()   {}
+func (_ StringValue) internalOnly()   {}
+func (_ IntegerValue) internalOnly()  {}
+func (_ UnsignedValue) internalOnly() {}
+func (_ BooleanValue) internalOnly()  {}
+func (_ FloatValue) internalOnly()    {}
 
 // Encode converts the values to a byte slice.  If there are no values,
 // this function panics.
@@ -148,13 +184,15 @@ func (a Values) Encode(buf []byte) ([]byte, error) {
 	}
 
 	switch a[0].(type) {
-	case *FloatValue:
+	case FloatValue:
 		return encodeFloatBlock(buf, a)
-	case *IntegerValue:
+	case IntegerValue:
 		return encodeIntegerBlock(buf, a)
-	case *BooleanValue:
+	case UnsignedValue:
+		return encodeUnsignedBlock(buf, a)
+	case BooleanValue:
 		return encodeBooleanBlock(buf, a)
-	case *StringValue:
+	case StringValue:
 		return encodeStringBlock(buf, a)
 	}
 
@@ -168,13 +206,15 @@ func (a Values) InfluxQLType() (influxql.DataType, error) {
 	}
 
 	switch a[0].(type) {
-	case *FloatValue:
+	case FloatValue:
 		return influxql.Float, nil
-	case *IntegerValue:
+	case IntegerValue:
 		return influxql.Integer, nil
-	case *BooleanValue:
+	case UnsignedValue:
+		return influxql.Unsigned, nil
+	case BooleanValue:
 		return influxql.Boolean, nil
-	case *StringValue:
+	case StringValue:
 		return influxql.String, nil
 	}
 
@@ -186,35 +226,36 @@ func (a Values) InfluxQLType() (influxql.DataType, error) {
 func BlockType(block []byte) (byte, error) {
 	blockType := block[0]
 	switch blockType {
-	case BlockFloat64, BlockInteger, BlockBoolean, BlockString:
+	case BlockFloat64, BlockInteger, BlockUnsigned, BlockBoolean, BlockString:
 		return blockType, nil
 	default:
 		return 0, fmt.Errorf("unknown block type: %d", blockType)
 	}
 }
 
-func BlockCount(block []byte) int {
+// BlockCount returns the number of timestamps encoded in block.
+func BlockCount(block []byte) (int, error) {
 	if len(block) <= encodedBlockHeaderSize {
-		panic(fmt.Sprintf("count of short block: got %v, exp %v", len(block), encodedBlockHeaderSize))
+		return 0, fmt.Errorf("count of short block: got %v, exp %v", len(block), encodedBlockHeaderSize)
 	}
 	// first byte is the block type
 	tb, _, err := unpackBlock(block[1:])
 	if err != nil {
-		panic(fmt.Sprintf("BlockCount: error unpacking block: %s", err.Error()))
+		return 0, fmt.Errorf("BlockCount: error unpacking block: %v", err)
 	}
-	return CountTimestamps(tb)
+	return CountTimestamps(tb), nil
 }
 
-// DecodeBlock takes a byte array and will decode into values of the appropriate type
+// DecodeBlock takes a byte slice and decodes it into values of the appropriate type
 // based on the block.
 func DecodeBlock(block []byte, vals []Value) ([]Value, error) {
 	if len(block) <= encodedBlockHeaderSize {
-		panic(fmt.Sprintf("decode of short block: got %v, exp %v", len(block), encodedBlockHeaderSize))
+		return nil, fmt.Errorf("decode of short block: got %v, exp %v", len(block), encodedBlockHeaderSize)
 	}
 
 	blockType, err := BlockType(block)
 	if err != nil {
-		return nil, err
+		return nil, fmt.Errorf("error decoding block type: %v", err)
 	}
 
 	switch blockType {
@@ -225,7 +266,7 @@ func DecodeBlock(block []byte, vals []Value) ([]Value, error) {
 			vals = make([]Value, len(decoded))
 		}
 		for i := range decoded {
-			vals[i] = &decoded[i]
+			vals[i] = decoded[i]
 		}
 		return vals[:len(decoded)], err
 	case BlockInteger:
@@ -235,7 +276,18 @@ func DecodeBlock(block []byte, vals []Value) ([]Value, error) {
 			vals = make([]Value, len(decoded))
 		}
 		for i := range decoded {
-			vals[i] = &decoded[i]
+			vals[i] = decoded[i]
+		}
+		return vals[:len(decoded)], err
+
+	case BlockUnsigned:
+		var buf []UnsignedValue
+		decoded, err := DecodeUnsignedBlock(block, &buf)
+		if len(vals) < len(decoded) {
+			vals = make([]Value, len(decoded))
+		}
+		for i := range decoded {
+			vals[i] = decoded[i]
 		}
 		return vals[:len(decoded)], err
 
@@ -246,7 +298,7 @@ func DecodeBlock(block []byte, vals []Value) ([]Value, error) {
 			vals = make([]Value, len(decoded))
 		}
 		for i := range decoded {
-			vals[i] = &decoded[i]
+			vals[i] = decoded[i]
 		}
 		return vals[:len(decoded)], err
 
@@ -257,34 +309,39 @@ func DecodeBlock(block []byte, vals []Value) ([]Value, error) {
 			vals = make([]Value, len(decoded))
 		}
 		for i := range decoded {
-			vals[i] = &decoded[i]
+			vals[i] = decoded[i]
 		}
 		return vals[:len(decoded)], err
 
 	default:
-		panic(fmt.Sprintf("unknown block type: %d", blockType))
+		return nil, fmt.Errorf("unknown block type: %d", blockType)
 	}
 }
 
+// FloatValue represents a float64 value.
 type FloatValue struct {
 	unixnano int64
 	value    float64
 }
 
-func (f *FloatValue) UnixNano() int64 {
-	return f.unixnano
+// UnixNano returns the timestamp of the value.
+func (v FloatValue) UnixNano() int64 {
+	return v.unixnano
 }
 
-func (f *FloatValue) Value() interface{} {
-	return f.value
+// Value returns the underlying float64 value.
+func (v FloatValue) Value() interface{} {
+	return v.value
 }
 
-func (f *FloatValue) Size() int {
+// Size returns the number of bytes necessary to represent the value and its timestamp.
+func (v FloatValue) Size() int {
 	return 16
 }
 
-func (f *FloatValue) String() string {
-	return fmt.Sprintf("%v %v", time.Unix(0, f.unixnano), f.value)
+// String returns the string representation of the value and its timestamp.
+func (v FloatValue) String() string {
+	return fmt.Sprintf("%v %v", time.Unix(0, v.unixnano), v.value)
 }
 
 func encodeFloatBlock(buf []byte, values []Value) ([]byte, error) {
@@ -296,37 +353,13 @@ func encodeFloatBlock(buf []byte, values []Value) ([]byte, error) {
 	// for timestamps and values.
 
 	// Encode values using Gorilla float compression
-	venc := getFloatEncoder()
+	venc := getFloatEncoder(len(values))
 
 	// Encode timestamps using an adaptive encoder that uses delta-encoding,
 	// frame-or-reference and run length encoding.
 	tsenc := getTimeEncoder(len(values))
 
-	var b []byte
-	err := func() error {
-		for _, v := range values {
-			tsenc.Write(v.UnixNano())
-			venc.Push(v.(*FloatValue).value)
-		}
-		venc.Finish()
-
-		// Encoded timestamp values
-		tb, err := tsenc.Bytes()
-		if err != nil {
-			return err
-		}
-		// Encoded float values
-		vb, err := venc.Bytes()
-		if err != nil {
-			return err
-		}
-
-		// Prepend the first timestamp of the block in the first 8 bytes and the block
-		// in the next byte, followed by the block
-		b = packBlock(buf, BlockFloat64, tb, vb)
-
-		return nil
-	}()
+	b, err := encodeFloatBlockUsing(buf, values, tsenc, venc)
 
 	putTimeEncoder(tsenc)
 	putFloatEncoder(venc)
@@ -334,6 +367,35 @@ func encodeFloatBlock(buf []byte, values []Value) ([]byte, error) {
 	return b, err
 }
 
+func encodeFloatBlockUsing(buf []byte, values []Value, tsenc TimeEncoder, venc *FloatEncoder) ([]byte, error) {
+	tsenc.Reset()
+	venc.Reset()
+
+	for _, v := range values {
+		vv := v.(FloatValue)
+		tsenc.Write(vv.unixnano)
+		venc.Write(vv.value)
+	}
+	venc.Flush()
+
+	// Encoded timestamp values
+	tb, err := tsenc.Bytes()
+	if err != nil {
+		return nil, err
+	}
+	// Encoded float values
+	vb, err := venc.Bytes()
+	if err != nil {
+		return nil, err
+	}
+
+	// Prepend the first timestamp of the block in the first 8 bytes and the block
+	// in the next byte, followed by the block
+	return packBlock(buf, BlockFloat64, tb, vb), nil
+}
+
+// DecodeFloatBlock decodes the float block from the byte slice
+// and appends the float values to a.
 func DecodeFloatBlock(block []byte, a *[]FloatValue) ([]FloatValue, error) {
 	// Block type is the next block, make sure we actually have a float block
 	blockType := block[0]
@@ -347,11 +409,19 @@ func DecodeFloatBlock(block []byte, a *[]FloatValue) ([]FloatValue, error) {
 		return nil, err
 	}
 
+	sz := CountTimestamps(tb)
+
+	if cap(*a) < sz {
+		*a = make([]FloatValue, sz)
+	} else {
+		*a = (*a)[:sz]
+	}
+
 	tdec := timeDecoderPool.Get(0).(*TimeDecoder)
 	vdec := floatDecoderPool.Get(0).(*FloatDecoder)
 
 	var i int
-	err = func() error {
+	err = func(a []FloatValue) error {
 		// Setup our timestamp and value decoders
 		tdec.Init(tb)
 		err = vdec.SetBytes(vb)
@@ -360,18 +430,12 @@ func DecodeFloatBlock(block []byte, a *[]FloatValue) ([]FloatValue, error) {
 		}
 
 		// Decode both a timestamp and value
-		for tdec.Next() && vdec.Next() {
-			ts := tdec.Read()
-			v := vdec.Values()
-			if i < len(*a) {
-				elem := &(*a)[i]
-				elem.unixnano = ts
-				elem.value = v
-			} else {
-				*a = append(*a, FloatValue{ts, v})
-			}
-			i++
+		j := 0
+		for j < len(a) && tdec.Next() && vdec.Next() {
+			a[j] = FloatValue{unixnano: tdec.Read(), value: vdec.Values()}
+			j++
 		}
+		i = j
 
 		// Did timestamp decoding have an error?
 		err = tdec.Error()
@@ -380,12 +444,8 @@ func DecodeFloatBlock(block []byte, a *[]FloatValue) ([]FloatValue, error) {
 		}
 
 		// Did float decoding have an error?
-		err = vdec.Error()
-		if err != nil {
-			return err
-		}
-		return nil
-	}()
+		return vdec.Error()
+	}(*a)
 
 	timeDecoderPool.Put(tdec)
 	floatDecoderPool.Put(vdec)
@@ -393,25 +453,30 @@ func DecodeFloatBlock(block []byte, a *[]FloatValue) ([]FloatValue, error) {
 	return (*a)[:i], err
 }
 
+// BooleanValue represents a boolean value.
 type BooleanValue struct {
 	unixnano int64
 	value    bool
 }
 
-func (b *BooleanValue) Size() int {
+// Size returns the number of bytes necessary to represent the value and its timestamp.
+func (v BooleanValue) Size() int {
 	return 9
 }
 
-func (b *BooleanValue) UnixNano() int64 {
-	return b.unixnano
+// UnixNano returns the timestamp of the value in nanoseconds since unix epoch.
+func (v BooleanValue) UnixNano() int64 {
+	return v.unixnano
 }
 
-func (b *BooleanValue) Value() interface{} {
-	return b.value
+// Value returns the underlying boolean value.
+func (v BooleanValue) Value() interface{} {
+	return v.value
 }
 
-func (f *BooleanValue) String() string {
-	return fmt.Sprintf("%v %v", time.Unix(0, f.unixnano), f.Value())
+// String returns the string representation of the value and its timestamp.
+func (v BooleanValue) String() string {
+	return fmt.Sprintf("%v %v", time.Unix(0, v.unixnano), v.Value())
 }
 
 func encodeBooleanBlock(buf []byte, values []Value) ([]byte, error) {
@@ -426,29 +491,7 @@ func encodeBooleanBlock(buf []byte, values []Value) ([]byte, error) {
 	// Encode timestamps using an adaptive encoder
 	tsenc := getTimeEncoder(len(values))
 
-	var b []byte
-	err := func() error {
-		for _, v := range values {
-			tsenc.Write(v.UnixNano())
-			venc.Write(v.(*BooleanValue).value)
-		}
-
-		// Encoded timestamp values
-		tb, err := tsenc.Bytes()
-		if err != nil {
-			return err
-		}
-		// Encoded float values
-		vb, err := venc.Bytes()
-		if err != nil {
-			return err
-		}
-
-		// Prepend the first timestamp of the block in the first 8 bytes and the block
-		// in the next byte, followed by the block
-		b = packBlock(buf, BlockBoolean, tb, vb)
-		return nil
-	}()
+	b, err := encodeBooleanBlockUsing(buf, values, tsenc, venc)
 
 	putTimeEncoder(tsenc)
 	putBooleanEncoder(venc)
@@ -456,6 +499,34 @@ func encodeBooleanBlock(buf []byte, values []Value) ([]byte, error) {
 	return b, err
 }
 
+func encodeBooleanBlockUsing(buf []byte, values []Value, tenc TimeEncoder, venc BooleanEncoder) ([]byte, error) {
+	tenc.Reset()
+	venc.Reset()
+
+	for _, v := range values {
+		vv := v.(BooleanValue)
+		tenc.Write(vv.unixnano)
+		venc.Write(vv.value)
+	}
+
+	// Encoded timestamp values
+	tb, err := tenc.Bytes()
+	if err != nil {
+		return nil, err
+	}
+	// Encoded float values
+	vb, err := venc.Bytes()
+	if err != nil {
+		return nil, err
+	}
+
+	// Prepend the first timestamp of the block in the first 8 bytes and the block
+	// in the next byte, followed by the block
+	return packBlock(buf, BlockBoolean, tb, vb), nil
+}
+
+// DecodeBooleanBlock decodes the boolean block from the byte slice
+// and appends the boolean values to a.
 func DecodeBooleanBlock(block []byte, a *[]BooleanValue) ([]BooleanValue, error) {
 	// Block type is the next block, make sure we actually have a float block
 	blockType := block[0]
@@ -469,28 +540,30 @@ func DecodeBooleanBlock(block []byte, a *[]BooleanValue) ([]BooleanValue, error)
 		return nil, err
 	}
 
+	sz := CountTimestamps(tb)
+
+	if cap(*a) < sz {
+		*a = make([]BooleanValue, sz)
+	} else {
+		*a = (*a)[:sz]
+	}
+
 	tdec := timeDecoderPool.Get(0).(*TimeDecoder)
 	vdec := booleanDecoderPool.Get(0).(*BooleanDecoder)
 
 	var i int
-	err = func() error {
+	err = func(a []BooleanValue) error {
 		// Setup our timestamp and value decoders
 		tdec.Init(tb)
 		vdec.SetBytes(vb)
 
 		// Decode both a timestamp and value
-		for tdec.Next() && vdec.Next() {
-			ts := tdec.Read()
-			v := vdec.Read()
-			if i < len(*a) {
-				elem := &(*a)[i]
-				elem.unixnano = ts
-				elem.value = v
-			} else {
-				*a = append(*a, BooleanValue{ts, v})
-			}
-			i++
+		j := 0
+		for j < len(a) && tdec.Next() && vdec.Next() {
+			a[j] = BooleanValue{unixnano: tdec.Read(), value: vdec.Read()}
+			j++
 		}
+		i = j
 
 		// Did timestamp decoding have an error?
 		err = tdec.Error()
@@ -498,12 +571,8 @@ func DecodeBooleanBlock(block []byte, a *[]BooleanValue) ([]BooleanValue, error)
 			return err
 		}
 		// Did boolean decoding have an error?
-		err = vdec.Error()
-		if err != nil {
-			return err
-		}
-		return nil
-	}()
+		return vdec.Error()
+	}(*a)
 
 	timeDecoderPool.Put(tdec)
 	booleanDecoderPool.Put(vdec)
@@ -511,60 +580,71 @@ func DecodeBooleanBlock(block []byte, a *[]BooleanValue) ([]BooleanValue, error)
 	return (*a)[:i], err
 }
 
+// IntegerValue represents an int64 value.
 type IntegerValue struct {
 	unixnano int64
 	value    int64
 }
 
-func (v *IntegerValue) Value() interface{} {
+// Value returns the underlying int64 value.
+func (v IntegerValue) Value() interface{} {
 	return v.value
 }
 
-func (v *IntegerValue) UnixNano() int64 {
+// UnixNano returns the timestamp of the value.
+func (v IntegerValue) UnixNano() int64 {
 	return v.unixnano
 }
 
-func (v *IntegerValue) Size() int {
+// Size returns the number of bytes necessary to represent the value and its timestamp.
+func (v IntegerValue) Size() int {
 	return 16
 }
 
-func (f *IntegerValue) String() string {
-	return fmt.Sprintf("%v %v", time.Unix(0, f.unixnano), f.Value())
+// String returns the string representation of the value and its timestamp.
+func (v IntegerValue) String() string {
+	return fmt.Sprintf("%v %v", time.Unix(0, v.unixnano), v.Value())
 }
 
 func encodeIntegerBlock(buf []byte, values []Value) ([]byte, error) {
-	tsEnc := getTimeEncoder(len(values))
-	vEnc := getIntegerEncoder(len(values))
+	tenc := getTimeEncoder(len(values))
+	venc := getIntegerEncoder(len(values))
 
-	var b []byte
-	err := func() error {
-		for _, v := range values {
-			tsEnc.Write(v.UnixNano())
-			vEnc.Write(v.(*IntegerValue).value)
-		}
+	b, err := encodeIntegerBlockUsing(buf, values, tenc, venc)
 
-		// Encoded timestamp values
-		tb, err := tsEnc.Bytes()
-		if err != nil {
-			return err
-		}
-		// Encoded int64 values
-		vb, err := vEnc.Bytes()
-		if err != nil {
-			return err
-		}
-
-		// Prepend the first timestamp of the block in the first 8 bytes
-		b = packBlock(buf, BlockInteger, tb, vb)
-		return nil
-	}()
-
-	putTimeEncoder(tsEnc)
-	putIntegerEncoder(vEnc)
+	putTimeEncoder(tenc)
+	putIntegerEncoder(venc)
 
 	return b, err
 }
 
+func encodeIntegerBlockUsing(buf []byte, values []Value, tenc TimeEncoder, venc IntegerEncoder) ([]byte, error) {
+	tenc.Reset()
+	venc.Reset()
+
+	for _, v := range values {
+		vv := v.(IntegerValue)
+		tenc.Write(vv.unixnano)
+		venc.Write(vv.value)
+	}
+
+	// Encoded timestamp values
+	tb, err := tenc.Bytes()
+	if err != nil {
+		return nil, err
+	}
+	// Encoded int64 values
+	vb, err := venc.Bytes()
+	if err != nil {
+		return nil, err
+	}
+
+	// Prepend the first timestamp of the block in the first 8 bytes
+	return packBlock(buf, BlockInteger, tb, vb), nil
+}
+
+// DecodeIntegerBlock decodes the integer block from the byte slice
+// and appends the integer values to a.
 func DecodeIntegerBlock(block []byte, a *[]IntegerValue) ([]IntegerValue, error) {
 	blockType := block[0]
 	if blockType != BlockInteger {
@@ -579,28 +659,30 @@ func DecodeIntegerBlock(block []byte, a *[]IntegerValue) ([]IntegerValue, error)
 		return nil, err
 	}
 
+	sz := CountTimestamps(tb)
+
+	if cap(*a) < sz {
+		*a = make([]IntegerValue, sz)
+	} else {
+		*a = (*a)[:sz]
+	}
+
 	tdec := timeDecoderPool.Get(0).(*TimeDecoder)
 	vdec := integerDecoderPool.Get(0).(*IntegerDecoder)
 
 	var i int
-	err = func() error {
+	err = func(a []IntegerValue) error {
 		// Setup our timestamp and value decoders
 		tdec.Init(tb)
 		vdec.SetBytes(vb)
 
 		// Decode both a timestamp and value
-		for tdec.Next() && vdec.Next() {
-			ts := tdec.Read()
-			v := vdec.Read()
-			if i < len(*a) {
-				elem := &(*a)[i]
-				elem.unixnano = ts
-				elem.value = v
-			} else {
-				*a = append(*a, IntegerValue{ts, v})
-			}
-			i++
+		j := 0
+		for j < len(a) && tdec.Next() && vdec.Next() {
+			a[j] = IntegerValue{unixnano: tdec.Read(), value: vdec.Read()}
+			j++
 		}
+		i = j
 
 		// Did timestamp decoding have an error?
 		err = tdec.Error()
@@ -608,12 +690,8 @@ func DecodeIntegerBlock(block []byte, a *[]IntegerValue) ([]IntegerValue, error)
 			return err
 		}
 		// Did int64 decoding have an error?
-		err = vdec.Error()
-		if err != nil {
-			return err
-		}
-		return nil
-	}()
+		return vdec.Error()
+	}(*a)
 
 	timeDecoderPool.Put(tdec)
 	integerDecoderPool.Put(vdec)
@@ -621,61 +699,190 @@ func DecodeIntegerBlock(block []byte, a *[]IntegerValue) ([]IntegerValue, error)
 	return (*a)[:i], err
 }
 
+// UnsignedValue represents an int64 value.
+type UnsignedValue struct {
+	unixnano int64
+	value    uint64
+}
+
+// Value returns the underlying int64 value.
+func (v UnsignedValue) Value() interface{} {
+	return v.value
+}
+
+// UnixNano returns the timestamp of the value.
+func (v UnsignedValue) UnixNano() int64 {
+	return v.unixnano
+}
+
+// Size returns the number of bytes necessary to represent the value and its timestamp.
+func (v UnsignedValue) Size() int {
+	return 16
+}
+
+// String returns the string representation of the value and its timestamp.
+func (v UnsignedValue) String() string {
+	return fmt.Sprintf("%v %v", time.Unix(0, v.unixnano), v.Value())
+}
+
+func encodeUnsignedBlock(buf []byte, values []Value) ([]byte, error) {
+	tenc := getTimeEncoder(len(values))
+	venc := getUnsignedEncoder(len(values))
+
+	b, err := encodeUnsignedBlockUsing(buf, values, tenc, venc)
+
+	putTimeEncoder(tenc)
+	putUnsignedEncoder(venc)
+
+	return b, err
+}
+
+func encodeUnsignedBlockUsing(buf []byte, values []Value, tenc TimeEncoder, venc IntegerEncoder) ([]byte, error) {
+	tenc.Reset()
+	venc.Reset()
+
+	for _, v := range values {
+		vv := v.(UnsignedValue)
+		tenc.Write(vv.unixnano)
+		venc.Write(int64(vv.value))
+	}
+
+	// Encoded timestamp values
+	tb, err := tenc.Bytes()
+	if err != nil {
+		return nil, err
+	}
+	// Encoded int64 values
+	vb, err := venc.Bytes()
+	if err != nil {
+		return nil, err
+	}
+
+	// Prepend the first timestamp of the block in the first 8 bytes
+	return packBlock(buf, BlockUnsigned, tb, vb), nil
+}
+
+// DecodeUnsignedBlock decodes the unsigned integer block from the byte slice
+// and appends the unsigned integer values to a.
+func DecodeUnsignedBlock(block []byte, a *[]UnsignedValue) ([]UnsignedValue, error) {
+	blockType := block[0]
+	if blockType != BlockUnsigned {
+		return nil, fmt.Errorf("invalid block type: exp %d, got %d", BlockUnsigned, blockType)
+	}
+
+	block = block[1:]
+
+	// The first 8 bytes is the minimum timestamp of the block
+	tb, vb, err := unpackBlock(block)
+	if err != nil {
+		return nil, err
+	}
+
+	sz := CountTimestamps(tb)
+
+	if cap(*a) < sz {
+		*a = make([]UnsignedValue, sz)
+	} else {
+		*a = (*a)[:sz]
+	}
+
+	tdec := timeDecoderPool.Get(0).(*TimeDecoder)
+	vdec := integerDecoderPool.Get(0).(*IntegerDecoder)
+
+	var i int
+	err = func(a []UnsignedValue) error {
+		// Setup our timestamp and value decoders
+		tdec.Init(tb)
+		vdec.SetBytes(vb)
+
+		// Decode both a timestamp and value
+		j := 0
+		for j < len(a) && tdec.Next() && vdec.Next() {
+			a[j] = UnsignedValue{unixnano: tdec.Read(), value: uint64(vdec.Read())}
+			j++
+		}
+		i = j
+
+		// Did timestamp decoding have an error?
+		err = tdec.Error()
+		if err != nil {
+			return err
+		}
+		// Did int64 decoding have an error?
+		return vdec.Error()
+	}(*a)
+
+	timeDecoderPool.Put(tdec)
+	integerDecoderPool.Put(vdec)
+
+	return (*a)[:i], err
+}
+
+// StringValue represents a string value.
 type StringValue struct {
 	unixnano int64
 	value    string
 }
 
-func (v *StringValue) Value() interface{} {
+// Value returns the underlying string value.
+func (v StringValue) Value() interface{} {
 	return v.value
 }
 
-func (v *StringValue) UnixNano() int64 {
+// UnixNano returns the timestamp of the value.
+func (v StringValue) UnixNano() int64 {
 	return v.unixnano
 }
 
-func (v *StringValue) Size() int {
+// Size returns the number of bytes necessary to represent the value and its timestamp.
+func (v StringValue) Size() int {
 	return 8 + len(v.value)
 }
 
-func (f *StringValue) String() string {
-	return fmt.Sprintf("%v %v", time.Unix(0, f.unixnano), f.Value())
+// String returns the string representation of the value and its timestamp.
+func (v StringValue) String() string {
+	return fmt.Sprintf("%v %v", time.Unix(0, v.unixnano), v.Value())
 }
 
 func encodeStringBlock(buf []byte, values []Value) ([]byte, error) {
-	tsEnc := getTimeEncoder(len(values))
-	vEnc := getStringEncoder(len(values) * len(values[0].(*StringValue).value))
+	tenc := getTimeEncoder(len(values))
+	venc := getStringEncoder(len(values) * len(values[0].(StringValue).value))
 
-	var b []byte
-	err := func() error {
-		for _, v := range values {
-			tsEnc.Write(v.UnixNano())
-			vEnc.Write(v.(*StringValue).value)
-		}
+	b, err := encodeStringBlockUsing(buf, values, tenc, venc)
 
-		// Encoded timestamp values
-		tb, err := tsEnc.Bytes()
-		if err != nil {
-			return err
-		}
-		// Encoded string values
-		vb, err := vEnc.Bytes()
-		if err != nil {
-			return err
-		}
-
-		// Prepend the first timestamp of the block in the first 8 bytes
-		b = packBlock(buf, BlockString, tb, vb)
-
-		return nil
-	}()
-
-	putTimeEncoder(tsEnc)
-	putStringEncoder(vEnc)
+	putTimeEncoder(tenc)
+	putStringEncoder(venc)
 
 	return b, err
 }
 
+func encodeStringBlockUsing(buf []byte, values []Value, tenc TimeEncoder, venc StringEncoder) ([]byte, error) {
+	tenc.Reset()
+	venc.Reset()
+
+	for _, v := range values {
+		vv := v.(StringValue)
+		tenc.Write(vv.unixnano)
+		venc.Write(vv.value)
+	}
+
+	// Encoded timestamp values
+	tb, err := tenc.Bytes()
+	if err != nil {
+		return nil, err
+	}
+	// Encoded string values
+	vb, err := venc.Bytes()
+	if err != nil {
+		return nil, err
+	}
+
+	// Prepend the first timestamp of the block in the first 8 bytes
+	return packBlock(buf, BlockString, tb, vb), nil
+}
+
+// DecodeStringBlock decodes the string block from the byte slice
+// and appends the string values to a.
 func DecodeStringBlock(block []byte, a *[]StringValue) ([]StringValue, error) {
 	blockType := block[0]
 	if blockType != BlockString {
@@ -690,11 +897,19 @@ func DecodeStringBlock(block []byte, a *[]StringValue) ([]StringValue, error) {
 		return nil, err
 	}
 
+	sz := CountTimestamps(tb)
+
+	if cap(*a) < sz {
+		*a = make([]StringValue, sz)
+	} else {
+		*a = (*a)[:sz]
+	}
+
 	tdec := timeDecoderPool.Get(0).(*TimeDecoder)
 	vdec := stringDecoderPool.Get(0).(*StringDecoder)
 
 	var i int
-	err = func() error {
+	err = func(a []StringValue) error {
 		// Setup our timestamp and value decoders
 		tdec.Init(tb)
 		err = vdec.SetBytes(vb)
@@ -703,18 +918,12 @@ func DecodeStringBlock(block []byte, a *[]StringValue) ([]StringValue, error) {
 		}
 
 		// Decode both a timestamp and value
-		for tdec.Next() && vdec.Next() {
-			ts := tdec.Read()
-			v := vdec.Read()
-			if i < len(*a) {
-				elem := &(*a)[i]
-				elem.unixnano = ts
-				elem.value = v
-			} else {
-				*a = append(*a, StringValue{ts, v})
-			}
-			i++
+		j := 0
+		for j < len(a) && tdec.Next() && vdec.Next() {
+			a[j] = StringValue{unixnano: tdec.Read(), value: vdec.Read()}
+			j++
 		}
+		i = j
 
 		// Did timestamp decoding have an error?
 		err = tdec.Error()
@@ -722,12 +931,8 @@ func DecodeStringBlock(block []byte, a *[]StringValue) ([]StringValue, error) {
 			return err
 		}
 		// Did string decoding have an error?
-		err = vdec.Error()
-		if err != nil {
-			return err
-		}
-		return nil
-	}()
+		return vdec.Error()
+	}(*a)
 
 	timeDecoderPool.Put(tdec)
 	stringDecoderPool.Put(vdec)
@@ -777,12 +982,12 @@ func unpackBlock(buf []byte) (ts, values []byte, err error) {
 }
 
 // ZigZagEncode converts a int64 to a uint64 by zig zagging negative and positive values
-// across even and odd numbers.  Eg. [0,-1,1,-2] becomes [0, 1, 2, 3]
+// across even and odd numbers.  Eg. [0,-1,1,-2] becomes [0, 1, 2, 3].
 func ZigZagEncode(x int64) uint64 {
 	return uint64(uint64(x<<1) ^ uint64((int64(x) >> 63)))
 }
 
-// ZigZagDecode converts a previously zigzag encoded uint64 back to a int64
+// ZigZagDecode converts a previously zigzag encoded uint64 back to a int64.
 func ZigZagDecode(v uint64) int64 {
 	return int64((v >> 1) ^ uint64((int64(v&1)<<63)>>63))
 }
@@ -800,8 +1005,15 @@ func getIntegerEncoder(sz int) IntegerEncoder {
 }
 func putIntegerEncoder(enc IntegerEncoder) { integerEncoderPool.Put(enc) }
 
-func getFloatEncoder() *FloatEncoder {
-	x := floatEncoderPool.Get(1024).(*FloatEncoder)
+func getUnsignedEncoder(sz int) IntegerEncoder {
+	x := integerEncoderPool.Get(sz).(IntegerEncoder)
+	x.Reset()
+	return x
+}
+func putUnsignedEncoder(enc IntegerEncoder) { integerEncoderPool.Put(enc) }
+
+func getFloatEncoder(sz int) *FloatEncoder {
+	x := floatEncoderPool.Get(sz).(*FloatEncoder)
 	x.Reset()
 	return x
 }
