@@ -66,6 +66,112 @@ func TestEngine_DeleteWALLoadMetadata(t *testing.T) {
 	}
 }
 
+// See https://github.com/influxdata/influxdb/issues/14229
+func TestEngine_DeleteSeriesAfterCacheSnapshot(t *testing.T) {
+	for _, index := range tsdb.RegisteredIndexes() {
+		t.Run(index, func(t *testing.T) {
+			e := MustOpenEngine(index)
+			defer e.Close()
+
+			if err := e.WritePointsString(
+				`cpu,host=A value=1.1 1000000000`,
+				`cpu,host=B value=1.2 2000000000`,
+			); err != nil {
+				t.Fatalf("failed to write points: %s", err.Error())
+			}
+
+			e.MeasurementFields([]byte("cpu")).CreateFieldIfNotExists([]byte("value"), influxql.Float)
+			e.CreateSeriesIfNotExists([]byte("cpu,host=A"), []byte("cpu"), models.NewTags(map[string]string{"host": "A"}))
+			e.CreateSeriesIfNotExists([]byte("cpu,host=B"), []byte("cpu"), models.NewTags(map[string]string{"host": "B"}))
+
+			// Verify series exist.
+			n, err := seriesExist(e, "cpu", []string{"host"})
+			if err != nil {
+				t.Fatal(err)
+			} else if got, exp := n, 2; got != exp {
+				t.Fatalf("got %d points, expected %d", got, exp)
+			}
+
+			// Simulate restart of server
+			if err := e.Reopen(); err != nil {
+				t.Fatal(err)
+			}
+
+			// Snapshot the cache
+			if err := e.WriteSnapshot(); err != nil {
+				t.Fatalf("failed to snapshot: %s", err.Error())
+			}
+
+			// Verify series exist.
+			n, err = seriesExist(e, "cpu", []string{"host"})
+			if err != nil {
+				t.Fatal(err)
+			} else if got, exp := n, 2; got != exp {
+				t.Fatalf("got %d points, expected %d", got, exp)
+			}
+
+			// Delete the series
+			itr := &seriesIterator{keys: [][]byte{
+				[]byte("cpu,host=A"),
+				[]byte("cpu,host=B"),
+			},
+			}
+			if err := e.DeleteSeriesRange(itr, math.MinInt64, math.MaxInt64); err != nil {
+				t.Fatalf("failed to delete series: %s", err.Error())
+			}
+
+			// Verify the series are no longer present.
+			n, err = seriesExist(e, "cpu", []string{"host"})
+			if err != nil {
+				t.Fatal(err)
+			} else if got, exp := n, 0; got != exp {
+				t.Fatalf("got %d points, expected %d", got, exp)
+			}
+
+			// Simulate restart of server
+			if err := e.Reopen(); err != nil {
+				t.Fatal(err)
+			}
+
+			// Verify the series are no longer present.
+			n, err = seriesExist(e, "cpu", []string{"host"})
+			if err != nil {
+				t.Fatal(err)
+			} else if got, exp := n, 0; got != exp {
+				t.Fatalf("got %d points, expected %d", got, exp)
+			}
+		})
+	}
+}
+
+func seriesExist(e *Engine, m string, dims []string) (int, error) {
+	itr, err := e.CreateIterator(context.Background(), "cpu", query.IteratorOptions{
+		Expr:       influxql.MustParseExpr(`value`),
+		Dimensions: []string{"host"},
+		StartTime:  influxql.MinTime,
+		EndTime:    influxql.MaxTime,
+		Ascending:  false,
+	})
+	if err != nil {
+		return 0, err
+	} else if itr == nil {
+		return 0, nil
+	}
+	defer itr.Close()
+	fitr := itr.(query.FloatIterator)
+
+	var n int
+	for {
+		p, err := fitr.Next()
+		if err != nil {
+			return 0, err
+		} else if p == nil {
+			return n, nil
+		}
+		n++
+	}
+}
+
 // Ensure that the engine can write & read shard digest files.
 func TestEngine_Digest(t *testing.T) {
 	e := MustOpenEngine(inmem.IndexName)
@@ -214,6 +320,57 @@ type span struct {
 	tspan *tsm1.DigestTimeSpan
 }
 
+// Ensure engine handles concurrent calls to Digest().
+func TestEngine_Digest_Concurrent(t *testing.T) {
+	e := MustOpenEngine(inmem.IndexName)
+	defer e.Close()
+
+	if err := e.Open(); err != nil {
+		t.Fatalf("failed to open tsm1 engine: %s", err.Error())
+	}
+
+	// Create a few points.
+	points := []models.Point{
+		MustParsePointString("cpu,host=A value=1.1 1000000000"),
+		MustParsePointString("cpu,host=B value=1.2 2000000000"),
+	}
+
+	if err := e.WritePoints(points); err != nil {
+		t.Fatalf("failed to write points: %s", err.Error())
+	}
+
+	// Force a compaction.
+	e.ScheduleFullCompaction()
+
+	// Start multiple waiting goroutines, ready to call Digest().
+	start := make(chan struct{})
+	errs := make(chan error)
+	wg := &sync.WaitGroup{}
+	for n := 0; n < 100; n++ {
+		wg.Add(1)
+		go func() {
+			defer wg.Done()
+			<-start
+			if _, _, err := e.Digest(); err != nil {
+				errs <- err
+			}
+		}()
+	}
+
+	// Goroutine to close errs channel after all routines have finished.
+	go func() { wg.Wait(); close(errs) }()
+
+	// Signal all goroutines to call Digest().
+	close(start)
+
+	// Check for digest errors.
+	for err := range errs {
+		if err != nil {
+			t.Fatal(err)
+		}
+	}
+}
+
 // Ensure that the engine will backup any TSM files created since the passed in time
 func TestEngine_Backup(t *testing.T) {
 	sfile := MustOpenSeriesFile()
@@ -320,6 +477,22 @@ func TestEngine_Backup(t *testing.T) {
 	if !strings.Contains(mostRecentFile, th.Name) || th.Name == "" {
 		t.Fatalf("file name doesn't match:\n\tgot: %s\n\texp: %s", th.Name, mostRecentFile)
 	}
+	storeDir := filepath.Dir(e.FileStore.Files()[0].Path())
+	dfd, err := os.Open(storeDir)
+	if err != nil {
+		t.Fatalf("cannot open filestore directory %s: %q", storeDir, err)
+	} else {
+		defer dfd.Close()
+	}
+	files, err := dfd.Readdirnames(0)
+	if err != nil {
+		t.Fatalf("cannot read directory %s: %q", storeDir, err)
+	}
+	for _, f := range files {
+		if strings.HasSuffix(f, tsm1.TmpTSMFileExtension) {
+			t.Fatalf("temporary directory for backup not cleaned up: %s", f)
+		}
+	}
 }
 
 func TestEngine_Export(t *testing.T) {
@@ -399,14 +572,14 @@ func TestEngine_Export(t *testing.T) {
 	}
 
 	// TEST 1: did we get any extra files not found in the store?
-	for k, _ := range fileData {
+	for k := range fileData {
 		if _, ok := fileNames[k]; !ok {
 			t.Errorf("exported a file not in the store: %s", k)
 		}
 	}
 
 	// TEST 2: did we miss any files that the store had?
-	for k, _ := range fileNames {
+	for k := range fileNames {
 		if _, ok := fileData[k]; !ok {
 			t.Errorf("failed to export a file from the store: %s", k)
 		}
@@ -816,7 +989,7 @@ func TestEngine_CreateIterator_TSM_Descending(t *testing.T) {
 	}
 }
 
-// Ensure engine can create an iterator with auxilary fields.
+// Ensure engine can create an iterator with auxiliary fields.
 func TestEngine_CreateIterator_Aux(t *testing.T) {
 	t.Parallel()
 
@@ -1799,12 +1972,12 @@ func TestEngine_CreateCursor_Ascending(t *testing.T) {
 			}
 			defer cur.Close()
 
-			fcur := cur.(tsdb.FloatBatchCursor)
-			ts, vs := fcur.Next()
-			if !cmp.Equal([]int64{2, 3, 10, 11}, ts) {
+			fcur := cur.(tsdb.FloatArrayCursor)
+			a := fcur.Next()
+			if !cmp.Equal([]int64{2, 3, 10, 11}, a.Timestamps) {
 				t.Fatal("unexpect timestamps")
 			}
-			if !cmp.Equal([]float64{1.2, 1.3, 10.1, 11.2}, vs) {
+			if !cmp.Equal([]float64{1.2, 1.3, 10.1, 11.2}, a.Values) {
 				t.Fatal("unexpect timestamps")
 			}
 		})
@@ -1859,12 +2032,12 @@ func TestEngine_CreateCursor_Descending(t *testing.T) {
 			}
 			defer cur.Close()
 
-			fcur := cur.(tsdb.FloatBatchCursor)
-			ts, vs := fcur.Next()
-			if !cmp.Equal([]int64{11, 10, 3, 2}, ts) {
+			fcur := cur.(tsdb.FloatArrayCursor)
+			a := fcur.Next()
+			if !cmp.Equal([]int64{11, 10, 3, 2}, a.Timestamps) {
 				t.Fatal("unexpect timestamps")
 			}
-			if !cmp.Equal([]float64{11.2, 10.1, 1.3, 1.2}, vs) {
+			if !cmp.Equal([]float64{11.2, 10.1, 1.3, 1.2}, a.Values) {
 				t.Fatal("unexpect timestamps")
 			}
 		})
@@ -1936,6 +2109,40 @@ func TestEngine_DisableEnableCompactions_Concurrent(t *testing.T) {
 			}
 		})
 	}
+
+}
+
+func TestEngine_WritePointsWithContext(t *testing.T) {
+	// Create a few points.
+	points := []models.Point{
+		MustParsePointString("cpu,host=A value=1.1 1000000000"),
+		MustParsePointString("cpu,host=B value=1.2,value2=8 2000000000"),
+	}
+
+	expectedPoints, expectedValues := int64(2), int64(3)
+
+	for _, index := range tsdb.RegisteredIndexes() {
+		t.Run(index, func(t *testing.T) {
+			e := MustOpenEngine(index)
+
+			var numPoints, numValues int64
+
+			ctx := context.WithValue(context.Background(), tsdb.StatPointsWritten, &numPoints)
+			ctx = context.WithValue(ctx, tsdb.StatValuesWritten, &numValues)
+
+			if err := e.WritePointsWithContext(ctx, points); err != nil {
+				t.Fatalf("failed to write points: %v", err)
+			}
+
+			if got, expected := numPoints, expectedPoints; got != expected {
+				t.Fatalf("Expected stats to return %d points; got %d", expected, got)
+			}
+
+			if got, expected := numValues, expectedValues; got != expected {
+				t.Fatalf("Expected stats to return %d points; got %d", expected, got)
+			}
+		})
+	}
 }
 
 func TestEngine_WritePoints_TypeConflict(t *testing.T) {
@@ -2017,6 +2224,40 @@ func TestEngine_WritePoints_Reload(t *testing.T) {
 	}
 }
 
+func TestEngine_Invalid_UTF8(t *testing.T) {
+	for _, index := range tsdb.RegisteredIndexes() {
+		t.Run(index, func(t *testing.T) {
+			name := []byte{255, 112, 114, 111, 99} // A known invalid UTF-8 string
+			field := []byte{255, 110, 101, 116}    // A known invalid UTF-8 string
+			p := MustParsePointString(fmt.Sprintf("%s,host=A %s=1.1 6000000000", name, field))
+
+			e, err := NewEngine(index)
+			if err != nil {
+				t.Fatal(err)
+			}
+
+			// mock the planner so compactions don't run during the test
+			e.CompactionPlan = &mockPlanner{}
+			if err := e.Open(); err != nil {
+				t.Fatal(err)
+			}
+			defer e.Close()
+
+			if err := e.CreateSeriesIfNotExists(p.Key(), p.Name(), p.Tags()); err != nil {
+				t.Fatalf("create series index error: %v", err)
+			}
+
+			if err := e.WritePoints([]models.Point{p}); err != nil {
+				t.Fatalf("failed to write points: %s", err.Error())
+			}
+
+			// Re-open the engine
+			if err := e.Reopen(); err != nil {
+				t.Fatal(err)
+			}
+		})
+	}
+}
 func BenchmarkEngine_WritePoints(b *testing.B) {
 	batchSizes := []int{10, 100, 1000, 5000, 10000}
 	for _, sz := range batchSizes {
@@ -2352,7 +2593,7 @@ func NewEngine(index string) (*Engine, error) {
 
 	opt := tsdb.NewEngineOptions()
 	opt.IndexVersion = index
-	if index == "inmem" {
+	if index == tsdb.InmemIndexName {
 		opt.InmemIndex = inmem.NewIndex(db, sfile)
 	}
 	// Initialise series id sets. Need to do this as it's normally done at the
