@@ -18,6 +18,7 @@ import (
 	"time"
 
 	"github.com/influxdata/influxdb/models"
+	"github.com/influxdata/influxdb/pkg/file"
 	"github.com/influxdata/influxdb/pkg/limiter"
 	"github.com/influxdata/influxdb/pkg/metrics"
 	"github.com/influxdata/influxdb/query"
@@ -45,10 +46,15 @@ type TSMFile interface {
 	// ReadAt returns all the values in the block identified by entry.
 	ReadAt(entry *IndexEntry, values []Value) ([]Value, error)
 	ReadFloatBlockAt(entry *IndexEntry, values *[]FloatValue) ([]FloatValue, error)
+	ReadFloatArrayBlockAt(entry *IndexEntry, values *tsdb.FloatArray) error
 	ReadIntegerBlockAt(entry *IndexEntry, values *[]IntegerValue) ([]IntegerValue, error)
+	ReadIntegerArrayBlockAt(entry *IndexEntry, values *tsdb.IntegerArray) error
 	ReadUnsignedBlockAt(entry *IndexEntry, values *[]UnsignedValue) ([]UnsignedValue, error)
+	ReadUnsignedArrayBlockAt(entry *IndexEntry, values *tsdb.UnsignedArray) error
 	ReadStringBlockAt(entry *IndexEntry, values *[]StringValue) ([]StringValue, error)
+	ReadStringArrayBlockAt(entry *IndexEntry, values *tsdb.StringArray) error
 	ReadBooleanBlockAt(entry *IndexEntry, values *[]BooleanValue) ([]BooleanValue, error)
+	ReadBooleanArrayBlockAt(entry *IndexEntry, values *tsdb.BooleanArray) error
 
 	// Entries returns the index entries for all blocks for the given key.
 	Entries(key []byte) []IndexEntry
@@ -104,9 +110,9 @@ type TSMFile interface {
 	// HasTombstones returns true if file contains values that have been deleted.
 	HasTombstones() bool
 
-	// TombstoneFiles returns the tombstone filestats if there are any tombstones
+	// TombstoneStats returns the tombstone filestats if there are any tombstones
 	// written for this file.
-	TombstoneFiles() []FileStat
+	TombstoneStats() TombstoneStat
 
 	// Close closes the underlying file resources.
 	Close() error
@@ -115,7 +121,7 @@ type TSMFile interface {
 	Size() uint32
 
 	// Rename renames the existing TSM file to a new name and replaces the mmap backing slice using the new
-	// file name.  Index and Reader state are not re-initialized.
+	// file name. Index and Reader state are not re-initialized.
 	Rename(path string) error
 
 	// Remove deletes the file from the filesystem.
@@ -184,6 +190,8 @@ type FileStore struct {
 
 	currentTempDirID int
 
+	parseFileName ParseFileNameFunc
+
 	obs tsdb.FileStoreObserver
 }
 
@@ -197,6 +205,14 @@ type FileStat struct {
 	MinKey, MaxKey   []byte
 }
 
+// TombstoneStat holds information about a possible tombstone file on disk.
+type TombstoneStat struct {
+	TombstoneExists bool
+	Path            string
+	LastModified    int64
+	Size            uint32
+}
+
 // OverlapsTimeRange returns true if the time range of the file intersect min and max.
 func (f FileStat) OverlapsTimeRange(min, max int64) bool {
 	return f.MinTime <= max && f.MaxTime >= min
@@ -226,7 +242,8 @@ func NewFileStore(dir string) *FileStore {
 			files:  map[string]TSMFile{},
 			logger: logger,
 		},
-		obs: noFileStoreObserver{},
+		obs:           noFileStoreObserver{},
+		parseFileName: DefaultParseFileName,
 	}
 	fs.purger.fileStore = fs
 	return fs
@@ -237,6 +254,14 @@ func (f *FileStore) WithObserver(obs tsdb.FileStoreObserver) {
 	f.obs = obs
 }
 
+func (f *FileStore) WithParseFileNameFunc(parseFileNameFunc ParseFileNameFunc) {
+	f.parseFileName = parseFileNameFunc
+}
+
+func (f *FileStore) ParseFileName(path string) (int, int, error) {
+	return f.parseFileName(path)
+}
+
 // enableTraceLogging must be called before the FileStore is opened.
 func (f *FileStore) enableTraceLogging(enabled bool) {
 	f.traceLogging = enabled
@@ -281,7 +306,7 @@ func (f *FileStore) Count() int {
 }
 
 // Files returns the slice of TSM files currently loaded. This is only used for
-// tests, and the files aren't guaranteed to stay valid in the presense of compactions.
+// tests, and the files aren't guaranteed to stay valid in the presence of compactions.
 func (f *FileStore) Files() []TSMFile {
 	f.mu.RLock()
 	defer f.mu.RUnlock()
@@ -493,7 +518,7 @@ func (f *FileStore) Open() error {
 	readerC := make(chan *res)
 	for i, fn := range files {
 		// Keep track of the latest ID
-		generation, _, err := ParseFileName(fn)
+		generation, _, err := f.parseFileName(fn)
 		if err != nil {
 			return err
 		}
@@ -525,11 +550,14 @@ func (f *FileStore) Open() error {
 			// the file, and continue loading the shard without it.
 			if err != nil {
 				f.logger.Error("Cannot read corrupt tsm file, renaming", zap.String("path", file.Name()), zap.Int("id", idx), zap.Error(err))
+				file.Close()
 				if e := os.Rename(file.Name(), file.Name()+"."+BadTSMFileExtension); e != nil {
 					f.logger.Error("Cannot rename corrupt tsm file", zap.String("path", file.Name()), zap.Int("id", idx), zap.Error(e))
 					readerC <- &res{r: df, err: fmt.Errorf("cannot rename corrupt file %s: %v", file.Name(), e)}
 					return
 				}
+				readerC <- &res{r: df, err: fmt.Errorf("cannot read corrupt file %s: %v", file.Name(), err)}
+				return
 			}
 
 			df.WithObserver(f.obs)
@@ -538,6 +566,7 @@ func (f *FileStore) Open() error {
 	}
 
 	var lm int64
+	isEmpty := true
 	for range files {
 		res := <-readerC
 		if res.err != nil {
@@ -549,7 +578,7 @@ func (f *FileStore) Open() error {
 
 		// Accumulate file store size stats
 		atomic.AddInt64(&f.stats.DiskBytes, int64(res.r.Size()))
-		for _, ts := range res.r.TombstoneFiles() {
+		if ts := res.r.TombstoneStats(); ts.TombstoneExists {
 			atomic.AddInt64(&f.stats.DiskBytes, int64(ts.Size))
 		}
 
@@ -557,9 +586,18 @@ func (f *FileStore) Open() error {
 		if res.r.LastModified() > lm {
 			lm = res.r.LastModified()
 		}
-
+		isEmpty = false
+	}
+	if isEmpty {
+		if fi, err := os.Stat(f.dir); err == nil {
+			f.lastModified = fi.ModTime().UTC()
+		} else {
+			close(readerC)
+			return err
+		}
+	} else {
+		f.lastModified = time.Unix(0, lm).UTC()
 	}
-	f.lastModified = time.Unix(0, lm).UTC()
 	close(readerC)
 
 	sort.Sort(tsmReaders(f.files))
@@ -712,17 +750,25 @@ func (f *FileStore) replace(oldFiles, newFiles []string, updatedFn func(r []TSMF
 			return err
 		}
 
-		var newName = file
-		if strings.HasSuffix(file, tsmTmpExt) {
+		var oldName, newName = file, file
+		if strings.HasSuffix(oldName, tsmTmpExt) {
 			// The new TSM files have a tmp extension.  First rename them.
 			newName = file[:len(file)-4]
-			if err := os.Rename(file, newName); err != nil {
+			if err := os.Rename(oldName, newName); err != nil {
 				return err
 			}
 		}
 
+		// Any error after this point should result in the file being bein named
+		// back to the original name. The caller then has the opportunity to
+		// remove it.
 		fd, err := os.Open(newName)
 		if err != nil {
+			if newName != oldName {
+				if err1 := os.Rename(newName, oldName); err1 != nil {
+					return err1
+				}
+			}
 			return err
 		}
 
@@ -735,6 +781,11 @@ func (f *FileStore) replace(oldFiles, newFiles []string, updatedFn func(r []TSMF
 
 		tsm, err := NewTSMReader(fd, WithMadviseWillNeed(f.tsmMMAPWillNeed))
 		if err != nil {
+			if newName != oldName {
+				if err1 := os.Rename(newName, oldName); err1 != nil {
+					return err1
+				}
+			}
 			return err
 		}
 		tsm.WithObserver(f.obs)
@@ -769,8 +820,8 @@ func (f *FileStore) replace(oldFiles, newFiles []string, updatedFn func(r []TSMF
 					return err
 				}
 
-				for _, t := range file.TombstoneFiles() {
-					if err := f.obs.FileUnlinking(t.Path); err != nil {
+				if ts := file.TombstoneStats(); ts.TombstoneExists {
+					if err := f.obs.FileUnlinking(ts.Path); err != nil {
 						return err
 					}
 				}
@@ -788,8 +839,8 @@ func (f *FileStore) replace(oldFiles, newFiles []string, updatedFn func(r []TSMF
 				if file.InUse() {
 					// Copy all the tombstones related to this TSM file
 					var deletes []string
-					for _, t := range file.TombstoneFiles() {
-						deletes = append(deletes, t.Path)
+					if ts := file.TombstoneStats(); ts.TombstoneExists {
+						deletes = append(deletes, ts.Path)
 					}
 
 					// Rename the TSM file used by this reader
@@ -826,7 +877,7 @@ func (f *FileStore) replace(oldFiles, newFiles []string, updatedFn func(r []TSMF
 		}
 	}
 
-	if err := syncDir(f.dir); err != nil {
+	if err := file.SyncDir(f.dir); err != nil {
 		return err
 	}
 
@@ -851,10 +902,9 @@ func (f *FileStore) replace(oldFiles, newFiles []string, updatedFn func(r []TSMF
 	var totalSize int64
 	for _, file := range f.files {
 		totalSize += int64(file.Size())
-		for _, ts := range file.TombstoneFiles() {
+		if ts := file.TombstoneStats(); ts.TombstoneExists {
 			totalSize += int64(ts.Size)
 		}
-
 	}
 	atomic.StoreInt64(&f.stats.DiskBytes, totalSize)
 
@@ -891,7 +941,9 @@ func (f *FileStore) BlockCount(path string, idx int) int {
 				}
 			}
 			_, _, _, _, _, block, _ := iter.Read()
-			return BlockCount(block)
+			// on Error, BlockCount(block) returns 0 for cnt
+			cnt, _ := BlockCount(block)
+			return cnt
 		}
 	}
 	return 0
@@ -1005,6 +1057,24 @@ func (f *FileStore) locations(key []byte, t int64, ascending bool) []*location {
 	return locations
 }
 
+// MakeSnapshotLinks creates hardlinks from the supplied TSMFiles to
+// corresponding files under a supplied directory.
+func (f *FileStore) MakeSnapshotLinks(destPath string, files []TSMFile) (returnErr error) {
+	for _, tsmf := range files {
+		newpath := filepath.Join(destPath, filepath.Base(tsmf.Path()))
+		if err := copyOrLink(tsmf.Path(), newpath); err != nil {
+			return err
+		}
+		if tf := tsmf.TombstoneStats(); tf.TombstoneExists {
+			newpath := filepath.Join(destPath, filepath.Base(tf.Path))
+			if err := copyOrLink(tf.Path, newpath); err != nil {
+				return err
+			}
+		}
+	}
+	return nil
+}
+
 // CreateSnapshot creates hardlinks for all tsm and tombstone files
 // in the path provided.
 func (f *FileStore) CreateSnapshot() (string, error) {
@@ -1024,29 +1094,24 @@ func (f *FileStore) CreateSnapshot() (string, error) {
 	// increment and keep track of the current temp dir for when we drop the lock.
 	// this ensures we are the only writer to the directory.
 	f.currentTempDirID += 1
-	tmpPath := fmt.Sprintf("%d.%s", f.currentTempDirID, TmpTSMFileExtension)
-	tmpPath = filepath.Join(f.dir, tmpPath)
+	tmpPath := filepath.Join(f.dir, fmt.Sprintf("%d.%s", f.currentTempDirID, TmpTSMFileExtension))
 	f.mu.Unlock()
 
-	// create the tmp directory and add the hard links. there is no longer any shared
-	// mutable state.
-	err := os.Mkdir(tmpPath, 0777)
-	if err != nil {
+	// create the tmp directory and add the hard links. there is no longer any
+	// shared mutable state.
+	if err := os.Mkdir(tmpPath, 0777); err != nil {
 		return "", err
 	}
-	for _, tsmf := range files {
-		newpath := filepath.Join(tmpPath, filepath.Base(tsmf.Path()))
-		if err := os.Link(tsmf.Path(), newpath); err != nil {
-			return "", fmt.Errorf("error creating tsm hard link: %q", err)
-		}
-		for _, tf := range tsmf.TombstoneFiles() {
-			newpath := filepath.Join(tmpPath, filepath.Base(tf.Path))
-			if err := os.Link(tf.Path, newpath); err != nil {
-				return "", fmt.Errorf("error creating tombstone hard link: %q", err)
-			}
-		}
-	}
 
+	if err := f.MakeSnapshotLinks(tmpPath, files); err != nil {
+		// remove temporary directory since we couldn't create our hard links.
+		if err := os.RemoveAll(tmpPath); err != nil {
+			// report if, for some reason, we couldn't remove our temporary
+			// directory.
+			return "", fmt.Errorf("CreateSnapshot() failed to create links and failed to remove temporary direcotry %v: %v", tmpPath, err)
+		}
+		return "", err
+	}
 	return tmpPath, nil
 }
 
@@ -1062,8 +1127,8 @@ func DefaultFormatFileName(generation, sequence int) string {
 // ParseFileNameFunc is executed when parsing a TSM filename into generation & sequence.
 type ParseFileNameFunc func(name string) (generation, sequence int, err error)
 
-// ParseFileName is used to parse the filenames of TSM files.
-var ParseFileName ParseFileNameFunc = func(name string) (int, int, error) {
+// DefaultParseFileName is used to parse the filenames of TSM files.
+func DefaultParseFileName(name string) (int, int, error) {
 	base := filepath.Base(name)
 	idx := strings.Index(base, ".")
 	if idx == -1 {
@@ -1308,41 +1373,6 @@ func (c *KeyCursor) nextDescending() {
 	}
 }
 
-func (c *KeyCursor) filterFloatValues(tombstones []TimeRange, values FloatValues) FloatValues {
-	for _, t := range tombstones {
-		values = values.Exclude(t.Min, t.Max)
-	}
-	return values
-}
-
-func (c *KeyCursor) filterIntegerValues(tombstones []TimeRange, values IntegerValues) IntegerValues {
-	for _, t := range tombstones {
-		values = values.Exclude(t.Min, t.Max)
-	}
-	return values
-}
-
-func (c *KeyCursor) filterUnsignedValues(tombstones []TimeRange, values UnsignedValues) UnsignedValues {
-	for _, t := range tombstones {
-		values = values.Exclude(t.Min, t.Max)
-	}
-	return values
-}
-
-func (c *KeyCursor) filterStringValues(tombstones []TimeRange, values StringValues) StringValues {
-	for _, t := range tombstones {
-		values = values.Exclude(t.Min, t.Max)
-	}
-	return values
-}
-
-func (c *KeyCursor) filterBooleanValues(tombstones []TimeRange, values BooleanValues) BooleanValues {
-	for _, t := range tombstones {
-		values = values.Exclude(t.Min, t.Max)
-	}
-	return values
-}
-
 type purger struct {
 	mu        sync.RWMutex
 	fileStore *FileStore
