@@ -13,6 +13,8 @@ import (
 	"sync/atomic"
 
 	"github.com/influxdata/influxdb/pkg/bytesutil"
+	"github.com/influxdata/influxdb/pkg/file"
+	"github.com/influxdata/influxdb/tsdb"
 )
 
 // ErrFileInUse is returned when attempting to remove or close a TSM file that is still being used.
@@ -28,7 +30,8 @@ type TSMReader struct {
 	refs   int64
 	refsWG sync.WaitGroup
 
-	mu sync.RWMutex
+	madviseWillNeed bool // Hint to the kernel with MADV_WILLNEED.
+	mu              sync.RWMutex
 
 	// accessor provides access and decoding of blocks for the reader.
 	accessor blockAccessor
@@ -195,6 +198,7 @@ func (b *BlockIterator) Read() (key []byte, minTime int64, maxTime int64, typ by
 	}
 	checksum, buf, err = b.r.ReadBytes(&b.entries[0], nil)
 	if err != nil {
+		b.err = err
 		return nil, 0, 0, 0, 0, nil, err
 	}
 	return b.key, b.entries[0].MinTime, b.entries[0].MaxTime, b.typ, checksum, buf, err
@@ -205,28 +209,21 @@ func (b *BlockIterator) Err() error {
 	return b.err
 }
 
-// blockAccessor abstracts a method of accessing blocks from a
-// TSM file.
-type blockAccessor interface {
-	init() (*indirectIndex, error)
-	read(key []byte, timestamp int64) ([]Value, error)
-	readAll(key []byte) ([]Value, error)
-	readBlock(entry *IndexEntry, values []Value) ([]Value, error)
-	readFloatBlock(entry *IndexEntry, values *[]FloatValue) ([]FloatValue, error)
-	readIntegerBlock(entry *IndexEntry, values *[]IntegerValue) ([]IntegerValue, error)
-	readUnsignedBlock(entry *IndexEntry, values *[]UnsignedValue) ([]UnsignedValue, error)
-	readStringBlock(entry *IndexEntry, values *[]StringValue) ([]StringValue, error)
-	readBooleanBlock(entry *IndexEntry, values *[]BooleanValue) ([]BooleanValue, error)
-	readBytes(entry *IndexEntry, buf []byte) (uint32, []byte, error)
-	rename(path string) error
-	path() string
-	close() error
-	free() error
+type tsmReaderOption func(*TSMReader)
+
+// WithMadviseWillNeed is an option for specifying whether to provide a MADV_WILL need hint to the kernel.
+var WithMadviseWillNeed = func(willNeed bool) tsmReaderOption {
+	return func(r *TSMReader) {
+		r.madviseWillNeed = willNeed
+	}
 }
 
 // NewTSMReader returns a new TSMReader from the given file.
-func NewTSMReader(f *os.File) (*TSMReader, error) {
+func NewTSMReader(f *os.File, options ...tsmReaderOption) (*TSMReader, error) {
 	t := &TSMReader{}
+	for _, option := range options {
+		option(t)
+	}
 
 	stat, err := f.Stat()
 	if err != nil {
@@ -235,7 +232,8 @@ func NewTSMReader(f *os.File) (*TSMReader, error) {
 	t.size = stat.Size()
 	t.lastModified = stat.ModTime().UnixNano()
 	t.accessor = &mmapAccessor{
-		f: f,
+		f:            f,
+		mmapWillNeed: t.madviseWillNeed,
 	}
 
 	index, err := t.accessor.init()
@@ -244,7 +242,7 @@ func NewTSMReader(f *os.File) (*TSMReader, error) {
 	}
 
 	t.index = index
-	t.tombstoner = &Tombstoner{Path: t.Path(), FilterFn: index.ContainsKey}
+	t.tombstoner = NewTombstoner(t.Path(), index.ContainsKey)
 
 	if err := t.applyTombstones(); err != nil {
 		return nil, err
@@ -253,6 +251,11 @@ func NewTSMReader(f *os.File) (*TSMReader, error) {
 	return t, nil
 }
 
+// WithObserver sets the observer for the TSM reader.
+func (t *TSMReader) WithObserver(obs tsdb.FileStoreObserver) {
+	t.tombstoner.WithObserver(obs)
+}
+
 func (t *TSMReader) applyTombstones() error {
 	var cur, prev Tombstone
 	batch := make([][]byte, 0, 4096)
@@ -329,46 +332,6 @@ func (t *TSMReader) ReadAt(entry *IndexEntry, vals []Value) ([]Value, error) {
 	return v, err
 }
 
-// ReadFloatBlockAt returns the float values corresponding to the given index entry.
-func (t *TSMReader) ReadFloatBlockAt(entry *IndexEntry, vals *[]FloatValue) ([]FloatValue, error) {
-	t.mu.RLock()
-	v, err := t.accessor.readFloatBlock(entry, vals)
-	t.mu.RUnlock()
-	return v, err
-}
-
-// ReadIntegerBlockAt returns the integer values corresponding to the given index entry.
-func (t *TSMReader) ReadIntegerBlockAt(entry *IndexEntry, vals *[]IntegerValue) ([]IntegerValue, error) {
-	t.mu.RLock()
-	v, err := t.accessor.readIntegerBlock(entry, vals)
-	t.mu.RUnlock()
-	return v, err
-}
-
-// ReadUnsignedBlockAt returns the unsigned integer values corresponding to the given index entry.
-func (t *TSMReader) ReadUnsignedBlockAt(entry *IndexEntry, vals *[]UnsignedValue) ([]UnsignedValue, error) {
-	t.mu.RLock()
-	v, err := t.accessor.readUnsignedBlock(entry, vals)
-	t.mu.RUnlock()
-	return v, err
-}
-
-// ReadStringBlockAt returns the string values corresponding to the given index entry.
-func (t *TSMReader) ReadStringBlockAt(entry *IndexEntry, vals *[]StringValue) ([]StringValue, error) {
-	t.mu.RLock()
-	v, err := t.accessor.readStringBlock(entry, vals)
-	t.mu.RUnlock()
-	return v, err
-}
-
-// ReadBooleanBlockAt returns the boolean values corresponding to the given index entry.
-func (t *TSMReader) ReadBooleanBlockAt(entry *IndexEntry, vals *[]BooleanValue) ([]BooleanValue, error) {
-	t.mu.RLock()
-	v, err := t.accessor.readBooleanBlock(entry, vals)
-	t.mu.RUnlock()
-	return v, err
-}
-
 // Read returns the values corresponding to the block at the given key and timestamp.
 func (t *TSMReader) Read(key []byte, timestamp int64) ([]Value, error) {
 	t.mu.RLock()
@@ -561,7 +524,7 @@ func (t *TSMReader) Size() uint32 {
 func (t *TSMReader) LastModified() int64 {
 	t.mu.RLock()
 	lm := t.lastModified
-	for _, ts := range t.tombstoner.TombstoneFiles() {
+	if ts := t.tombstoner.TombstoneStats(); ts.TombstoneExists {
 		if ts.LastModified > lm {
 			lm = ts.LastModified
 		}
@@ -579,9 +542,9 @@ func (t *TSMReader) HasTombstones() bool {
 }
 
 // TombstoneFiles returns any tombstone files associated with this TSM file.
-func (t *TSMReader) TombstoneFiles() []FileStat {
+func (t *TSMReader) TombstoneStats() TombstoneStat {
 	t.mu.RLock()
-	fs := t.tombstoner.TombstoneFiles()
+	fs := t.tombstoner.TombstoneStats()
 	t.mu.RUnlock()
 	return fs
 }
@@ -751,7 +714,7 @@ type indirectIndex struct {
 	// Using this offset slice we can find `Key 2` by doing a binary search
 	// over the offsets slice.  Instead of comparing the value in the offsets
 	// (e.g. `62`), we use that as an index into the underlying index to
-	// retrieve the key at postion `62` and perform our comparisons with that.
+	// retrieve the key at position `62` and perform our comparisons with that.
 
 	// When we have identified the correct position in the index for a given
 	// key, we could perform another binary search or a linear scan.  This
@@ -839,8 +802,13 @@ func (d *indirectIndex) searchOffset(key []byte) int {
 // search returns the byte position of key in the index.  If key is not
 // in the index, len(index) is returned.
 func (d *indirectIndex) search(key []byte) int {
+	if !d.ContainsKey(key) {
+		return len(d.b)
+	}
+
 	// We use a binary search across our indirect offsets (pointers to all the keys
 	// in the index slice).
+	// TODO(sgc): this should be inlined to `indirectIndex` as it is only used here
 	i := bytesutil.SearchBytesFixed(d.offsets, 4, func(x []byte) bool {
 		// i is the position in offsets we are at so get offset it points to
 		offset := int32(binary.BigEndian.Uint32(x))
@@ -858,7 +826,7 @@ func (d *indirectIndex) search(key []byte) int {
 		_, k := readKey(d.b[ofs:])
 
 		// The search may have returned an i == 0 which could indicated that the value
-		// searched should be inserted at postion 0.  Make sure the key in the index
+		// searched should be inserted at position 0.  Make sure the key in the index
 		// matches the search value.
 		if !bytes.Equal(key, k) {
 			return len(d.b)
@@ -1010,7 +978,7 @@ func (d *indirectIndex) Delete(keys [][]byte) {
 
 		if len(keys) > 0 && bytes.Equal(keys[0], indexKey) {
 			keys = keys[1:]
-			copy(d.offsets[i:i+4], nilOffset[:])
+			copy(d.offsets[i:i+4], nilOffset)
 		}
 	}
 	d.offsets = bytesutil.Pack(d.offsets, 4, 255)
@@ -1333,15 +1301,15 @@ func (d *indirectIndex) Close() error {
 // mmapAccess is mmap based block accessor.  It access blocks through an
 // MMAP file interface.
 type mmapAccessor struct {
-	// Counter incremented everytime the mmapAccessor is accessed
-	accessCount uint64
-	// Counter to determine whether the accessor can free its resources
-	freeCount uint64
+	accessCount uint64 // Counter incremented everytime the mmapAccessor is accessed
+	freeCount   uint64 // Counter to determine whether the accessor can free its resources
+
+	mmapWillNeed bool // If true then mmap advise value MADV_WILLNEED will be provided the kernel for b.
 
 	mu sync.RWMutex
+	b  []byte
+	f  *os.File
 
-	f     *os.File
-	b     []byte
 	index *indirectIndex
 }
 
@@ -1372,6 +1340,15 @@ func (m *mmapAccessor) init() (*indirectIndex, error) {
 		return nil, fmt.Errorf("mmapAccessor: byte slice too small for indirectIndex")
 	}
 
+	// Hint to the kernel that we will be reading the file.  It would be better to hint
+	// that we will be reading the index section, but that's not been
+	// implemented as yet.
+	if m.mmapWillNeed {
+		if err := madviseWillNeed(m.b); err != nil {
+			return nil, err
+		}
+	}
+
 	indexOfsPos := len(m.b) - 8
 	indexStart := binary.BigEndian.Uint64(m.b[indexOfsPos : indexOfsPos+8])
 	if indexStart >= uint64(indexOfsPos) {
@@ -1436,7 +1413,7 @@ func (m *mmapAccessor) rename(path string) error {
 		return err
 	}
 
-	if err := renameFile(m.f.Name(), path); err != nil {
+	if err := file.RenameFile(m.f.Name(), path); err != nil {
 		return err
 	}
 
@@ -1455,7 +1432,14 @@ func (m *mmapAccessor) rename(path string) error {
 	}
 
 	m.b, err = mmap(m.f, 0, int(stat.Size()))
-	return err
+	if err != nil {
+		return err
+	}
+
+	if m.mmapWillNeed {
+		return madviseWillNeed(m.b)
+	}
+	return nil
 }
 
 func (m *mmapAccessor) read(key []byte, timestamp int64) ([]Value, error) {
@@ -1486,101 +1470,6 @@ func (m *mmapAccessor) readBlock(entry *IndexEntry, values []Value) ([]Value, er
 	return values, nil
 }
 
-func (m *mmapAccessor) readFloatBlock(entry *IndexEntry, values *[]FloatValue) ([]FloatValue, error) {
-	m.incAccess()
-
-	m.mu.RLock()
-	if int64(len(m.b)) < entry.Offset+int64(entry.Size) {
-		m.mu.RUnlock()
-		return nil, ErrTSMClosed
-	}
-
-	a, err := DecodeFloatBlock(m.b[entry.Offset+4:entry.Offset+int64(entry.Size)], values)
-	m.mu.RUnlock()
-
-	if err != nil {
-		return nil, err
-	}
-
-	return a, nil
-}
-
-func (m *mmapAccessor) readIntegerBlock(entry *IndexEntry, values *[]IntegerValue) ([]IntegerValue, error) {
-	m.incAccess()
-
-	m.mu.RLock()
-	if int64(len(m.b)) < entry.Offset+int64(entry.Size) {
-		m.mu.RUnlock()
-		return nil, ErrTSMClosed
-	}
-
-	a, err := DecodeIntegerBlock(m.b[entry.Offset+4:entry.Offset+int64(entry.Size)], values)
-	m.mu.RUnlock()
-
-	if err != nil {
-		return nil, err
-	}
-
-	return a, nil
-}
-
-func (m *mmapAccessor) readUnsignedBlock(entry *IndexEntry, values *[]UnsignedValue) ([]UnsignedValue, error) {
-	m.incAccess()
-
-	m.mu.RLock()
-	if int64(len(m.b)) < entry.Offset+int64(entry.Size) {
-		m.mu.RUnlock()
-		return nil, ErrTSMClosed
-	}
-
-	a, err := DecodeUnsignedBlock(m.b[entry.Offset+4:entry.Offset+int64(entry.Size)], values)
-	m.mu.RUnlock()
-
-	if err != nil {
-		return nil, err
-	}
-
-	return a, nil
-}
-
-func (m *mmapAccessor) readStringBlock(entry *IndexEntry, values *[]StringValue) ([]StringValue, error) {
-	m.incAccess()
-
-	m.mu.RLock()
-	if int64(len(m.b)) < entry.Offset+int64(entry.Size) {
-		m.mu.RUnlock()
-		return nil, ErrTSMClosed
-	}
-
-	a, err := DecodeStringBlock(m.b[entry.Offset+4:entry.Offset+int64(entry.Size)], values)
-	m.mu.RUnlock()
-
-	if err != nil {
-		return nil, err
-	}
-
-	return a, nil
-}
-
-func (m *mmapAccessor) readBooleanBlock(entry *IndexEntry, values *[]BooleanValue) ([]BooleanValue, error) {
-	m.incAccess()
-
-	m.mu.RLock()
-	if int64(len(m.b)) < entry.Offset+int64(entry.Size) {
-		m.mu.RUnlock()
-		return nil, ErrTSMClosed
-	}
-
-	a, err := DecodeBooleanBlock(m.b[entry.Offset+4:entry.Offset+int64(entry.Size)], values)
-	m.mu.RUnlock()
-
-	if err != nil {
-		return nil, err
-	}
-
-	return a, nil
-}
-
 func (m *mmapAccessor) readBytes(entry *IndexEntry, b []byte) (uint32, []byte, error) {
 	m.incAccess()
 
