@@ -1,11 +1,13 @@
 package tsm1_test
 
 import (
+	"bytes"
 	"io/ioutil"
 	"os"
 	"testing"
 
 	"github.com/influxdata/influxdb/tsdb/engine/tsm1"
+	"github.com/stretchr/testify/require"
 )
 
 func TestTombstoner_Add(t *testing.T) {
@@ -13,17 +15,15 @@ func TestTombstoner_Add(t *testing.T) {
 	defer func() { os.RemoveAll(dir) }()
 
 	f := MustTempFile(dir)
-	ts := &tsm1.Tombstoner{Path: f.Name()}
+	ts := tsm1.NewTombstoner(f.Name(), nil)
 
 	entries := mustReadAll(ts)
 	if got, exp := len(entries), 0; got != exp {
 		t.Fatalf("length mismatch: got %v, exp %v", got, exp)
 	}
 
-	stats := ts.TombstoneFiles()
-	if got, exp := len(stats), 0; got != exp {
-		t.Fatalf("stat length mismatch: got %v, exp %v", got, exp)
-	}
+	stats := ts.TombstoneStats()
+	require.False(t, stats.TombstoneExists)
 
 	ts.Add([][]byte{[]byte("foo")})
 
@@ -32,22 +32,11 @@ func TestTombstoner_Add(t *testing.T) {
 	}
 
 	entries = mustReadAll(ts)
-	stats = ts.TombstoneFiles()
-	if got, exp := len(stats), 1; got != exp {
-		t.Fatalf("stat length mismatch: got %v, exp %v", got, exp)
-	}
-
-	if stats[0].Size == 0 {
-		t.Fatalf("got size %v, exp > 0", stats[0].Size)
-	}
-
-	if stats[0].LastModified == 0 {
-		t.Fatalf("got lastModified %v, exp > 0", stats[0].LastModified)
-	}
-
-	if stats[0].Path == "" {
-		t.Fatalf("got path %v, exp != ''", stats[0].Path)
-	}
+	stats = ts.TombstoneStats()
+	require.True(t, stats.TombstoneExists)
+	require.NotZero(t, stats.Size)
+	require.NotZero(t, stats.LastModified)
+	require.NotEmpty(t, stats.Path)
 
 	if got, exp := len(entries), 1; got != exp {
 		t.Fatalf("length mismatch: got %v, exp %v", got, exp)
@@ -58,7 +47,7 @@ func TestTombstoner_Add(t *testing.T) {
 	}
 
 	// Use a new Tombstoner to verify values are persisted
-	ts = &tsm1.Tombstoner{Path: f.Name()}
+	ts = tsm1.NewTombstoner(f.Name(), nil)
 	entries = mustReadAll(ts)
 	if got, exp := len(entries), 1; got != exp {
 		t.Fatalf("length mismatch: got %v, exp %v", got, exp)
@@ -69,23 +58,70 @@ func TestTombstoner_Add(t *testing.T) {
 	}
 }
 
-func TestTombstoner_Add_Multiple(t *testing.T) {
+func TestTombstoner_Add_LargeKey(t *testing.T) {
 	dir := MustTempDir()
 	defer func() { os.RemoveAll(dir) }()
 
 	f := MustTempFile(dir)
-	ts := &tsm1.Tombstoner{Path: f.Name()}
+	ts := tsm1.NewTombstoner(f.Name(), nil)
 
 	entries := mustReadAll(ts)
 	if got, exp := len(entries), 0; got != exp {
 		t.Fatalf("length mismatch: got %v, exp %v", got, exp)
 	}
 
-	stats := ts.TombstoneFiles()
-	if got, exp := len(stats), 0; got != exp {
-		t.Fatalf("stat length mismatch: got %v, exp %v", got, exp)
+	stats := ts.TombstoneStats()
+	require.False(t, stats.TombstoneExists)
+
+	key := bytes.Repeat([]byte{'a'}, 4096)
+	ts.Add([][]byte{key})
+
+	if err := ts.Flush(); err != nil {
+		t.Fatalf("unexpected error flushing tombstone: %v", err)
 	}
 
+	entries = mustReadAll(ts)
+	stats = ts.TombstoneStats()
+	require.True(t, stats.TombstoneExists)
+	require.NotZero(t, stats.Size)
+	require.NotZero(t, stats.LastModified)
+	require.NotEmpty(t, stats.Path)
+
+	if got, exp := len(entries), 1; got != exp {
+		t.Fatalf("length mismatch: got %v, exp %v", got, exp)
+	}
+
+	if got, exp := string(entries[0].Key), string(key); got != exp {
+		t.Fatalf("value mismatch: got %v, exp %v", got, exp)
+	}
+
+	// Use a new Tombstoner to verify values are persisted
+	ts = tsm1.NewTombstoner(f.Name(), nil)
+	entries = mustReadAll(ts)
+	if got, exp := len(entries), 1; got != exp {
+		t.Fatalf("length mismatch: got %v, exp %v", got, exp)
+	}
+
+	if got, exp := string(entries[0].Key), string(key); got != exp {
+		t.Fatalf("value mismatch: got %v, exp %v", got, exp)
+	}
+}
+
+func TestTombstoner_Add_Multiple(t *testing.T) {
+	dir := MustTempDir()
+	defer func() { os.RemoveAll(dir) }()
+
+	f := MustTempFile(dir)
+	ts := tsm1.NewTombstoner(f.Name(), nil)
+
+	entries := mustReadAll(ts)
+	if got, exp := len(entries), 0; got != exp {
+		t.Fatalf("length mismatch: got %v, exp %v", got, exp)
+	}
+
+	stats := ts.TombstoneStats()
+	require.False(t, stats.TombstoneExists)
+
 	ts.Add([][]byte{[]byte("foo")})
 
 	if err := ts.Flush(); err != nil {
@@ -99,22 +135,11 @@ func TestTombstoner_Add_Multiple(t *testing.T) {
 	}
 
 	entries = mustReadAll(ts)
-	stats = ts.TombstoneFiles()
-	if got, exp := len(stats), 1; got != exp {
-		t.Fatalf("stat length mismatch: got %v, exp %v", got, exp)
-	}
-
-	if stats[0].Size == 0 {
-		t.Fatalf("got size %v, exp > 0", stats[0].Size)
-	}
-
-	if stats[0].LastModified == 0 {
-		t.Fatalf("got lastModified %v, exp > 0", stats[0].LastModified)
-	}
-
-	if stats[0].Path == "" {
-		t.Fatalf("got path %v, exp != ''", stats[0].Path)
-	}
+	stats = ts.TombstoneStats()
+	require.True(t, stats.TombstoneExists)
+	require.NotZero(t, stats.Size)
+	require.NotZero(t, stats.LastModified)
+	require.NotEmpty(t, stats.Path)
 
 	if got, exp := len(entries), 2; got != exp {
 		t.Fatalf("length mismatch: got %v, exp %v", got, exp)
@@ -129,7 +154,7 @@ func TestTombstoner_Add_Multiple(t *testing.T) {
 	}
 
 	// Use a new Tombstoner to verify values are persisted
-	ts = &tsm1.Tombstoner{Path: f.Name()}
+	ts = tsm1.NewTombstoner(f.Name(), nil)
 	entries = mustReadAll(ts)
 	if got, exp := len(entries), 2; got != exp {
 		t.Fatalf("length mismatch: got %v, exp %v", got, exp)
@@ -150,7 +175,7 @@ func TestTombstoner_Add_Empty(t *testing.T) {
 	defer func() { os.RemoveAll(dir) }()
 
 	f := MustTempFile(dir)
-	ts := &tsm1.Tombstoner{Path: f.Name()}
+	ts := tsm1.NewTombstoner(f.Name(), nil)
 
 	entries := mustReadAll(ts)
 	if got, exp := len(entries), 0; got != exp {
@@ -164,17 +189,14 @@ func TestTombstoner_Add_Empty(t *testing.T) {
 	}
 
 	// Use a new Tombstoner to verify values are persisted
-	ts = &tsm1.Tombstoner{Path: f.Name()}
+	ts = tsm1.NewTombstoner(f.Name(), nil)
 	entries = mustReadAll(ts)
 	if got, exp := len(entries), 0; got != exp {
 		t.Fatalf("length mismatch: got %v, exp %v", got, exp)
 	}
 
-	stats := ts.TombstoneFiles()
-	if got, exp := len(stats), 0; got != exp {
-		t.Fatalf("stat length mismatch: got %v, exp %v", got, exp)
-	}
-
+	stats := ts.TombstoneStats()
+	require.False(t, stats.TombstoneExists)
 }
 
 func TestTombstoner_Delete(t *testing.T) {
@@ -182,7 +204,7 @@ func TestTombstoner_Delete(t *testing.T) {
 	defer func() { os.RemoveAll(dir) }()
 
 	f := MustTempFile(dir)
-	ts := &tsm1.Tombstoner{Path: f.Name()}
+	ts := tsm1.NewTombstoner(f.Name(), nil)
 
 	ts.Add([][]byte{[]byte("foo")})
 
@@ -191,7 +213,7 @@ func TestTombstoner_Delete(t *testing.T) {
 	}
 
 	// Use a new Tombstoner to verify values are persisted
-	ts = &tsm1.Tombstoner{Path: f.Name()}
+	ts = tsm1.NewTombstoner(f.Name(), nil)
 	entries := mustReadAll(ts)
 	if got, exp := len(entries), 1; got != exp {
 		t.Fatalf("length mismatch: got %v, exp %v", got, exp)
@@ -205,12 +227,10 @@ func TestTombstoner_Delete(t *testing.T) {
 		fatal(t, "delete tombstone", err)
 	}
 
-	stats := ts.TombstoneFiles()
-	if got, exp := len(stats), 0; got != exp {
-		t.Fatalf("stat length mismatch: got %v, exp %v", got, exp)
-	}
+	stats := ts.TombstoneStats()
+	require.False(t, stats.TombstoneExists)
 
-	ts = &tsm1.Tombstoner{Path: f.Name()}
+	ts = tsm1.NewTombstoner(f.Name(), nil)
 	entries = mustReadAll(ts)
 	if got, exp := len(entries), 0; got != exp {
 		t.Fatalf("length mismatch: got %v, exp %v", got, exp)
@@ -227,11 +247,11 @@ func TestTombstoner_ReadV1(t *testing.T) {
 	}
 	f.Close()
 
-	if err := os.Rename(f.Name(), f.Name()+".tombstone"); err != nil {
+	if err := os.Rename(f.Name(), f.Name()+"."+tsm1.TombstoneFileExtension); err != nil {
 		t.Fatalf("rename tombstone failed: %v", err)
 	}
 
-	ts := &tsm1.Tombstoner{Path: f.Name()}
+	ts := tsm1.NewTombstoner(f.Name(), nil)
 
 	// Read once
 	_ = mustReadAll(ts)
@@ -248,7 +268,7 @@ func TestTombstoner_ReadV1(t *testing.T) {
 	}
 
 	// Use a new Tombstoner to verify values are persisted
-	ts = &tsm1.Tombstoner{Path: f.Name()}
+	ts = tsm1.NewTombstoner(f.Name(), nil)
 	entries = mustReadAll(ts)
 	if got, exp := len(entries), 1; got != exp {
 		t.Fatalf("length mismatch: got %v, exp %v", got, exp)
@@ -266,11 +286,11 @@ func TestTombstoner_ReadEmptyV1(t *testing.T) {
 	f := MustTempFile(dir)
 	f.Close()
 
-	if err := os.Rename(f.Name(), f.Name()+".tombstone"); err != nil {
+	if err := os.Rename(f.Name(), f.Name()+"."+tsm1.TombstoneFileExtension); err != nil {
 		t.Fatalf("rename tombstone failed: %v", err)
 	}
 
-	ts := &tsm1.Tombstoner{Path: f.Name()}
+	ts := tsm1.NewTombstoner(f.Name(), nil)
 
 	_ = mustReadAll(ts)
 
