@@ -31,9 +31,8 @@ const IndexName = tsdb.TSI1IndexName
 var ErrCompactionInterrupted = errors.New("tsi1: compaction interrupted")
 
 func init() {
-	// FIXME(edd): Remove this.
-	if os.Getenv("TSI_PARTITIONS") != "" {
-		i, err := strconv.Atoi(os.Getenv("TSI_PARTITIONS"))
+	if os.Getenv("INFLUXDB_EXP_TSI_PARTITIONS") != "" {
+		i, err := strconv.Atoi(os.Getenv("INFLUXDB_EXP_TSI_PARTITIONS"))
 		if err != nil {
 			panic(err)
 		}
@@ -41,7 +40,11 @@ func init() {
 	}
 
 	tsdb.RegisterIndex(IndexName, func(_ uint64, db, path string, _ *tsdb.SeriesIDSet, sfile *tsdb.SeriesFile, opt tsdb.EngineOptions) tsdb.Index {
-		idx := NewIndex(sfile, db, WithPath(path), WithMaximumLogFileSize(int64(opt.Config.MaxIndexLogFileSize)))
+		idx := NewIndex(sfile, db,
+			WithPath(path),
+			WithMaximumLogFileSize(int64(opt.Config.MaxIndexLogFileSize)),
+			WithSeriesIDCacheSize(opt.Config.SeriesIDSetCacheSize),
+		)
 		return idx
 	})
 }
@@ -110,12 +113,23 @@ var WithLogFileBufferSize = func(sz int) IndexOption {
 	}
 }
 
+// WithSeriesIDCacheSize sets the size of the series id set cache.
+// If set to 0, then the cache is disabled.
+var WithSeriesIDCacheSize = func(sz int) IndexOption {
+	return func(i *Index) {
+		i.tagValueCacheSize = sz
+	}
+}
+
 // Index represents a collection of layered index files and WAL.
 type Index struct {
 	mu         sync.RWMutex
 	partitions []*Partition
 	opened     bool
 
+	tagValueCache     *TagValueSeriesIDCache
+	tagValueCacheSize int
+
 	// The following may be set when initializing an Index.
 	path               string      // Root directory of the index partitions.
 	disableCompactions bool        // Initially disables compactions on the index.
@@ -146,22 +160,24 @@ func (i *Index) UniqueReferenceID() uintptr {
 // NewIndex returns a new instance of Index.
 func NewIndex(sfile *tsdb.SeriesFile, database string, options ...IndexOption) *Index {
 	idx := &Index{
-		maxLogFileSize: tsdb.DefaultMaxIndexLogFileSize,
-		logger:         zap.NewNop(),
-		version:        Version,
-		sfile:          sfile,
-		database:       database,
-		mSketch:        hll.NewDefaultPlus(),
-		mTSketch:       hll.NewDefaultPlus(),
-		sSketch:        hll.NewDefaultPlus(),
-		sTSketch:       hll.NewDefaultPlus(),
-		PartitionN:     DefaultPartitionN,
+		tagValueCacheSize: tsdb.DefaultSeriesIDSetCacheSize,
+		maxLogFileSize:    tsdb.DefaultMaxIndexLogFileSize,
+		logger:            zap.NewNop(),
+		version:           Version,
+		sfile:             sfile,
+		database:          database,
+		mSketch:           hll.NewDefaultPlus(),
+		mTSketch:          hll.NewDefaultPlus(),
+		sSketch:           hll.NewDefaultPlus(),
+		sTSketch:          hll.NewDefaultPlus(),
+		PartitionN:        DefaultPartitionN,
 	}
 
 	for _, option := range options {
 		option(idx)
 	}
 
+	idx.tagValueCache = NewTagValueSeriesIDCache(idx.tagValueCacheSize)
 	return idx
 }
 
@@ -367,7 +383,6 @@ func (i *Index) availableThreads() int {
 
 // updateMeasurementSketches rebuilds the cached measurement sketches.
 func (i *Index) updateMeasurementSketches() error {
-	i.mSketch, i.mTSketch = hll.NewDefaultPlus(), hll.NewDefaultPlus()
 	for j := 0; j < int(i.PartitionN); j++ {
 		if s, t, err := i.partitions[j].MeasurementsSketches(); err != nil {
 			return err
@@ -382,7 +397,6 @@ func (i *Index) updateMeasurementSketches() error {
 
 // updateSeriesSketches rebuilds the cached series sketches.
 func (i *Index) updateSeriesSketches() error {
-	i.sSketch, i.sTSketch = hll.NewDefaultPlus(), hll.NewDefaultPlus()
 	for j := 0; j < int(i.PartitionN); j++ {
 		if s, t, err := i.partitions[j].SeriesSketches(); err != nil {
 			return err
@@ -536,7 +550,7 @@ func (i *Index) fetchByteValues(fn func(idx int) ([][]byte, error)) ([][]byte, e
 	}
 
 	// It's now safe to read from names.
-	return slices.MergeSortedBytes(names[:]...), nil
+	return slices.MergeSortedBytes(names...), nil
 }
 
 // MeasurementIterator returns an iterator over all measurements.
@@ -649,7 +663,54 @@ func (i *Index) CreateSeriesListIfNotExists(keys [][]byte, names [][]byte, tagsS
 				if idx >= len(i.partitions) {
 					return // No more work.
 				}
-				errC <- i.partitions[idx].createSeriesListIfNotExists(pNames[idx], pTags[idx])
+
+				ids, err := i.partitions[idx].createSeriesListIfNotExists(pNames[idx], pTags[idx])
+
+				var updateCache bool
+				for _, id := range ids {
+					if id != 0 {
+						updateCache = true
+						break
+					}
+				}
+
+				if !updateCache {
+					errC <- err
+					continue
+				}
+
+				// Some cached bitset results may need to be updated.
+				i.tagValueCache.RLock()
+				for j, id := range ids {
+					if id == 0 {
+						continue
+					}
+
+					name := pNames[idx][j]
+					tags := pTags[idx][j]
+					if i.tagValueCache.measurementContainsSets(name) {
+						for _, pair := range tags {
+							// TODO(edd): It's not clear to me yet whether it will be better to take a lock
+							// on every series id set, or whether to gather them all up under the cache rlock
+							// and then take the cache lock and update them all at once (without invoking a lock
+							// on each series id set).
+							//
+							// Taking the cache lock will block all queries, but is one lock. Taking each series set
+							// lock might be many lock/unlocks but will only block a query that needs that particular set.
+							//
+							// Need to think on it, but I think taking a lock on each series id set is the way to go.
+							//
+							// One other option here is to take a lock on the series id set when we first encounter it
+							// and then keep it locked until we're done with all the ids.
+							//
+							// Note: this will only add `id` to the set if it exists.
+							i.tagValueCache.addToSet(name, pair.Key, pair.Value, id) // Takes a lock on the series id set
+						}
+					}
+				}
+				i.tagValueCache.RUnlock()
+
+				errC <- err
 			}
 		}()
 	}
@@ -677,7 +738,8 @@ func (i *Index) CreateSeriesListIfNotExists(keys [][]byte, names [][]byte, tagsS
 
 // CreateSeriesIfNotExists creates a series if it doesn't exist or is deleted.
 func (i *Index) CreateSeriesIfNotExists(key, name []byte, tags models.Tags) error {
-	if err := i.partition(key).createSeriesListIfNotExists([][]byte{name}, []models.Tags{tags}); err != nil {
+	ids, err := i.partition(key).createSeriesListIfNotExists([][]byte{name}, []models.Tags{tags})
+	if err != nil {
 		return err
 	}
 
@@ -685,6 +747,31 @@ func (i *Index) CreateSeriesIfNotExists(key, name []byte, tags models.Tags) erro
 	i.sSketch.Add(key)
 	i.mSketch.Add(name)
 	i.mu.Unlock()
+
+	if ids[0] == 0 {
+		return nil // No new series, nothing further to update.
+	}
+
+	// If there are cached sets for any of the tag pairs, they will need to be
+	// updated with the series id.
+	i.tagValueCache.RLock()
+	if i.tagValueCache.measurementContainsSets(name) {
+		for _, pair := range tags {
+			// TODO(edd): It's not clear to me yet whether it will be better to take a lock
+			// on every series id set, or whether to gather them all up under the cache rlock
+			// and then take the cache lock and update them all at once (without invoking a lock
+			// on each series id set).
+			//
+			// Taking the cache lock will block all queries, but is one lock. Taking each series set
+			// lock might be many lock/unlocks but will only block a query that needs that particular set.
+			//
+			// Need to think on it, but I think taking a lock on each series id set is the way to go.
+			//
+			// Note this will only add `id` to the set if it exists.
+			i.tagValueCache.addToSet(name, pair.Key, pair.Value, ids[0]) // Takes a lock on the series id set
+		}
+	}
+	i.tagValueCache.RUnlock()
 	return nil
 }
 
@@ -710,8 +797,18 @@ func (i *Index) DropSeries(seriesID uint64, key []byte, cascade bool) error {
 		return nil
 	}
 
-	// Extract measurement name.
-	name, _ := models.ParseKeyBytes(key)
+	// Extract measurement name & tags.
+	name, tags := models.ParseKeyBytes(key)
+
+	// If there are cached sets for any of the tag pairs, they will need to be
+	// updated with the series id.
+	i.tagValueCache.RLock()
+	if i.tagValueCache.measurementContainsSets(name) {
+		for _, pair := range tags {
+			i.tagValueCache.delete(name, pair.Key, pair.Value, seriesID) // Takes a lock on the series id set
+		}
+	}
+	i.tagValueCache.RUnlock()
 
 	// Check if that was the last series for the measurement in the entire index.
 	if ok, err := i.MeasurementHasSeries(name); err != nil {
@@ -727,21 +824,79 @@ func (i *Index) DropSeries(seriesID uint64, key []byte, cascade bool) error {
 	return nil
 }
 
+// DropSeries drops the provided series from the index.  If cascade is true
+// and this is the last series to the measurement, the measurment will also be dropped.
+func (i *Index) DropSeriesList(seriesIDs []uint64, keys [][]byte, _ bool) error {
+	// All slices must be of equal length.
+	if len(seriesIDs) != len(keys) {
+		return errors.New("seriesIDs/keys length mismatch in index")
+	}
+
+	// We need to move different series into collections for each partition
+	// to process.
+	pSeriesIDs := make([][]uint64, i.PartitionN)
+	pKeys := make([][][]byte, i.PartitionN)
+
+	for idx, key := range keys {
+		pidx := i.partitionIdx(key)
+		pSeriesIDs[pidx] = append(pSeriesIDs[pidx], seriesIDs[idx])
+		pKeys[pidx] = append(pKeys[pidx], key)
+	}
+
+	// Process each subset of series on each partition.
+	n := i.availableThreads()
+
+	// Store errors.
+	errC := make(chan error, i.PartitionN)
+
+	var pidx uint32 // Index of maximum Partition being worked on.
+	for k := 0; k < n; k++ {
+		go func() {
+			for {
+				idx := int(atomic.AddUint32(&pidx, 1) - 1) // Get next partition to work on.
+				if idx >= len(i.partitions) {
+					return // No more work.
+				}
+
+				// Drop from partition.
+				err := i.partitions[idx].DropSeriesList(pSeriesIDs[idx])
+				errC <- err
+			}
+		}()
+	}
+
+	// Check for error
+	for i := 0; i < cap(errC); i++ {
+		if err := <-errC; err != nil {
+			return err
+		}
+	}
+
+	// Add sketch tombstone.
+	i.mu.Lock()
+	for _, key := range keys {
+		i.sTSketch.Add(key)
+	}
+	i.mu.Unlock()
+
+	return nil
+}
+
 // DropSeriesGlobal is a no-op on the tsi1 index.
 func (i *Index) DropSeriesGlobal(key []byte) error { return nil }
 
 // DropMeasurementIfSeriesNotExist drops a measurement only if there are no more
 // series for the measurment.
-func (i *Index) DropMeasurementIfSeriesNotExist(name []byte) error {
+func (i *Index) DropMeasurementIfSeriesNotExist(name []byte) (bool, error) {
 	// Check if that was the last series for the measurement in the entire index.
 	if ok, err := i.MeasurementHasSeries(name); err != nil {
-		return err
+		return false, err
 	} else if ok {
-		return nil
+		return false, nil
 	}
 
 	// If no more series exist in the measurement then delete the measurement.
-	return i.DropMeasurement(name)
+	return true, i.DropMeasurement(name)
 }
 
 // MeasurementsSketches returns the two measurement sketches for the index.
@@ -885,8 +1040,10 @@ func (i *Index) TagValueIterator(name, key []byte) (tsdb.TagValueIterator, error
 func (i *Index) TagKeySeriesIDIterator(name, key []byte) (tsdb.SeriesIDIterator, error) {
 	a := make([]tsdb.SeriesIDIterator, 0, len(i.partitions))
 	for _, p := range i.partitions {
-		itr := p.TagKeySeriesIDIterator(name, key)
-		if itr != nil {
+		itr, err := p.TagKeySeriesIDIterator(name, key)
+		if err != nil {
+			return nil, err
+		} else if itr != nil {
 			a = append(a, itr)
 		}
 	}
@@ -895,14 +1052,35 @@ func (i *Index) TagKeySeriesIDIterator(name, key []byte) (tsdb.SeriesIDIterator,
 
 // TagValueSeriesIDIterator returns a series iterator for a single tag value.
 func (i *Index) TagValueSeriesIDIterator(name, key, value []byte) (tsdb.SeriesIDIterator, error) {
+	// Check series ID set cache...
+	if i.tagValueCacheSize > 0 {
+		if ss := i.tagValueCache.Get(name, key, value); ss != nil {
+			// Return a clone because the set is mutable.
+			return tsdb.NewSeriesIDSetIterator(ss.Clone()), nil
+		}
+	}
+
 	a := make([]tsdb.SeriesIDIterator, 0, len(i.partitions))
 	for _, p := range i.partitions {
-		itr := p.TagValueSeriesIDIterator(name, key, value)
-		if itr != nil {
+		itr, err := p.TagValueSeriesIDIterator(name, key, value)
+		if err != nil {
+			return nil, err
+		} else if itr != nil {
 			a = append(a, itr)
 		}
 	}
-	return tsdb.MergeSeriesIDIterators(a...), nil
+
+	itr := tsdb.MergeSeriesIDIterators(a...)
+	if i.tagValueCacheSize == 0 {
+		return itr, nil
+	}
+
+	// Check if the iterator contains only series id sets. Cache them...
+	if ssitr, ok := itr.(tsdb.SeriesIDSetIterator); ok {
+		ss := ssitr.SeriesIDSet()
+		i.tagValueCache.Put(name, key, value, ss)
+	}
+	return itr, nil
 }
 
 // MeasurementTagKeysByExpr extracts the tag keys wanted by the expression.
