@@ -12,6 +12,7 @@ import (
 	"regexp"
 	"runtime"
 	"sort"
+	"strconv"
 	"strings"
 	"sync"
 	"sync/atomic"
@@ -40,6 +41,7 @@ const (
 	statWritePointsErr     = "writePointsErr"
 	statWritePointsDropped = "writePointsDropped"
 	statWritePointsOK      = "writePointsOk"
+	statWriteValuesOK      = "writeValuesOk"
 	statWriteBytes         = "writeBytes"
 	statDiskBytes          = "diskBytes"
 )
@@ -148,6 +150,10 @@ type Shard struct {
 	logger     *zap.Logger
 
 	EnableOnOpen bool
+
+	// CompactionDisabled specifies the shard should not schedule compactions.
+	// This option is intended for offline tooling.
+	CompactionDisabled bool
 }
 
 // NewShard returns a new initialized Shard. walPath doesn't apply to the b1 type index
@@ -202,7 +208,7 @@ func (s *Shard) SetEnabled(enabled bool) {
 	s.mu.Lock()
 	// Prevent writes and queries
 	s.enabled = enabled
-	if s._engine != nil {
+	if s._engine != nil && !s.CompactionDisabled {
 		// Disable background compactions and snapshotting
 		s._engine.SetEnabled(enabled)
 	}
@@ -242,6 +248,7 @@ type ShardStatistics struct {
 	WritePointsErr     int64
 	WritePointsDropped int64
 	WritePointsOK      int64
+	WriteValuesOK      int64
 	BytesWritten       int64
 	DiskBytes          int64
 }
@@ -260,6 +267,13 @@ func (s *Shard) Statistics(tags map[string]string) []models.Statistic {
 	seriesN := engine.SeriesN()
 
 	tags = s.defaultTags.Merge(tags)
+
+	// Set the index type on the tags.  N.B this needs to be checked since it's
+	// only set when the shard is opened.
+	if indexType := s.IndexType(); indexType != "" {
+		tags["indexType"] = indexType
+	}
+
 	statistics := []models.Statistic{{
 		Name: "shard",
 		Tags: tags,
@@ -272,6 +286,7 @@ func (s *Shard) Statistics(tags map[string]string) []models.Statistic {
 			statWritePointsErr:     atomic.LoadInt64(&s.stats.WritePointsErr),
 			statWritePointsDropped: atomic.LoadInt64(&s.stats.WritePointsDropped),
 			statWritePointsOK:      atomic.LoadInt64(&s.stats.WritePointsOK),
+			statWriteValuesOK:      atomic.LoadInt64(&s.stats.WriteValuesOK),
 			statWriteBytes:         atomic.LoadInt64(&s.stats.BytesWritten),
 			statDiskBytes:          atomic.LoadInt64(&s.stats.DiskBytes),
 		},
@@ -305,12 +320,13 @@ func (s *Shard) Open() error {
 			return err
 		}
 
+		idx.WithLogger(s.baseLogger)
+
 		// Open index.
 		if err := idx.Open(); err != nil {
 			return err
 		}
 		s.index = idx
-		idx.WithLogger(s.baseLogger)
 
 		// Initialize underlying engine.
 		e, err := NewEngine(s.id, idx, s.path, s.walPath, s.sfile, s.options)
@@ -419,7 +435,9 @@ func (s *Shard) Index() (Index, error) {
 	return s.index, nil
 }
 
-func (s *Shard) seriesFile() (*SeriesFile, error) {
+// SeriesFile returns a reference the underlying series file. If return an error
+// if the series file is nil.
+func (s *Shard) SeriesFile() (*SeriesFile, error) {
 	s.mu.RLock()
 	defer s.mu.RUnlock()
 	if err := s.ready(); err != nil {
@@ -429,10 +447,10 @@ func (s *Shard) seriesFile() (*SeriesFile, error) {
 }
 
 // IsIdle return true if the shard is not receiving writes and is fully compacted.
-func (s *Shard) IsIdle() bool {
+func (s *Shard) IsIdle() (state bool, reason string) {
 	engine, err := s.Engine()
 	if err != nil {
-		return true
+		return true, ""
 	}
 	return engine.IsIdle()
 }
@@ -478,8 +496,27 @@ type FieldCreate struct {
 	Field       *Field
 }
 
-// WritePoints will write the raw data points and any new metadata to the index in the shard.
+// WritePoints() is a thin wrapper for WritePointsWithContext().
 func (s *Shard) WritePoints(points []models.Point) error {
+	return s.WritePointsWithContext(context.Background(), points)
+}
+
+type ConetextKey int
+
+const (
+	StatPointsWritten = ConetextKey(iota)
+	StatValuesWritten
+)
+
+// WritePointsWithContext() will write the raw data points and any new metadata
+// to the index in the shard.
+//
+// If a context key of type ConetextKey is passed in, WritePointsWithContext()
+// will store points written stats into the int64 pointer associated with
+// StatPointsWritten and the number of values written in the int64 pointer
+// stored in the StatValuesWritten context values.
+//
+func (s *Shard) WritePointsWithContext(ctx context.Context, points []models.Point) error {
 	s.mu.RLock()
 	defer s.mu.RUnlock()
 
@@ -507,15 +544,45 @@ func (s *Shard) WritePoints(points []models.Point) error {
 		return err
 	}
 
-	// Write to the engine.
-	if err := engine.WritePoints(points); err != nil {
-		atomic.AddInt64(&s.stats.WritePointsErr, int64(len(points)))
-		atomic.AddInt64(&s.stats.WriteReqErr, 1)
-		return fmt.Errorf("engine: %s", err)
+	// see if our engine is capable of WritePointsWithContext
+	type contextWriter interface {
+		WritePointsWithContext(context.Context, []models.Point) error
 	}
-	atomic.AddInt64(&s.stats.WritePointsOK, int64(len(points)))
+	switch eng := engine.(type) {
+	case contextWriter:
+		if err := eng.WritePointsWithContext(ctx, points); err != nil {
+			atomic.AddInt64(&s.stats.WritePointsErr, int64(len(points)))
+			atomic.AddInt64(&s.stats.WriteReqErr, 1)
+			return fmt.Errorf("engine: %s", err)
+		}
+	default:
+		// Write to the engine.
+		if err := engine.WritePoints(points); err != nil {
+			atomic.AddInt64(&s.stats.WritePointsErr, int64(len(points)))
+			atomic.AddInt64(&s.stats.WriteReqErr, 1)
+			return fmt.Errorf("engine: %s", err)
+		}
+	}
+
+	// increment the number OK write requests
 	atomic.AddInt64(&s.stats.WriteReqOK, 1)
 
+	// Increment the number of points written.  If was a StatPointsWritten
+	// request is sent to this function via a context, use the value that the
+	// engine reported.  otherwise, use the length of our points slice.
+	if npoints, ok := ctx.Value(StatPointsWritten).(*int64); ok {
+		// use engine counted points
+		atomic.AddInt64(&s.stats.WritePointsOK, *npoints)
+	} else {
+		// fallback to assuming that len(points) is accurate
+		atomic.AddInt64(&s.stats.WritePointsOK, int64(len(points)))
+	}
+
+	// Increment the number of values stored if available
+	if nvalues, ok := ctx.Value(StatValuesWritten).(*int64); ok {
+		atomic.AddInt64(&s.stats.WriteValuesOK, *nvalues)
+	}
+
 	return writeError
 }
 
@@ -533,6 +600,9 @@ func (s *Shard) validateSeriesAndFields(points []models.Point) ([]models.Point,
 	names := make([][]byte, len(points))
 	tagsSlice := make([]models.Tags, len(points))
 
+	// Check if keys should be unicode validated.
+	validateKeys := s.options.Config.ValidateKeys
+
 	var j int
 	for i, p := range points {
 		tags := p.Tags()
@@ -548,6 +618,15 @@ func (s *Shard) validateSeriesAndFields(points []models.Point) ([]models.Point,
 			continue
 		}
 
+		// Drop any series with invalid unicode characters in the key.
+		if validateKeys && !models.ValidKeyTokens(string(p.Name()), tags) {
+			dropped++
+			if reason == "" {
+				reason = fmt.Sprintf("key contains invalid unicode: \"%s\"", string(p.Key()))
+			}
+			continue
+		}
+
 		keys[j] = p.Key()
 		names[j] = p.Name()
 		tagsSlice[j] = tags
@@ -578,8 +657,6 @@ func (s *Shard) validateSeriesAndFields(points []models.Point) ([]models.Point,
 		}
 	}
 
-	// Create a MeasurementFields cache.
-	mfCache := make(map[string]*MeasurementFields, 16)
 	j = 0
 	for i, p := range points {
 		// Skip any points with only invalid fields.
@@ -607,13 +684,8 @@ func (s *Shard) validateSeriesAndFields(points []models.Point) ([]models.Point,
 			continue
 		}
 
-		// Grab the MeasurementFields checking the local cache to avoid lock contention.
 		name := p.Name()
-		mf := mfCache[string(name)]
-		if mf == nil {
-			mf = engine.MeasurementFields(name).Clone()
-			mfCache[string(name)] = mf
-		}
+		mf := engine.MeasurementFields(name)
 
 		// Check with the field validator.
 		if err := s.options.FieldValidator.Validate(mf, p); err != nil {
@@ -689,11 +761,7 @@ func (s *Shard) createFieldsAndMeasurements(fieldsToCreate []*FieldCreate) error
 		s.index.SetFieldName(f.Measurement, f.Field.Name)
 	}
 
-	if len(fieldsToCreate) > 0 {
-		return engine.MeasurementFieldSet().Save()
-	}
-
-	return nil
+	return engine.MeasurementFieldSet().Save()
 }
 
 // DeleteSeriesRange deletes all values from for seriesKeys between min and max (inclusive)
@@ -771,7 +839,7 @@ func (s *Shard) MeasurementTagKeysByExpr(name []byte, expr influxql.Expr) (map[s
 
 // MeasurementTagKeyValuesByExpr returns all the tag keys values for the
 // provided expression.
-func (s *Shard) MeasurementTagKeyValuesByExpr(auth query.Authorizer, name []byte, key []string, expr influxql.Expr, keysSorted bool) ([][]string, error) {
+func (s *Shard) MeasurementTagKeyValuesByExpr(auth query.FineAuthorizer, name []byte, key []string, expr influxql.Expr, keysSorted bool) ([][]string, error) {
 	index, err := s.Index()
 	if err != nil {
 		return nil, err
@@ -780,9 +848,17 @@ func (s *Shard) MeasurementTagKeyValuesByExpr(auth query.Authorizer, name []byte
 	return indexSet.MeasurementTagKeyValuesByExpr(auth, name, key, expr, keysSorted)
 }
 
+// MeasurementNamesByPredicate returns fields for a measurement filtered by an expression.
+func (s *Shard) MeasurementNamesByPredicate(expr influxql.Expr) ([][]byte, error) {
+	index, err := s.Index()
+	if err != nil {
+		return nil, err
+	}
+	indexSet := IndexSet{Indexes: []Index{index}, SeriesFile: s.sfile}
+	return indexSet.MeasurementNamesByPredicate(query.OpenAuthorizer, expr)
+}
+
 // MeasurementFields returns fields for a measurement.
-// TODO(edd): This method is currently only being called from tests; do we
-// really need it?
 func (s *Shard) MeasurementFields(name []byte) *MeasurementFields {
 	engine, err := s.Engine()
 	if err != nil {
@@ -1098,12 +1174,12 @@ func (s *Shard) Import(r io.Reader, basePath string) error {
 
 // CreateSnapshot will return a path to a temp directory
 // containing hard links to the underlying shard files.
-func (s *Shard) CreateSnapshot() (string, error) {
+func (s *Shard) CreateSnapshot(skipCacheOk bool) (string, error) {
 	engine, err := s.Engine()
 	if err != nil {
 		return "", err
 	}
-	return engine.CreateSnapshot()
+	return engine.CreateSnapshot(skipCacheOk)
 }
 
 // ForEachMeasurementName iterates over each measurement in the shard.
@@ -1124,19 +1200,20 @@ func (s *Shard) TagKeyCardinality(name, key []byte) int {
 }
 
 // Digest returns a digest of the shard.
-func (s *Shard) Digest() (io.ReadCloser, int64, error) {
+func (s *Shard) Digest() (io.ReadCloser, int64, error, string) {
 	engine, err := s.Engine()
 	if err != nil {
-		return nil, 0, err
+		return nil, 0, err, ""
 	}
 
 	// Make sure the shard is idle/cold. (No use creating a digest of a
 	// hot shard that is rapidly changing.)
-	if !engine.IsIdle() {
-		return nil, 0, ErrShardNotIdle
+	if isIdle, reason := engine.IsIdle(); !isIdle {
+		return nil, 0, ErrShardNotIdle, reason
 	}
 
-	return engine.Digest()
+	readCloser, size, err := engine.Digest()
+	return readCloser, size, err, ""
 }
 
 // engine safely (under an RLock) returns a reference to the shard's Engine, or
@@ -1237,6 +1314,38 @@ func (a Shards) FieldKeysByMeasurement(name []byte) []string {
 	return slices.MergeSortedStrings(all...)
 }
 
+// MeasurementNamesByPredicate returns the measurements that match the given predicate.
+func (a Shards) MeasurementNamesByPredicate(expr influxql.Expr) ([][]byte, error) {
+	if len(a) == 1 {
+		return a[0].MeasurementNamesByPredicate(expr)
+	}
+
+	all := make([][][]byte, len(a))
+	for i, shard := range a {
+		names, err := shard.MeasurementNamesByPredicate(expr)
+		if err != nil {
+			return nil, err
+		}
+		all[i] = names
+	}
+	return slices.MergeSortedBytes(all...), nil
+}
+
+// FieldKeysByPredicate returns the field keys for series that match
+// the given predicate.
+func (a Shards) FieldKeysByPredicate(expr influxql.Expr) (map[string][]string, error) {
+	names, err := a.MeasurementNamesByPredicate(expr)
+	if err != nil {
+		return nil, err
+	}
+
+	all := make(map[string][]string, len(names))
+	for _, name := range names {
+		all[string(name)] = a.FieldKeysByMeasurement(name)
+	}
+	return all, nil
+}
+
 func (a Shards) FieldDimensions(measurements []string) (fields map[string]influxql.DataType, dimensions map[string]struct{}, err error) {
 	fields = make(map[string]influxql.DataType)
 	dimensions = make(map[string]struct{})
@@ -1322,7 +1431,7 @@ func (a Shards) createSeriesIterator(ctx context.Context, opt query.IteratorOpti
 			idxs = append(idxs, idx)
 		}
 		if sfile == nil {
-			sfile, _ = sh.seriesFile()
+			sfile, _ = sh.SeriesFile()
 		}
 	}
 
@@ -1395,7 +1504,7 @@ func (a Shards) CreateSeriesCursor(ctx context.Context, req SeriesCursorRequest,
 			idxs = append(idxs, idx)
 		}
 		if sfile == nil {
-			sfile, _ = sh.seriesFile()
+			sfile, _ = sh.SeriesFile()
 		}
 	}
 
@@ -1446,22 +1555,23 @@ func (a Shards) ExpandSources(sources influxql.Sources) (influxql.Sources, error
 
 // MeasurementFields holds the fields of a measurement and their codec.
 type MeasurementFields struct {
-	mu sync.RWMutex
+	mu sync.Mutex
 
-	fields map[string]*Field
+	fields atomic.Value // map[string]*Field
 }
 
 // NewMeasurementFields returns an initialised *MeasurementFields value.
 func NewMeasurementFields() *MeasurementFields {
-	return &MeasurementFields{fields: make(map[string]*Field)}
+	fields := make(map[string]*Field)
+	mf := &MeasurementFields{}
+	mf.fields.Store(fields)
+	return mf
 }
 
 func (m *MeasurementFields) FieldKeys() []string {
-	m.mu.RLock()
-	defer m.mu.RUnlock()
-
-	a := make([]string, 0, len(m.fields))
-	for key := range m.fields {
+	fields := m.fields.Load().(map[string]*Field)
+	a := make([]string, 0, len(fields))
+	for key := range fields {
 		a = append(a, key)
 	}
 	sort.Strings(a)
@@ -1471,14 +1581,13 @@ func (m *MeasurementFields) FieldKeys() []string {
 // bytes estimates the memory footprint of this MeasurementFields, in bytes.
 func (m *MeasurementFields) bytes() int {
 	var b int
-	m.mu.RLock()
 	b += 24 // mu RWMutex is 24 bytes
-	b += int(unsafe.Sizeof(m.fields))
-	for k, v := range m.fields {
+	fields := m.fields.Load().(map[string]*Field)
+	b += int(unsafe.Sizeof(fields))
+	for k, v := range fields {
 		b += int(unsafe.Sizeof(k)) + len(k)
 		b += int(unsafe.Sizeof(v)+unsafe.Sizeof(*v)) + len(v.Name)
 	}
-	m.mu.RUnlock()
 	return b
 }
 
@@ -1486,53 +1595,52 @@ func (m *MeasurementFields) bytes() int {
 // Returns an error if 255 fields have already been created on the measurement or
 // the fields already exists with a different type.
 func (m *MeasurementFields) CreateFieldIfNotExists(name []byte, typ influxql.DataType) error {
-	m.mu.RLock()
+	fields := m.fields.Load().(map[string]*Field)
 
 	// Ignore if the field already exists.
-	if f := m.fields[string(name)]; f != nil {
+	if f := fields[string(name)]; f != nil {
 		if f.Type != typ {
-			m.mu.RUnlock()
 			return ErrFieldTypeConflict
 		}
-		m.mu.RUnlock()
 		return nil
 	}
-	m.mu.RUnlock()
 
 	m.mu.Lock()
 	defer m.mu.Unlock()
 
+	fields = m.fields.Load().(map[string]*Field)
 	// Re-check field and type under write lock.
-	if f := m.fields[string(name)]; f != nil {
+	if f := fields[string(name)]; f != nil {
 		if f.Type != typ {
 			return ErrFieldTypeConflict
 		}
 		return nil
 	}
 
+	fieldsUpdate := make(map[string]*Field, len(fields)+1)
+	for k, v := range fields {
+		fieldsUpdate[k] = v
+	}
 	// Create and append a new field.
 	f := &Field{
-		ID:   uint8(len(m.fields) + 1),
+		ID:   uint8(len(fields) + 1),
 		Name: string(name),
 		Type: typ,
 	}
-	m.fields[string(name)] = f
+	fieldsUpdate[string(name)] = f
+	m.fields.Store(fieldsUpdate)
 
 	return nil
 }
 
 func (m *MeasurementFields) FieldN() int {
-	m.mu.RLock()
-	n := len(m.fields)
-	m.mu.RUnlock()
+	n := len(m.fields.Load().(map[string]*Field))
 	return n
 }
 
 // Field returns the field for name, or nil if there is no field for name.
 func (m *MeasurementFields) Field(name string) *Field {
-	m.mu.RLock()
-	f := m.fields[name]
-	m.mu.RUnlock()
+	f := m.fields.Load().(map[string]*Field)[name]
 	return f
 }
 
@@ -1540,9 +1648,7 @@ func (m *MeasurementFields) HasField(name string) bool {
 	if m == nil {
 		return false
 	}
-	m.mu.RLock()
-	f := m.fields[name]
-	m.mu.RUnlock()
+	f := m.fields.Load().(map[string]*Field)[name]
 	return f != nil
 }
 
@@ -1551,62 +1657,48 @@ func (m *MeasurementFields) HasField(name string) bool {
 // it avoids a string allocation, which can't be avoided if the caller converts
 // the []byte to a string and calls Field.
 func (m *MeasurementFields) FieldBytes(name []byte) *Field {
-	m.mu.RLock()
-	f := m.fields[string(name)]
-	m.mu.RUnlock()
+	f := m.fields.Load().(map[string]*Field)[string(name)]
 	return f
 }
 
 // FieldSet returns the set of fields and their types for the measurement.
 func (m *MeasurementFields) FieldSet() map[string]influxql.DataType {
-	m.mu.RLock()
-	defer m.mu.RUnlock()
-
-	fields := make(map[string]influxql.DataType)
-	for name, f := range m.fields {
-		fields[name] = f.Type
+	fields := m.fields.Load().(map[string]*Field)
+	fieldTypes := make(map[string]influxql.DataType)
+	for name, f := range fields {
+		fieldTypes[name] = f.Type
 	}
-	return fields
+	return fieldTypes
 }
 
 func (m *MeasurementFields) ForEachField(fn func(name string, typ influxql.DataType) bool) {
-	m.mu.RLock()
-	defer m.mu.RUnlock()
-	for name, f := range m.fields {
+	fields := m.fields.Load().(map[string]*Field)
+	for name, f := range fields {
 		if !fn(name, f.Type) {
 			return
 		}
 	}
 }
 
-// Clone returns copy of the MeasurementFields
-func (m *MeasurementFields) Clone() *MeasurementFields {
-	m.mu.RLock()
-	defer m.mu.RUnlock()
-	fields := make(map[string]*Field, len(m.fields))
-	for key, field := range m.fields {
-		fields[key] = field
-	}
-	return &MeasurementFields{
-		fields: fields,
-	}
-}
-
 // MeasurementFieldSet represents a collection of fields by measurement.
 // This safe for concurrent use.
 type MeasurementFieldSet struct {
 	mu     sync.RWMutex
 	fields map[string]*MeasurementFields
-
 	// path is the location to persist field sets
 	path string
+	// ephemeral counters for updating the file on disk
+	memoryVersion  uint64
+	writtenVersion uint64
 }
 
 // NewMeasurementFieldSet returns a new instance of MeasurementFieldSet.
 func NewMeasurementFieldSet(path string) (*MeasurementFieldSet, error) {
 	fs := &MeasurementFieldSet{
-		fields: make(map[string]*MeasurementFields),
-		path:   path,
+		fields:         make(map[string]*MeasurementFields),
+		path:           path,
+		memoryVersion:  0,
+		writtenVersion: 0,
 	}
 
 	// If there is a load error, return the error and an empty set so
@@ -1691,21 +1783,41 @@ func (fs *MeasurementFieldSet) IsEmpty() bool {
 	return len(fs.fields) == 0
 }
 
-func (fs *MeasurementFieldSet) Save() error {
-	fs.mu.Lock()
-	defer fs.mu.Unlock()
+func (fs *MeasurementFieldSet) Save() (err error) {
+	// current version
+	var v uint64
+	// Is the MeasurementFieldSet empty?
+	isEmpty := false
+	// marshaled MeasurementFieldSet
 
-	return fs.saveNoLock()
-}
+	b, err := func() ([]byte, error) {
+		fs.mu.Lock()
+		defer fs.mu.Unlock()
+		fs.memoryVersion += 1
+		v = fs.memoryVersion
+		// If no fields left, remove the fields index file
+		if len(fs.fields) == 0 {
+			isEmpty = true
+			if err := os.RemoveAll(fs.path); err != nil {
+				return nil, err
+			} else {
+				fs.writtenVersion = fs.memoryVersion
+				return nil, nil
+			}
+		}
+		return fs.marshalMeasurementFieldSetNoLock()
+	}()
 
-func (fs *MeasurementFieldSet) saveNoLock() error {
-	// No fields left, remove the fields index file
-	if len(fs.fields) == 0 {
-		return os.RemoveAll(fs.path)
+	if err != nil {
+		return err
+	} else if isEmpty {
+		return nil
 	}
 
 	// Write the new index to a temp file and rename when it's sync'd
-	path := fs.path + ".tmp"
+	// if it is still the most recent memoryVersion of the MeasurementFields
+	path := fs.path + "." + strconv.FormatUint(v, 10) + ".tmp"
+
 	fd, err := os.OpenFile(path, os.O_CREATE|os.O_RDWR|os.O_EXCL|os.O_SYNC, 0666)
 	if err != nil {
 		return err
@@ -1716,28 +1828,6 @@ func (fs *MeasurementFieldSet) saveNoLock() error {
 		return err
 	}
 
-	pb := internal.MeasurementFieldSet{
-		Measurements: make([]*internal.MeasurementFields, 0, len(fs.fields)),
-	}
-	for name, mf := range fs.fields {
-		fs := &internal.MeasurementFields{
-			Name:   name,
-			Fields: make([]*internal.Field, 0, mf.FieldN()),
-		}
-
-		mf.ForEachField(func(field string, typ influxql.DataType) bool {
-			fs.Fields = append(fs.Fields, &internal.Field{Name: field, Type: int32(typ)})
-			return true
-		})
-
-		pb.Measurements = append(pb.Measurements, fs)
-	}
-
-	b, err := proto.Marshal(&pb)
-	if err != nil {
-		return err
-	}
-
 	if _, err := fd.Write(b); err != nil {
 		return err
 	}
@@ -1751,11 +1841,52 @@ func (fs *MeasurementFieldSet) saveNoLock() error {
 		return err
 	}
 
+	fs.mu.Lock()
+	defer fs.mu.Unlock()
+
+	// Check if a later modification and save of fields has superseded ours
+	// If so, we are successfully done! We were beaten by a later call
+	// to this function
+	if fs.writtenVersion > v {
+		return nil
+	}
+
 	if err := file.RenameFile(path, fs.path); err != nil {
 		return err
 	}
 
-	return file.SyncDir(filepath.Dir(fs.path))
+	if err = file.SyncDir(filepath.Dir(fs.path)); err != nil {
+		return err
+	}
+	// Update the written version to the current version
+	fs.writtenVersion = v
+	return nil
+}
+
+func (fs *MeasurementFieldSet) marshalMeasurementFieldSetNoLock() (marshalled []byte, err error) {
+	pb := internal.MeasurementFieldSet{
+		Measurements: make([]*internal.MeasurementFields, 0, len(fs.fields)),
+	}
+
+	for name, mf := range fs.fields {
+		imf := &internal.MeasurementFields{
+			Name:   []byte(name),
+			Fields: make([]*internal.Field, 0, mf.FieldN()),
+		}
+
+		mf.ForEachField(func(field string, typ influxql.DataType) bool {
+			imf.Fields = append(imf.Fields, &internal.Field{Name: []byte(field), Type: int32(typ)})
+			return true
+		})
+
+		pb.Measurements = append(pb.Measurements, imf)
+	}
+	b, err := proto.Marshal(&pb)
+	if err != nil {
+		return nil, err
+	} else {
+		return b, nil
+	}
 }
 
 func (fs *MeasurementFieldSet) load() error {
@@ -1791,18 +1922,18 @@ func (fs *MeasurementFieldSet) load() error {
 
 	fs.fields = make(map[string]*MeasurementFields, len(pb.GetMeasurements()))
 	for _, measurement := range pb.GetMeasurements() {
-		set := &MeasurementFields{
-			fields: make(map[string]*Field, len(measurement.GetFields())),
-		}
+		fields := make(map[string]*Field, len(measurement.GetFields()))
 		for _, field := range measurement.GetFields() {
-			set.fields[field.GetName()] = &Field{Name: field.GetName(), Type: influxql.DataType(field.GetType())}
+			fields[string(field.GetName())] = &Field{Name: string(field.GetName()), Type: influxql.DataType(field.GetType())}
 		}
-		fs.fields[measurement.GetName()] = set
+		set := &MeasurementFields{}
+		set.fields.Store(fields)
+		fs.fields[string(measurement.GetName())] = set
 	}
 	return nil
 }
 
-// Field represents a series field.
+// Field represents a series field. All of the fields must be hashable.
 type Field struct {
 	ID   uint8             `json:"id,omitempty"`
 	Name string            `json:"name,omitempty"`
