@@ -456,7 +456,7 @@ func TestShard_WritePoints_FieldConflictConcurrent(t *testing.T) {
 			}
 
 			_ = sh.WritePoints(points[:500])
-			if f, err := sh.CreateSnapshot(); err == nil {
+			if f, err := sh.CreateSnapshot(false); err == nil {
 				os.RemoveAll(f)
 			}
 
@@ -472,7 +472,7 @@ func TestShard_WritePoints_FieldConflictConcurrent(t *testing.T) {
 			}
 
 			_ = sh.WritePoints(points[500:])
-			if f, err := sh.CreateSnapshot(); err == nil {
+			if f, err := sh.CreateSnapshot(false); err == nil {
 				os.RemoveAll(f)
 			}
 		}
@@ -491,6 +491,7 @@ func TestShard_WritePoints_FieldConflictConcurrent(t *testing.T) {
 }
 
 func TestShard_WritePoints_FieldConflictConcurrentQuery(t *testing.T) {
+	t.Skip("https://github.com/influxdata/influxdb/issues/14267")
 	if testing.Short() {
 		t.Skip()
 	}
@@ -1695,6 +1696,71 @@ func TestMeasurementFieldSet_InvalidFormat(t *testing.T) {
 	}
 }
 
+func TestMeasurementFieldSet_ConcurrentSave(t *testing.T) {
+	var iterations int
+	dir, cleanup := MustTempDir()
+	defer cleanup()
+
+	if testing.Short() {
+		iterations = 50
+	} else {
+		iterations = 200
+	}
+
+	mt := []string{"cpu", "dpu", "epu", "fpu"}
+	ft := make([][]string, len(mt), len(mt))
+	for mi, m := range mt {
+		ft[mi] = make([]string, iterations, iterations)
+		for i := 0; i < iterations; i += 1 {
+			ft[mi][i] = fmt.Sprintf("%s_%s_%d", m, "value", i)
+		}
+	}
+
+	path := filepath.Join(dir, "fields.idx")
+	mfs, err := tsdb.NewMeasurementFieldSet(path)
+	if err != nil {
+		t.Fatalf("NewMeasurementFieldSet error: %v", err)
+	}
+	var wg sync.WaitGroup
+
+	wg.Add(len(ft))
+	for i, fs := range ft {
+		go testFieldMaker(t, &wg, mfs, mt[i], fs)
+	}
+	wg.Wait()
+
+	mfs2, err := tsdb.NewMeasurementFieldSet(path)
+	if err != nil {
+		t.Fatalf("NewMeasurementFieldSet error: %v", err)
+	}
+	for i, fs := range ft {
+		mf := mfs.Fields([]byte(mt[i]))
+		mf2 := mfs2.Fields([]byte(mt[i]))
+		for _, f := range fs {
+			if mf2.Field(f) == nil {
+				t.Fatalf("Created field not found on reloaded MeasurementFieldSet %s", f)
+			}
+			if mf.Field(f) == nil {
+				t.Fatalf("Created field not found in original MeasureMentFieldSet: %s", f)
+			}
+		}
+	}
+
+}
+
+func testFieldMaker(t *testing.T, wg *sync.WaitGroup, mf *tsdb.MeasurementFieldSet, measurement string, fieldNames []string) {
+	defer wg.Done()
+	fields := mf.CreateFieldsIfNotExists([]byte(measurement))
+	for _, fieldName := range fieldNames {
+		if err := fields.CreateFieldIfNotExists([]byte(fieldName), influxql.Float); err != nil {
+			t.Fatalf("create field error: %v", err)
+		}
+		if err := mf.Save(); err != nil {
+			t.Fatalf("save error: %v", err)
+		}
+	}
+}
+
 func BenchmarkWritePoints_NewSeries_1K(b *testing.B)   { benchmarkWritePoints(b, 38, 3, 3, 1) }
 func BenchmarkWritePoints_NewSeries_100K(b *testing.B) { benchmarkWritePoints(b, 32, 5, 5, 1) }
 func BenchmarkWritePoints_NewSeries_250K(b *testing.B) { benchmarkWritePoints(b, 80, 5, 5, 1) }
@@ -1747,12 +1813,18 @@ func BenchmarkWritePoints_NewSeries_5000_Measurement_1_TagKey_1_TagValue(b *test
 func BenchmarkWritePoints_NewSeries_10000_Measurement_1_TagKey_1_TagValue(b *testing.B) {
 	benchmarkWritePoints(b, 10000, 1, 1, 1)
 }
+
+func BenchmarkWritePoints_NewSeries_1000_Measurement_10_TagKey_1_TagValue(b *testing.B) {
+	benchmarkWritePoints(b, 1000, 10, 1, 1)
+}
+
 func BenchmarkWritePoints_NewSeries_50000_Measurement_1_TagKey_1_TagValue(b *testing.B) {
 	benchmarkWritePoints(b, 50000, 1, 1, 1)
 }
 func BenchmarkWritePoints_NewSeries_100000_Measurement_1_TagKey_1_TagValue(b *testing.B) {
 	benchmarkWritePoints(b, 100000, 1, 1, 1)
 }
+
 func BenchmarkWritePoints_NewSeries_500000_Measurement_1_TagKey_1_TagValue(b *testing.B) {
 	benchmarkWritePoints(b, 500000, 1, 1, 1)
 }
@@ -1815,6 +1887,7 @@ func BenchmarkWritePoints_ExistingSeries_1K(b *testing.B) {
 func BenchmarkWritePoints_ExistingSeries_100K(b *testing.B) {
 	benchmarkWritePointsExistingSeries(b, 32, 5, 5, 1)
 }
+
 func BenchmarkWritePoints_ExistingSeries_250K(b *testing.B) {
 	benchmarkWritePointsExistingSeries(b, 80, 5, 5, 1)
 }
@@ -1825,6 +1898,23 @@ func BenchmarkWritePoints_ExistingSeries_1M(b *testing.B) {
 	benchmarkWritePointsExistingSeries(b, 320, 5, 5, 1)
 }
 
+// The following two benchmarks measure time to write 10k points at a time for comparing performance with different measurement cardinalities.
+func BenchmarkWritePoints_ExistingSeries_100K_1_1(b *testing.B) {
+	benchmarkWritePointsExistingSeriesEqualBatches(b, 100000, 1, 1, 1)
+}
+
+func BenchmarkWritePoints_ExistingSeries_10K_10_1(b *testing.B) {
+	benchmarkWritePointsExistingSeriesEqualBatches(b, 10000, 10, 1, 1)
+}
+
+func BenchmarkWritePoints_ExistingSeries_100K_1_1_Fields(b *testing.B) {
+	benchmarkWritePointsExistingSeriesFields(b, 100000, 1, 1, 1)
+}
+
+func BenchmarkWritePoints_ExistingSeries_10K_10_1_Fields(b *testing.B) {
+	benchmarkWritePointsExistingSeriesFields(b, 10000, 10, 1, 1)
+}
+
 // benchmarkWritePoints benchmarks writing new series to a shard.
 // mCnt - measurement count
 // tkCnt - tag key count
@@ -1851,14 +1941,11 @@ func benchmarkWritePoints(b *testing.B, mCnt, tkCnt, tvCnt, pntCnt int) {
 
 	// Run the benchmark loop.
 	for n := 0; n < b.N; n++ {
-		tmpDir, _ := ioutil.TempDir("", "shard_test")
-		tmpShard := filepath.Join(tmpDir, "shard")
-		tmpWal := filepath.Join(tmpDir, "wal")
-		opts := tsdb.NewEngineOptions()
-		opts.Config.WALDir = tmpWal
-		opts.InmemIndex = inmem.NewIndex(filepath.Base(tmpDir), sfile.SeriesFile)
-		shard := tsdb.NewShard(1, tmpShard, tmpWal, sfile.SeriesFile, opts)
-		shard.Open()
+		shard, tmpDir, err := openShard(sfile)
+		if err != nil {
+			shard.Close()
+			b.Fatal(err)
+		}
 
 		b.StartTimer()
 		// Call the function being benchmarked.
@@ -1890,16 +1977,12 @@ func benchmarkWritePointsExistingSeries(b *testing.B, mCnt, tkCnt, tvCnt, pntCnt
 	sfile := MustOpenSeriesFile()
 	defer sfile.Close()
 
-	tmpDir, _ := ioutil.TempDir("", "")
-	defer os.RemoveAll(tmpDir)
-	tmpShard := filepath.Join(tmpDir, "shard")
-	tmpWal := filepath.Join(tmpDir, "wal")
-	opts := tsdb.NewEngineOptions()
-	opts.Config.WALDir = tmpWal
-	opts.InmemIndex = inmem.NewIndex(filepath.Base(tmpDir), sfile.SeriesFile)
-	shard := tsdb.NewShard(1, tmpShard, tmpWal, sfile.SeriesFile, opts)
-	shard.Open()
+	shard, tmpDir, err := openShard(sfile)
+	if err != nil {
+		b.Fatal(err)
+	}
 	defer shard.Close()
+
 	chunkedWrite(shard, points)
 
 	// Reset timers and mem-stats before the main benchmark loop.
@@ -1908,6 +1991,7 @@ func benchmarkWritePointsExistingSeries(b *testing.B, mCnt, tkCnt, tvCnt, pntCnt
 	// Run the benchmark loop.
 	for n := 0; n < b.N; n++ {
 		b.StopTimer()
+
 		for _, p := range points {
 			p.SetTime(p.Time().Add(time.Second))
 		}
@@ -1916,6 +2000,119 @@ func benchmarkWritePointsExistingSeries(b *testing.B, mCnt, tkCnt, tvCnt, pntCnt
 		// Call the function being benchmarked.
 		chunkedWrite(shard, points)
 	}
+	os.RemoveAll(tmpDir)
+}
+
+func benchmarkWritePointsExistingSeriesFields(b *testing.B, mCnt, tkCnt, tvCnt, pntCnt int) {
+	// Generate test series (measurements + unique tag sets).
+	series := genTestSeries(mCnt, tkCnt, tvCnt)
+	// Generate point data to write to the shard.
+	points := []models.Point{}
+	for _, s := range series {
+		i := 0
+		for val := 0.0; val < float64(pntCnt); val++ {
+			field := fmt.Sprintf("v%d", i%256)
+			p := models.MustNewPoint(s.Measurement, s.Tags, map[string]interface{}{field: val}, time.Now())
+			points = append(points, p)
+			i++
+		}
+	}
+
+	sfile := MustOpenSeriesFile()
+	defer sfile.Close()
+
+	shard, tmpDir, err := openShard(sfile)
+	if err != nil {
+		b.Fatal(err)
+	}
+	defer shard.Close()
+
+	chunkedWrite(shard, points)
+
+	// Reset timers and mem-stats before the main benchmark loop.
+	b.ResetTimer()
+
+	// Run the benchmark loop.
+	for n := 0; n < b.N; n++ {
+		b.StopTimer()
+
+		for _, p := range points {
+			p.SetTime(p.Time().Add(time.Second))
+		}
+
+		b.StartTimer()
+		// Call the function being benchmarked.
+		chunkedWrite(shard, points)
+	}
+	os.RemoveAll(tmpDir)
+}
+
+func benchmarkWritePointsExistingSeriesEqualBatches(b *testing.B, mCnt, tkCnt, tvCnt, pntCnt int) {
+	// Generate test series (measurements + unique tag sets).
+	series := genTestSeries(mCnt, tkCnt, tvCnt)
+	// Generate point data to write to the shard.
+	points := []models.Point{}
+	for _, s := range series {
+		for val := 0.0; val < float64(pntCnt); val++ {
+			p := models.MustNewPoint(s.Measurement, s.Tags, map[string]interface{}{"value": val}, time.Now())
+			points = append(points, p)
+		}
+	}
+
+	sfile := MustOpenSeriesFile()
+	defer sfile.Close()
+
+	shard, tmpDir, err := openShard(sfile)
+	if err != nil {
+		b.Fatal(err)
+	}
+	defer shard.Close()
+
+	chunkedWrite(shard, points)
+
+	// Reset timers and mem-stats before the main benchmark loop.
+	b.ResetTimer()
+
+	// Run the benchmark loop.
+	nPts := len(points)
+	chunkSz := 10000
+	start := 0
+	end := chunkSz
+	for n := 0; n < b.N; n++ {
+		b.StopTimer()
+
+		if end > nPts {
+			end = nPts
+		}
+		if end-start == 0 {
+			start = 0
+			end = chunkSz
+		}
+
+		for _, p := range points[start:end] {
+			p.SetTime(p.Time().Add(time.Second))
+		}
+
+		b.StartTimer()
+		shard.WritePoints(points[start:end])
+		b.StopTimer()
+
+		start = end
+		end += chunkSz
+	}
+	os.RemoveAll(tmpDir)
+}
+
+func openShard(sfile *SeriesFile) (*tsdb.Shard, string, error) {
+	tmpDir, _ := ioutil.TempDir("", "shard_test")
+	tmpShard := filepath.Join(tmpDir, "shard")
+	tmpWal := filepath.Join(tmpDir, "wal")
+	opts := tsdb.NewEngineOptions()
+	opts.Config.WALDir = tmpWal
+	opts.InmemIndex = inmem.NewIndex(filepath.Base(tmpDir), sfile.SeriesFile)
+	shard := tsdb.NewShard(1, tmpShard, tmpWal, sfile.SeriesFile, opts)
+	err := shard.Open()
+	return shard, tmpDir, err
 }
 
 func BenchmarkCreateIterator(b *testing.B) {
@@ -2060,7 +2257,7 @@ func NewShards(index string, n int) Shards {
 		opt := tsdb.NewEngineOptions()
 		opt.IndexVersion = index
 		opt.Config.WALDir = filepath.Join(dir, "wal")
-		if index == "inmem" {
+		if index == tsdb.InmemIndexName {
 			opt.InmemIndex = inmem.NewIndex(filepath.Base(dir), sfile.SeriesFile)
 		}
 
