@@ -2,6 +2,7 @@ package tsdb // import "github.com/influxdata/influxdb/tsdb"
 
 import (
 	"bytes"
+	"context"
 	"errors"
 	"fmt"
 	"io"
@@ -33,6 +34,9 @@ var (
 	ErrStoreClosed = fmt.Errorf("store is closed")
 	// ErrShardDeletion is returned when trying to create a shard that is being deleted
 	ErrShardDeletion = errors.New("shard is being deleted")
+	// ErrMultipleIndexTypes is returned when trying to do deletes on a database with
+	// multiple index types.
+	ErrMultipleIndexTypes = errors.New("cannot delete data. DB contains shards using both inmem and tsi1 indexes. Please convert all shards to use the same index type to delete data.")
 )
 
 // Statistics gathered by the store.
@@ -45,11 +49,35 @@ const (
 // a database.
 const SeriesFileDirectory = "_series"
 
+// databaseState keeps track of the state of a database.
+type databaseState struct{ indexTypes map[string]int }
+
+// addIndexType records that the database has a shard with the given index type.
+func (d *databaseState) addIndexType(indexType string) {
+	if d.indexTypes == nil {
+		d.indexTypes = make(map[string]int)
+	}
+	d.indexTypes[indexType]++
+}
+
+// addIndexType records that the database no longer has a shard with the given index type.
+func (d *databaseState) removeIndexType(indexType string) {
+	if d.indexTypes != nil {
+		d.indexTypes[indexType]--
+		if d.indexTypes[indexType] <= 0 {
+			delete(d.indexTypes, indexType)
+		}
+	}
+}
+
+// hasMultipleIndexTypes returns true if the database has multiple index types.
+func (d *databaseState) hasMultipleIndexTypes() bool { return d != nil && len(d.indexTypes) > 1 }
+
 // Store manages shards and indexes for databases.
 type Store struct {
 	mu                sync.RWMutex
 	shards            map[uint64]*Shard
-	databases         map[string]struct{}
+	databases         map[string]*databaseState
 	sfiles            map[string]*SeriesFile
 	SeriesFileMaxSize int64 // Determines size of series file mmap. Can be altered in tests.
 	path              string
@@ -61,6 +89,10 @@ type Store struct {
 	// This prevents new shards from being created while old ones are being deleted.
 	pendingShardDeletes map[uint64]struct{}
 
+	// Epoch tracker helps serialize writes and deletes that may conflict. It
+	// is stored by shard.
+	epochs map[uint64]*epochTracker
+
 	EngineOptions EngineOptions
 
 	baseLogger *zap.Logger
@@ -76,11 +108,12 @@ type Store struct {
 func NewStore(path string) *Store {
 	logger := zap.NewNop()
 	return &Store{
-		databases:           make(map[string]struct{}),
+		databases:           make(map[string]*databaseState),
 		path:                path,
 		sfiles:              make(map[string]*SeriesFile),
 		indexes:             make(map[string]interface{}),
 		pendingShardDeletes: make(map[uint64]struct{}),
+		epochs:              make(map[uint64]*epochTracker),
 		EngineOptions:       NewEngineOptions(),
 		Logger:              logger,
 		baseLogger:          logger,
@@ -101,19 +134,18 @@ func (s *Store) Statistics(tags map[string]string) []models.Statistic {
 	s.mu.RLock()
 	shards := s.shardsSlice()
 	s.mu.RUnlock()
-
 	// Add all the series and measurements cardinality estimations.
 	databases := s.Databases()
 	statistics := make([]models.Statistic, 0, len(databases))
 	for _, database := range databases {
 		log := s.Logger.With(logger.Database(database))
-		sc, err := s.SeriesCardinality(database)
+		sc, err := s.SeriesCardinality(context.Background(), database)
 		if err != nil {
 			log.Info("Cannot retrieve series cardinality", zap.Error(err))
 			continue
 		}
 
-		mc, err := s.MeasurementsCardinality(database)
+		mc, err := s.MeasurementsCardinality(context.Background(), database)
 		if err != nil {
 			log.Info("Cannot retrieve measurement cardinality", zap.Error(err))
 			continue
@@ -191,10 +223,13 @@ func (s *Store) Open() error {
 	}
 
 	s.opened = true
-	s.wg.Add(1)
 
 	if !s.EngineOptions.MonitorDisabled {
-		go s.monitorShards()
+		s.wg.Add(1)
+		go func() {
+			s.wg.Done()
+			s.monitorShards()
+		}()
 	}
 
 	return nil
@@ -227,15 +262,28 @@ func (s *Store) loadShards() error {
 
 	s.EngineOptions.CompactionLimiter = limiter.NewFixed(lim)
 
-	// Env var to disable throughput limiter.  This will be moved to a config option in 1.5.
 	compactionSettings := []zapcore.Field{zap.Int("max_concurrent_compactions", lim)}
-	if os.Getenv("INFLUXDB_DATA_COMPACTION_THROUGHPUT") == "" {
-		rate, burst := 48*1024*1024, 48*1024*1024
-		s.EngineOptions.CompactionThroughputLimiter = limiter.NewRate(48*1024*1024, 48*1024*1024)
-		compactionSettings = append(compactionSettings, zap.Int("throughput_bytes_per_second", rate), zap.Int("throughput_burst_bytes", burst))
+	throughput := int(s.EngineOptions.Config.CompactThroughput)
+	throughputBurst := int(s.EngineOptions.Config.CompactThroughputBurst)
+	if throughput > 0 {
+		if throughputBurst < throughput {
+			throughputBurst = throughput
+		}
+
+		compactionSettings = append(
+			compactionSettings,
+			zap.Int("throughput_bytes_per_second", throughput),
+			zap.Int("throughput_bytes_per_second_burst", throughputBurst),
+		)
+		s.EngineOptions.CompactionThroughputLimiter = limiter.NewRate(throughput, throughputBurst)
 	} else {
-		compactionSettings = append(compactionSettings, zap.String("throughput_bytes_per_second", "unlimited"), zap.String("throughput_burst", "unlimited"))
+		compactionSettings = append(
+			compactionSettings,
+			zap.String("throughput_bytes_per_second", "unlimited"),
+			zap.String("throughput_bytes_per_second_burst", "unlimited"),
+		)
 	}
+
 	s.Logger.Info("Compaction settings", compactionSettings...)
 
 	log, logEnd := logger.NewOperation(s.Logger, "Open store", "tsdb_open")
@@ -304,6 +352,12 @@ func (s *Store) loadShards() error {
 			}
 
 			for _, sh := range shardDirs {
+				// Series file should not be in a retention policy but skip just in case.
+				if sh.Name() == SeriesFileDirectory {
+					log.Warn("Skipping series file in retention policy dir", zap.String("path", filepath.Join(s.path, db.Name(), rp.Name())))
+					continue
+				}
+
 				n++
 				go func(db, rp, sh string) {
 					t.Take()
@@ -344,6 +398,7 @@ func (s *Store) loadShards() error {
 
 					// Disable compactions, writes and queries until all shards are loaded
 					shard.EnableOnOpen = false
+					shard.CompactionDisabled = s.EngineOptions.CompactionDisabled
 					shard.WithLogger(s.baseLogger)
 
 					err = shard.Open()
@@ -360,10 +415,6 @@ func (s *Store) loadShards() error {
 		}
 	}
 
-	// indexVersions tracks counts of the number of different types of index
-	// being used within each database.
-	indexVersions := make(map[string]map[string]int)
-
 	// Gather results of opening shards concurrently, keeping track of how
 	// many databases we are managing.
 	for i := 0; i < n; i++ {
@@ -372,20 +423,19 @@ func (s *Store) loadShards() error {
 			continue
 		}
 		s.shards[res.s.id] = res.s
-		s.databases[res.s.database] = struct{}{}
-
-		if _, ok := indexVersions[res.s.database]; !ok {
-			indexVersions[res.s.database] = make(map[string]int, 2)
+		s.epochs[res.s.id] = newEpochTracker()
+		if _, ok := s.databases[res.s.database]; !ok {
+			s.databases[res.s.database] = new(databaseState)
 		}
-		indexVersions[res.s.database][res.s.IndexType()]++
+		s.databases[res.s.database].addIndexType(res.s.IndexType())
 	}
 	close(resC)
 
 	// Check if any databases are running multiple index types.
-	for db, idxVersions := range indexVersions {
-		if len(idxVersions) > 1 {
+	for db, state := range s.databases {
+		if state.hasMultipleIndexTypes() {
 			var fields []zapcore.Field
-			for idx, cnt := range idxVersions {
+			for idx, cnt := range state.indexTypes {
 				fields = append(fields, zap.Int(fmt.Sprintf("%s_count", idx), cnt))
 			}
 			s.Logger.Warn("Mixed shard index types", append(fields, logger.Database(db))...)
@@ -395,7 +445,7 @@ func (s *Store) loadShards() error {
 	// Enable all shards
 	for _, sh := range s.shards {
 		sh.SetEnabled(true)
-		if sh.IsIdle() {
+		if isIdle, _ := sh.IsIdle(); isIdle {
 			if err := sh.Free(); err != nil {
 				return err
 			}
@@ -433,7 +483,7 @@ func (s *Store) Close() error {
 		}
 	}
 
-	s.databases = make(map[string]struct{})
+	s.databases = make(map[string]*databaseState)
 	s.sfiles = map[string]*SeriesFile{}
 	s.indexes = make(map[string]interface{})
 	s.pendingShardDeletes = make(map[uint64]struct{})
@@ -443,6 +493,16 @@ func (s *Store) Close() error {
 	return nil
 }
 
+// epochsForShards returns a copy of the epoch trackers only including what is necessary
+// for the provided shards. Must be called under the lock.
+func (s *Store) epochsForShards(shards []*Shard) map[uint64]*epochTracker {
+	out := make(map[uint64]*epochTracker)
+	for _, sh := range shards {
+		out[sh.id] = s.epochs[sh.id]
+	}
+	return out
+}
+
 // openSeriesFile either returns or creates a series file for the provided
 // database. It must be called under a full lock.
 func (s *Store) openSeriesFile(database string) (*SeriesFile, error) {
@@ -451,6 +511,7 @@ func (s *Store) openSeriesFile(database string) (*SeriesFile, error) {
 	}
 
 	sfile := NewSeriesFile(filepath.Join(s.path, database, SeriesFileDirectory))
+	sfile.WithMaxCompactionConcurrency(s.EngineOptions.Config.SeriesFileMaxConcurrentSnapshotCompactions)
 	sfile.Logger = s.baseLogger
 	if err := sfile.Open(); err != nil {
 		return nil, err
@@ -532,7 +593,8 @@ func (s *Store) ShardDigest(id uint64) (io.ReadCloser, int64, error) {
 		return nil, 0, ErrShardNotFound
 	}
 
-	return sh.Digest()
+	readCloser, size, err, _ := sh.Digest()
+	return readCloser, size, err
 }
 
 // CreateShard creates a shard with the given id and retention policy on a database.
@@ -595,20 +657,31 @@ func (s *Store) CreateShard(database, retentionPolicy string, shardID uint64, en
 	}
 
 	s.shards[shardID] = shard
-	s.databases[database] = struct{}{} // Ensure we are tracking any new db.
+	s.epochs[shardID] = newEpochTracker()
+	if _, ok := s.databases[database]; !ok {
+		s.databases[database] = new(databaseState)
+	}
+	s.databases[database].addIndexType(shard.IndexType())
+	if state := s.databases[database]; state.hasMultipleIndexTypes() {
+		var fields []zapcore.Field
+		for idx, cnt := range state.indexTypes {
+			fields = append(fields, zap.Int(fmt.Sprintf("%s_count", idx), cnt))
+		}
+		s.Logger.Warn("Mixed shard index types", append(fields, logger.Database(database))...)
+	}
 
 	return nil
 }
 
 // CreateShardSnapShot will create a hard link to the underlying shard and return a path.
 // The caller is responsible for cleaning up (removing) the file path returned.
-func (s *Store) CreateShardSnapshot(id uint64) (string, error) {
+func (s *Store) CreateShardSnapshot(id uint64, skipCacheOk bool) (string, error) {
 	sh := s.Shard(id)
 	if sh == nil {
 		return "", ErrShardNotFound
 	}
 
-	return sh.CreateSnapshot()
+	return sh.CreateSnapshot(skipCacheOk)
 }
 
 // SetShardEnabled enables or disables a shard for read and writes.
@@ -643,13 +716,21 @@ func (s *Store) DeleteShard(shardID uint64) error {
 	}
 	delete(s.shards, shardID)
 	s.pendingShardDeletes[shardID] = struct{}{}
+
+	db := sh.Database()
+	// Determine if the shard contained any series that are not present in any
+	// other shards in the database.
+	shards := s.filterShards(byDatabase(db))
+	epoch := s.epochs[shardID]
 	s.mu.Unlock()
 
 	// Ensure the pending deletion flag is cleared on exit.
 	defer func() {
 		s.mu.Lock()
 		defer s.mu.Unlock()
+		delete(s.epochs, shardID)
 		delete(s.pendingShardDeletes, shardID)
+		s.databases[db].removeIndexType(sh.IndexType())
 	}()
 
 	// Get the shard's local bitset of series IDs.
@@ -660,12 +741,6 @@ func (s *Store) DeleteShard(shardID uint64) error {
 
 	ss := index.SeriesIDSet()
 
-	db := sh.Database()
-
-	// Determine if the shard contained any series that are not present in any
-	// other shards in the database.
-	shards := s.filterShards(byDatabase(db))
-
 	s.walkShards(shards, func(sh *Shard) error {
 		index, err := sh.Index()
 		if err != nil {
@@ -714,6 +789,15 @@ func (s *Store) DeleteShard(shardID uint64) error {
 
 	}
 
+	// enter the epoch tracker
+	guards, gen := epoch.StartWrite()
+	defer epoch.EndWrite(gen)
+
+	// wait for any guards before closing the shard
+	for _, guard := range guards {
+		guard.Wait()
+	}
+
 	// Close the shard.
 	if err := sh.Close(); err != nil {
 		return err
@@ -735,9 +819,8 @@ func (s *Store) DeleteDatabase(name string) error {
 		// no files locally, so nothing to do
 		return nil
 	}
-	shards := s.filterShards(func(sh *Shard) bool {
-		return sh.database == name
-	})
+	shards := s.filterShards(byDatabase(name))
+	epochs := s.epochsForShards(shards)
 	s.mu.RUnlock()
 
 	if err := s.walkShards(shards, func(sh *Shard) error {
@@ -745,6 +828,16 @@ func (s *Store) DeleteDatabase(name string) error {
 			return nil
 		}
 
+		epoch := epochs[sh.id]
+		// enter the epoch tracker
+		guards, gen := epoch.StartWrite()
+		defer epoch.EndWrite(gen)
+
+		// wait for any guards before closing the shard
+		for _, guard := range guards {
+			guard.Wait()
+		}
+
 		return sh.Close()
 	}); err != nil {
 		return err
@@ -780,6 +873,7 @@ func (s *Store) DeleteDatabase(name string) error {
 
 	for _, sh := range shards {
 		delete(s.shards, sh.id)
+		delete(s.epochs, sh.id)
 	}
 
 	// Remove database from store list of databases
@@ -837,8 +931,10 @@ func (s *Store) DeleteRetentionPolicy(database, name string) error {
 	}
 
 	s.mu.Lock()
+	state := s.databases[database]
 	for _, sh := range shards {
 		delete(s.shards, sh.id)
+		state.removeIndexType(sh.IndexType())
 	}
 	s.mu.Unlock()
 	return nil
@@ -847,7 +943,12 @@ func (s *Store) DeleteRetentionPolicy(database, name string) error {
 // DeleteMeasurement removes a measurement and all associated series from a database.
 func (s *Store) DeleteMeasurement(database, name string) error {
 	s.mu.RLock()
+	if s.databases[database].hasMultipleIndexTypes() {
+		s.mu.RUnlock()
+		return ErrMultipleIndexTypes
+	}
 	shards := s.filterShards(byDatabase(database))
+	epochs := s.epochsForShards(shards)
 	s.mu.RUnlock()
 
 	// Limit to 1 delete for each shard since expanding the measurement into the list
@@ -857,12 +958,20 @@ func (s *Store) DeleteMeasurement(database, name string) error {
 		limit.Take()
 		defer limit.Release()
 
+		// install our guard and wait for any prior deletes to finish. the
+		// guard ensures future deletes that could conflict wait for us.
+		guard := newGuard(influxql.MinTime, influxql.MaxTime, []string{name}, nil)
+		waiter := epochs[sh.id].WaitDelete(guard)
+		waiter.Wait()
+		defer waiter.Done()
+
 		return sh.DeleteMeasurement([]byte(name))
 	})
 }
 
 // filterShards returns a slice of shards where fn returns true
 // for the shard. If the provided predicate is nil then all shards are returned.
+// filterShards should be called under a lock.
 func (s *Store) filterShards(fn func(sh *Shard) bool) []*Shard {
 	var shards []*Shard
 	if fn == nil {
@@ -955,7 +1064,7 @@ func (s *Store) Databases() []string {
 	defer s.mu.RUnlock()
 
 	databases := make([]string, 0, len(s.databases))
-	for k, _ := range s.databases {
+	for k := range s.databases {
 		databases = append(databases, k)
 	}
 	return databases
@@ -1022,7 +1131,7 @@ func (s *Store) sketchesForDatabase(dbName string, getSketches func(*Shard) (est
 // Cardinality is calculated exactly by unioning all shards' bitsets of series
 // IDs. The result of this method cannot be combined with any other results.
 //
-func (s *Store) SeriesCardinality(database string) (int64, error) {
+func (s *Store) SeriesCardinality(ctx context.Context, database string) (int64, error) {
 	s.mu.RLock()
 	shards := s.filterShards(byDatabase(database))
 	s.mu.RUnlock()
@@ -1030,23 +1139,35 @@ func (s *Store) SeriesCardinality(database string) (int64, error) {
 	var setMu sync.Mutex
 	others := make([]*SeriesIDSet, 0, len(shards))
 
-	s.walkShards(shards, func(sh *Shard) error {
-		index, err := sh.Index()
-		if err != nil {
-			return err
+	err := s.walkShards(shards, func(sh *Shard) error {
+		select {
+		case <-ctx.Done():
+			return ctx.Err()
+		default:
+			index, err := sh.Index()
+			if err != nil {
+				return err
+			}
+
+			seriesIDs := index.SeriesIDSet()
+			setMu.Lock()
+			others = append(others, seriesIDs)
+			setMu.Unlock()
+
+			return nil
 		}
-
-		seriesIDs := index.SeriesIDSet()
-		setMu.Lock()
-		others = append(others, seriesIDs)
-		setMu.Unlock()
-
-		return nil
 	})
-
+	if err != nil {
+		return 0, err
+	}
 	ss := NewSeriesIDSet()
 	ss.Merge(others...)
-	return int64(ss.Cardinality()), nil
+	select {
+	case <-ctx.Done():
+		return 0, ctx.Err()
+	default:
+		return int64(ss.Cardinality()), nil
+	}
 }
 
 // SeriesSketches returns the sketches associated with the series data in all
@@ -1054,8 +1175,13 @@ func (s *Store) SeriesCardinality(database string) (int64, error) {
 //
 // The returned sketches can be combined with other sketches to provide an
 // estimation across distributed databases.
-func (s *Store) SeriesSketches(database string) (estimator.Sketch, estimator.Sketch, error) {
+func (s *Store) SeriesSketches(ctx context.Context, database string) (estimator.Sketch, estimator.Sketch, error) {
 	return s.sketchesForDatabase(database, func(sh *Shard) (estimator.Sketch, estimator.Sketch, error) {
+		select {
+		case <-ctx.Done():
+			return nil, nil, ctx.Err()
+		default:
+		}
 		if sh == nil {
 			return nil, nil, errors.New("shard nil, can't get cardinality")
 		}
@@ -1068,13 +1194,8 @@ func (s *Store) SeriesSketches(database string) (estimator.Sketch, estimator.Ske
 //
 // Cardinality is calculated using a sketch-based estimation. The result of this
 // method cannot be combined with any other results.
-func (s *Store) MeasurementsCardinality(database string) (int64, error) {
-	ss, ts, err := s.sketchesForDatabase(database, func(sh *Shard) (estimator.Sketch, estimator.Sketch, error) {
-		if sh == nil {
-			return nil, nil, errors.New("shard nil, can't get cardinality")
-		}
-		return sh.MeasurementsSketches()
-	})
+func (s *Store) MeasurementsCardinality(ctx context.Context, database string) (int64, error) {
+	ss, ts, err := s.MeasurementsSketches(ctx, database)
 
 	if err != nil {
 		return 0, err
@@ -1087,12 +1208,18 @@ func (s *Store) MeasurementsCardinality(database string) (int64, error) {
 //
 // The returned sketches can be combined with other sketches to provide an
 // estimation across distributed databases.
-func (s *Store) MeasurementsSketches(database string) (estimator.Sketch, estimator.Sketch, error) {
+func (s *Store) MeasurementsSketches(ctx context.Context, database string) (estimator.Sketch, estimator.Sketch, error) {
 	return s.sketchesForDatabase(database, func(sh *Shard) (estimator.Sketch, estimator.Sketch, error) {
-		if sh == nil {
-			return nil, nil, errors.New("shard nil, can't get cardinality")
+		// every iteration, check for timeout.
+		select {
+		case <-ctx.Done():
+			return nil, nil, ctx.Err()
+		default:
+			if sh == nil {
+				return nil, nil, errors.New("shard nil, can't get cardinality")
+			}
+			return sh.MeasurementsSketches()
 		}
-		return sh.MeasurementsSketches()
 	})
 }
 
@@ -1201,6 +1328,10 @@ func (s *Store) DeleteSeries(database string, sources []influxql.Source, conditi
 	}
 
 	s.mu.RLock()
+	if s.databases[database].hasMultipleIndexTypes() {
+		s.mu.RUnlock()
+		return ErrMultipleIndexTypes
+	}
 	sfile := s.sfiles[database]
 	if sfile == nil {
 		s.mu.RUnlock()
@@ -1208,6 +1339,7 @@ func (s *Store) DeleteSeries(database string, sources []influxql.Source, conditi
 		return nil
 	}
 	shards := s.filterShards(byDatabase(database))
+	epochs := s.epochsForShards(shards)
 	s.mu.RUnlock()
 
 	// Limit to 1 delete for each shard since expanding the measurement into the list
@@ -1235,6 +1367,12 @@ func (s *Store) DeleteSeries(database string, sources []influxql.Source, conditi
 		limit.Take()
 		defer limit.Release()
 
+		// install our guard and wait for any prior deletes to finish. the
+		// guard ensures future deletes that could conflict wait for us.
+		waiter := epochs[sh.id].WaitDelete(newGuard(min, max, names, condition))
+		waiter.Wait()
+		defer waiter.Done()
+
 		index, err := sh.Index()
 		if err != nil {
 			return err
@@ -1272,6 +1410,10 @@ func (s *Store) ExpandSources(sources influxql.Sources) (influxql.Sources, error
 
 // WriteToShard writes a list of points to a shard identified by its ID.
 func (s *Store) WriteToShard(shardID uint64, points []models.Point) error {
+	return s.WriteToShardWithContext(context.Background(), shardID, points)
+}
+
+func (s *Store) WriteToShardWithContext(ctx context.Context, shardID uint64, points []models.Point) error {
 	s.mu.RLock()
 
 	select {
@@ -1286,21 +1428,35 @@ func (s *Store) WriteToShard(shardID uint64, points []models.Point) error {
 		s.mu.RUnlock()
 		return ErrShardNotFound
 	}
+
+	epoch := s.epochs[shardID]
+
 	s.mu.RUnlock()
 
+	// enter the epoch tracker
+	guards, gen := epoch.StartWrite()
+	defer epoch.EndWrite(gen)
+
+	// wait for any guards before writing the points.
+	for _, guard := range guards {
+		if guard.Matches(points) {
+			guard.Wait()
+		}
+	}
+
 	// Ensure snapshot compactions are enabled since the shard might have been cold
 	// and disabled by the monitor.
-	if sh.IsIdle() {
+	if isIdle, _ := sh.IsIdle(); isIdle {
 		sh.SetCompactionsEnabled(true)
 	}
 
-	return sh.WritePoints(points)
+	return sh.WritePointsWithContext(ctx, points)
 }
 
 // MeasurementNames returns a slice of all measurements. Measurements accepts an
 // optional condition expression. If cond is nil, then all measurements for the
 // database will be returned.
-func (s *Store) MeasurementNames(auth query.Authorizer, database string, cond influxql.Expr) ([][]byte, error) {
+func (s *Store) MeasurementNames(ctx context.Context, auth query.FineAuthorizer, database string, cond influxql.Expr) ([][]byte, error) {
 	s.mu.RLock()
 	shards := s.filterShards(byDatabase(database))
 	s.mu.RUnlock()
@@ -1320,6 +1476,11 @@ func (s *Store) MeasurementNames(auth query.Authorizer, database string, cond in
 		is.Indexes = append(is.Indexes, index)
 	}
 	is = is.DedupeInmemIndexes()
+	select {
+	case <-ctx.Done():
+		return nil, ctx.Err()
+	default:
+	}
 	return is.MeasurementNamesByExpr(auth, cond)
 }
 
@@ -1342,7 +1503,7 @@ func (a TagKeysSlice) Swap(i, j int)      { a[i], a[j] = a[j], a[i] }
 func (a TagKeysSlice) Less(i, j int) bool { return a[i].Measurement < a[j].Measurement }
 
 // TagKeys returns the tag keys in the given database, matching the condition.
-func (s *Store) TagKeys(auth query.Authorizer, shardIDs []uint64, cond influxql.Expr) ([]TagKeys, error) {
+func (s *Store) TagKeys(ctx context.Context, auth query.FineAuthorizer, shardIDs []uint64, cond influxql.Expr) ([]TagKeys, error) {
 	if len(shardIDs) == 0 {
 		return nil, nil
 	}
@@ -1387,10 +1548,20 @@ func (s *Store) TagKeys(auth query.Authorizer, shardIDs []uint64, cond influxql.
 		}
 
 		if is.SeriesFile == nil {
-			is.SeriesFile = shard.sfile
+			sfile, err := shard.SeriesFile()
+			if err != nil {
+				s.mu.RUnlock()
+				return nil, err
+			}
+			is.SeriesFile = sfile
 		}
 
-		is.Indexes = append(is.Indexes, shard.index)
+		index, err := shard.Index()
+		if err != nil {
+			s.mu.RUnlock()
+			return nil, err
+		}
+		is.Indexes = append(is.Indexes, index)
 	}
 	s.mu.RUnlock()
 
@@ -1405,6 +1576,12 @@ func (s *Store) TagKeys(auth query.Authorizer, shardIDs []uint64, cond influxql.
 	var results []TagKeys
 	for _, name := range names {
 
+		// Check for timeouts
+		select {
+		case <-ctx.Done():
+			return nil, ctx.Err()
+		default:
+		}
 		// Build keyset over all indexes for measurement.
 		tagKeySet, err := is.MeasurementTagKeysByExpr(name, nil)
 		if err != nil {
@@ -1418,6 +1595,12 @@ func (s *Store) TagKeys(auth query.Authorizer, shardIDs []uint64, cond influxql.
 		// If they have authorized series associated with them.
 		if filterExpr == nil {
 			for tagKey := range tagKeySet {
+				// check for timeouts
+				select {
+				case <-ctx.Done():
+					return nil, ctx.Err()
+				default:
+				}
 				ok, err := is.TagKeyHasAuthorizedSeries(auth, []byte(name), []byte(tagKey))
 				if err != nil {
 					return nil, err
@@ -1498,7 +1681,7 @@ func (a tagValuesSlice) Less(i, j int) bool { return bytes.Compare(a[i].name, a[
 
 // TagValues returns the tag keys and values for the provided shards, where the
 // tag values satisfy the provided condition.
-func (s *Store) TagValues(auth query.Authorizer, shardIDs []uint64, cond influxql.Expr) ([]TagValues, error) {
+func (s *Store) TagValues(ctx context.Context, auth query.FineAuthorizer, shardIDs []uint64, cond influxql.Expr) ([]TagValues, error) {
 	if cond == nil {
 		return nil, errors.New("a condition is required")
 	}
@@ -1543,9 +1726,21 @@ func (s *Store) TagValues(auth query.Authorizer, shardIDs []uint64, cond influxq
 		}
 
 		if is.SeriesFile == nil {
-			is.SeriesFile = shard.sfile
+			sfile, err := shard.SeriesFile()
+			if err != nil {
+				s.mu.RUnlock()
+				return nil, err
+			}
+			is.SeriesFile = sfile
 		}
-		is.Indexes = append(is.Indexes, shard.index)
+
+		index, err := shard.Index()
+		if err != nil {
+			s.mu.RUnlock()
+			return nil, err
+		}
+
+		is.Indexes = append(is.Indexes, index)
 	}
 	s.mu.RUnlock()
 	is = is.DedupeInmemIndexes()
@@ -1575,6 +1770,13 @@ func (s *Store) TagValues(auth query.Authorizer, shardIDs []uint64, cond influxq
 	// values from matching series. Series may be filtered using a WHERE
 	// filter.
 	for _, name := range names {
+		// check for timeouts
+		select {
+		case <-ctx.Done():
+			return nil, ctx.Err()
+		default:
+		}
+
 		// Determine a list of keys from condition.
 		keySet, err := is.MeasurementTagKeysByExpr(name, cond)
 		if err != nil {
@@ -1637,6 +1839,13 @@ func (s *Store) TagValues(auth query.Authorizer, shardIDs []uint64, cond influxq
 	// instances of tagValues for a given measurement.
 	idxBuf := make([][2]int, 0, len(is.Indexes))
 	for i < len(allResults) {
+		// check for timeouts
+		select {
+		case <-ctx.Done():
+			return nil, ctx.Err()
+		default:
+		}
+
 		// Gather all occurrences of the same measurement for merging.
 		for j+1 < len(allResults) && bytes.Equal(allResults[j+1].name, allResults[i].name) {
 			j++
@@ -1767,7 +1976,6 @@ func mergeTagValues(valueIdxs [][2]int, tvs ...tagValues) TagValues {
 }
 
 func (s *Store) monitorShards() {
-	defer s.wg.Done()
 	t := time.NewTicker(10 * time.Second)
 	defer t.Stop()
 	t2 := time.NewTicker(time.Minute)
@@ -1779,7 +1987,7 @@ func (s *Store) monitorShards() {
 		case <-t.C:
 			s.mu.RLock()
 			for _, sh := range s.shards {
-				if sh.IsIdle() {
+				if isIdle, _ := sh.IsIdle(); isIdle {
 					if err := sh.Free(); err != nil {
 						s.Logger.Warn("Error while freeing cold shard resources",
 							zap.Error(err),
