@@ -3,6 +3,7 @@ package tsdb_test
 import (
 	"bytes"
 	"context"
+	"errors"
 	"fmt"
 	"io/ioutil"
 	"math"
@@ -13,18 +14,19 @@ import (
 	"regexp"
 	"sort"
 	"strings"
+	"sync"
 	"testing"
 	"time"
 
-	"github.com/influxdata/influxdb/tsdb/index/inmem"
-
 	"github.com/davecgh/go-spew/spew"
 	"github.com/influxdata/influxdb/internal"
 	"github.com/influxdata/influxdb/logger"
 	"github.com/influxdata/influxdb/models"
 	"github.com/influxdata/influxdb/pkg/deep"
+	"github.com/influxdata/influxdb/pkg/slices"
 	"github.com/influxdata/influxdb/query"
 	"github.com/influxdata/influxdb/tsdb"
+	"github.com/influxdata/influxdb/tsdb/index/inmem"
 	"github.com/influxdata/influxql"
 )
 
@@ -142,6 +144,243 @@ func TestStore_CreateShard(t *testing.T) {
 	}
 }
 
+func TestStore_CreateMixedShards(t *testing.T) {
+	t.Parallel()
+
+	test := func(index1 string, index2 string) {
+		s := MustOpenStore(index1)
+		defer s.Close()
+
+		// Create a new shard and verify that it exists.
+		if err := s.CreateShard("db0", "rp0", 1, true); err != nil {
+			t.Fatal(err)
+		} else if sh := s.Shard(1); sh == nil {
+			t.Fatalf("expected shard")
+		}
+
+		s.EngineOptions.IndexVersion = index2
+		s.index = index2
+		if err := s.Reopen(); err != nil {
+			t.Fatal(err)
+		}
+
+		// Create another shard and verify that it exists.
+		if err := s.CreateShard("db0", "rp0", 2, true); err != nil {
+			t.Fatal(err)
+		} else if sh := s.Shard(2); sh == nil {
+			t.Fatalf("expected shard")
+		}
+
+		// Reopen shard and recheck.
+		if err := s.Reopen(); err != nil {
+			t.Fatal(err)
+		} else if sh := s.Shard(1); sh == nil {
+			t.Fatalf("expected shard(1)")
+		} else if sh = s.Shard(2); sh == nil {
+			t.Fatalf("expected shard(2)")
+		}
+
+		sh := s.Shard(1)
+		if sh.IndexType() != index1 {
+			t.Fatalf("got index %v, expected %v", sh.IndexType(), index1)
+		}
+
+		sh = s.Shard(2)
+		if sh.IndexType() != index2 {
+			t.Fatalf("got index %v, expected %v", sh.IndexType(), index2)
+		}
+	}
+
+	indexes := tsdb.RegisteredIndexes()
+	for i := range indexes {
+		j := (i + 1) % len(indexes)
+		index1 := indexes[i]
+		index2 := indexes[j]
+		t.Run(fmt.Sprintf("%s-%s", index1, index2), func(t *testing.T) { test(index1, index2) })
+	}
+}
+
+func TestStore_DropMeasurementMixedShards(t *testing.T) {
+	t.Parallel()
+
+	test := func(index1 string, index2 string) {
+		s := MustOpenStore(index1)
+		defer s.Close()
+
+		if err := s.CreateShard("db0", "rp0", 1, true); err != nil {
+			t.Fatal(err)
+		}
+
+		s.MustWriteToShardString(1, "mem,server=a v=1 10")
+
+		s.EngineOptions.IndexVersion = index2
+		s.index = index2
+		if err := s.Reopen(); err != nil {
+			t.Fatal(err)
+		}
+
+		if err := s.CreateShard("db0", "rp0", 2, true); err != nil {
+			t.Fatal(err)
+		}
+
+		s.MustWriteToShardString(2, "mem,server=b v=1 20")
+
+		s.MustWriteToShardString(1, "cpu,server=a v=1 10")
+		s.MustWriteToShardString(2, "cpu,server=b v=1 20")
+
+		err := s.DeleteMeasurement("db0", "cpu")
+		if err != tsdb.ErrMultipleIndexTypes {
+			t.Fatal(err)
+		} else if err == nil {
+			t.Fatal("expect failure deleting measurement on multiple index types")
+		}
+	}
+
+	indexes := tsdb.RegisteredIndexes()
+	for i := range indexes {
+		j := (i + 1) % len(indexes)
+		index1 := indexes[i]
+		index2 := indexes[j]
+		t.Run(fmt.Sprintf("%s-%s", index1, index2), func(t *testing.T) { test(index1, index2) })
+	}
+}
+
+func TestStore_DropConcurrentWriteMultipleShards(t *testing.T) {
+	t.Parallel()
+
+	test := func(index string) {
+		s := MustOpenStore(index)
+		defer s.Close()
+
+		if err := s.CreateShard("db0", "rp0", 1, true); err != nil {
+			t.Fatal(err)
+		}
+
+		s.MustWriteToShardString(1, "mem,server=a v=1 10")
+
+		if err := s.CreateShard("db0", "rp0", 2, true); err != nil {
+			t.Fatal(err)
+		}
+
+		s.MustWriteToShardString(2, "mem,server=b v=1 20")
+
+		errCh := make(chan error)
+		go func() {
+			for i := 0; i < 50; i++ {
+				s.MustWriteToShardString(1, "cpu,server=a v=1 10")
+				s.MustWriteToShardString(2, "cpu,server=b v=1 20")
+			}
+			errCh <- nil
+		}()
+
+		go func() {
+			for i := 0; i < 50; i++ {
+				if err := s.DeleteMeasurement("db0", "cpu"); err != nil {
+					errCh <- err
+					return
+				}
+			}
+			errCh <- nil
+		}()
+
+		for i := 0; i < 2; i++ {
+			if err := <-errCh; err != nil {
+				t.Fatal(err)
+			}
+		}
+		err := s.DeleteMeasurement("db0", "cpu")
+		if err != nil {
+			t.Fatal(err)
+		}
+
+		measurements, err := s.MeasurementNames(context.Background(), query.OpenAuthorizer, "db0", nil)
+		if err != nil {
+			t.Fatal(err)
+		}
+
+		exp := [][]byte{[]byte("mem")}
+		if got, exp := measurements, exp; !reflect.DeepEqual(got, exp) {
+			t.Fatal(fmt.Errorf("got measurements %v, expected %v", got, exp))
+		}
+	}
+
+	for _, index := range tsdb.RegisteredIndexes() {
+		t.Run(index, func(t *testing.T) { test(index) })
+	}
+}
+
+func TestStore_WriteMixedShards(t *testing.T) {
+	t.Parallel()
+
+	test := func(index1 string, index2 string) {
+		s := MustOpenStore(index1)
+		defer s.Close()
+
+		if err := s.CreateShard("db0", "rp0", 1, true); err != nil {
+			t.Fatal(err)
+		}
+
+		s.MustWriteToShardString(1, "mem,server=a v=1 10")
+
+		s.EngineOptions.IndexVersion = index2
+		s.index = index2
+		if err := s.Reopen(); err != nil {
+			t.Fatal(err)
+		}
+
+		if err := s.CreateShard("db0", "rp0", 2, true); err != nil {
+			t.Fatal(err)
+		}
+
+		s.MustWriteToShardString(2, "mem,server=b v=1 20")
+
+		var wg sync.WaitGroup
+		wg.Add(2)
+
+		go func() {
+			defer wg.Done()
+			for i := 0; i < 50; i++ {
+				s.MustWriteToShardString(1, fmt.Sprintf("cpu,server=a,f%0.2d=a v=1", i*2))
+			}
+		}()
+
+		go func() {
+			defer wg.Done()
+			for i := 0; i < 50; i++ {
+				s.MustWriteToShardString(2, fmt.Sprintf("cpu,server=b,f%0.2d=b v=1 20", i*2+1))
+			}
+		}()
+
+		wg.Wait()
+
+		keys, err := s.TagKeys(context.Background(), nil, []uint64{1, 2}, nil)
+		if err != nil {
+			t.Fatal(err)
+		}
+
+		cpuKeys := make([]string, 101)
+		for i := 0; i < 100; i++ {
+			cpuKeys[i] = fmt.Sprintf("f%0.2d", i)
+		}
+		cpuKeys[100] = "server"
+		expKeys := []tsdb.TagKeys{
+			{Measurement: "cpu", Keys: cpuKeys},
+			{Measurement: "mem", Keys: []string{"server"}},
+		}
+		if got, exp := keys, expKeys; !reflect.DeepEqual(got, exp) {
+			t.Fatalf("got keys %v, expected %v", got, exp)
+		}
+	}
+
+	indexes := tsdb.RegisteredIndexes()
+	for i := range indexes {
+		j := (i + 1) % len(indexes)
+		index1 := indexes[i]
+		index2 := indexes[j]
+		t.Run(fmt.Sprintf("%s-%s", index1, index2), func(t *testing.T) { test(index1, index2) })
+	}
+}
+
 // Ensure the store does not return an error when delete from a non-existent db.
 func TestStore_DeleteSeries_NonExistentDB(t *testing.T) {
 	t.Parallel()
@@ -214,7 +453,7 @@ func TestStore_DeleteShard(t *testing.T) {
 		// cpu,serverb=b should be removed from the series file for db0 because
 		// shard 1 was the only owner of that series.
 		// Verify by getting  all tag keys.
-		keys, err := s.TagKeys(nil, []uint64{2}, nil)
+		keys, err := s.TagKeys(context.Background(), nil, []uint64{2}, nil)
 		if err != nil {
 			return err
 		}
@@ -229,7 +468,7 @@ func TestStore_DeleteShard(t *testing.T) {
 
 		// Verify that the same series was not removed from other databases'
 		// series files.
-		if keys, err = s.TagKeys(nil, []uint64{3}, nil); err != nil {
+		if keys, err = s.TagKeys(context.Background(), nil, []uint64{3}, nil); err != nil {
 			return err
 		}
 
@@ -264,7 +503,7 @@ func TestStore_CreateShardSnapShot(t *testing.T) {
 			t.Fatalf("expected shard")
 		}
 
-		dir, e := s.CreateShardSnapshot(1)
+		dir, e := s.CreateShardSnapshot(1, false)
 		if e != nil {
 			t.Fatal(e)
 		}
@@ -627,7 +866,7 @@ func TestStore_MeasurementNames_Deduplicate(t *testing.T) {
 			`cpu value=3 20`,
 		)
 
-		meas, err := s.MeasurementNames(query.OpenAuthorizer, "db0", nil)
+		meas, err := s.MeasurementNames(context.Background(), query.OpenAuthorizer, "db0", nil)
 		if err != nil {
 			t.Fatalf("unexpected error with MeasurementNames: %v", err)
 		}
@@ -668,7 +907,7 @@ func testStoreCardinalityTombstoning(t *testing.T, store *Store) {
 	}
 
 	// Delete all the series for each measurement.
-	mnames, err := store.MeasurementNames(nil, "db", nil)
+	mnames, err := store.MeasurementNames(context.Background(), nil, "db", nil)
 	if err != nil {
 		t.Fatal(err)
 	}
@@ -680,7 +919,7 @@ func testStoreCardinalityTombstoning(t *testing.T, store *Store) {
 	}
 
 	// Estimate the series cardinality...
-	cardinality, err := store.Store.SeriesCardinality("db")
+	cardinality, err := store.Store.SeriesCardinality(context.Background(), "db")
 	if err != nil {
 		t.Fatal(err)
 	}
@@ -692,7 +931,7 @@ func testStoreCardinalityTombstoning(t *testing.T, store *Store) {
 
 	// Since all the series have been deleted, all the measurements should have
 	// been removed from the index too.
-	if cardinality, err = store.Store.MeasurementsCardinality("db"); err != nil {
+	if cardinality, err = store.Store.MeasurementsCardinality(context.Background(), "db"); err != nil {
 		t.Fatal(err)
 	}
 
@@ -746,7 +985,7 @@ func testStoreCardinalityUnique(t *testing.T, store *Store) {
 	}
 
 	// Estimate the series cardinality...
-	cardinality, err := store.Store.SeriesCardinality("db")
+	cardinality, err := store.Store.SeriesCardinality(context.Background(), "db")
 	if err != nil {
 		t.Fatal(err)
 	}
@@ -757,7 +996,7 @@ func testStoreCardinalityUnique(t *testing.T, store *Store) {
 	}
 
 	// Estimate the measurement cardinality...
-	if cardinality, err = store.Store.MeasurementsCardinality("db"); err != nil {
+	if cardinality, err = store.Store.MeasurementsCardinality(context.Background(), "db"); err != nil {
 		t.Fatal(err)
 	}
 
@@ -828,7 +1067,7 @@ func testStoreCardinalityDuplicates(t *testing.T, store *Store) {
 	}
 
 	// Estimate the series cardinality...
-	cardinality, err := store.Store.SeriesCardinality("db")
+	cardinality, err := store.Store.SeriesCardinality(context.Background(), "db")
 	if err != nil {
 		t.Fatal(err)
 	}
@@ -839,7 +1078,7 @@ func testStoreCardinalityDuplicates(t *testing.T, store *Store) {
 	}
 
 	// Estimate the measurement cardinality...
-	if cardinality, err = store.Store.MeasurementsCardinality("db"); err != nil {
+	if cardinality, err = store.Store.MeasurementsCardinality(context.Background(), "db"); err != nil {
 		t.Fatal(err)
 	}
 
@@ -872,6 +1111,150 @@ func TestStore_Cardinality_Duplicates(t *testing.T) {
 	}
 }
 
+func TestStore_MetaQuery_Timeout(t *testing.T) {
+	if testing.Short() || os.Getenv("APPVEYOR") != "" {
+		t.Skip("Skipping test in short and appveyor mode.")
+	}
+
+	test := func(index string) {
+		store := NewStore(index)
+		store.EngineOptions.Config.MaxSeriesPerDatabase = 0
+		if err := store.Open(); err != nil {
+			panic(err)
+		}
+		defer store.Close()
+		testStoreMetaQueryTimeout(t, store, index)
+	}
+
+	for _, index := range tsdb.RegisteredIndexes() {
+		test(index)
+	}
+}
+
+func testStoreMetaQueryTimeout(t *testing.T, store *Store, index string) {
+	shards := testStoreMetaQuerySetup(t, store)
+
+	testStoreMakeTimedFuncs(func(ctx context.Context) (string, error) {
+		const funcName = "SeriesCardinality"
+		_, err := store.Store.SeriesCardinality(ctx, "db")
+		return funcName, err
+	}, index)(t)
+
+	testStoreMakeTimedFuncs(func(ctx context.Context) (string, error) {
+		const funcName = "MeasurementsCardinality"
+		_, err := store.Store.MeasurementsCardinality(ctx, "db")
+		return funcName, err
+	}, index)(t)
+
+	keyCondition, allCondition := testStoreMetaQueryCondition()
+
+	testStoreMakeTimedFuncs(func(ctx context.Context) (string, error) {
+		const funcName = "TagValues"
+		_, err := store.Store.TagValues(ctx, nil, shards, allCondition)
+		return funcName, err
+	}, index)(t)
+
+	testStoreMakeTimedFuncs(func(ctx context.Context) (string, error) {
+		const funcName = "TagKeys"
+		_, err := store.Store.TagKeys(ctx, nil, shards, keyCondition)
+		return funcName, err
+	}, index)(t)
+
+	testStoreMakeTimedFuncs(func(ctx context.Context) (string, error) {
+		const funcName = "MeasurementNames"
+		_, err := store.Store.MeasurementNames(ctx, nil, "db", nil)
+		return funcName, err
+	}, index)(t)
+}
+
+func testStoreMetaQueryCondition() (influxql.Expr, influxql.Expr) {
+	keyCondition := &influxql.ParenExpr{
+		Expr: &influxql.BinaryExpr{
+			Op: influxql.OR,
+			LHS: &influxql.BinaryExpr{
+				Op:  influxql.EQ,
+				LHS: &influxql.VarRef{Val: "_tagKey"},
+				RHS: &influxql.StringLiteral{Val: "tagKey4"},
+			},
+			RHS: &influxql.BinaryExpr{
+				Op:  influxql.EQ,
+				LHS: &influxql.VarRef{Val: "_tagKey"},
+				RHS: &influxql.StringLiteral{Val: "tagKey5"},
+			},
+		},
+	}
+
+	whereCondition := &influxql.ParenExpr{
+		Expr: &influxql.BinaryExpr{
+			Op: influxql.AND,
+			LHS: &influxql.ParenExpr{
+				Expr: &influxql.BinaryExpr{
+					Op:  influxql.EQ,
+					LHS: &influxql.VarRef{Val: "tagKey1"},
+					RHS: &influxql.StringLiteral{Val: "tagValue2"},
+				},
+			},
+			RHS: keyCondition,
+		},
+	}
+
+	allCondition := &influxql.BinaryExpr{
+		Op: influxql.AND,
+		LHS: &influxql.ParenExpr{
+			Expr: &influxql.BinaryExpr{
+				Op:  influxql.EQREGEX,
+				LHS: &influxql.VarRef{Val: "tagKey3"},
+				RHS: &influxql.RegexLiteral{Val: regexp.MustCompile(`tagValue\d`)},
+			},
+		},
+		RHS: whereCondition,
+	}
+	return keyCondition, allCondition
+}
+
+func testStoreMetaQuerySetup(t *testing.T, store *Store) []uint64 {
+	const measurementCnt = 64
+	const tagCnt = 5
+	const valueCnt = 5
+	const pointsPerShard = 20000
+
+	// Generate point data to write to the shards.
+	series := genTestSeries(measurementCnt, tagCnt, valueCnt)
+
+	points := make([]models.Point, 0, len(series))
+	for _, s := range series {
+		points = append(points, models.MustNewPoint(s.Measurement, s.Tags, map[string]interface{}{"value": 1.0}, time.Now()))
+	}
+	// Create requested number of shards in the store & write points across
+	// shards such that we never write the same series to multiple shards.
+	shards := make([]uint64, len(points)/pointsPerShard)
+	for shardID := 0; shardID < len(points)/pointsPerShard; shardID++ {
+		if err := store.CreateShard("db", "rp", uint64(shardID), true); err != nil {
+			t.Fatalf("create shard: %s", err)
+		}
+		if err := store.BatchWrite(shardID, points[shardID*pointsPerShard:(shardID+1)*pointsPerShard]); err != nil {
+			t.Fatalf("batch write: %s", err)
+		}
+		shards[shardID] = uint64(shardID)
+	}
+	return shards
+}
+
+func testStoreMakeTimedFuncs(tested func(context.Context) (string, error), index string) func(*testing.T) {
+	cancelTested := func(t *testing.T) {
+		ctx, cancel := context.WithTimeout(context.Background(), time.Duration(0))
+		defer cancel()
+
+		funcName, err := tested(ctx)
+		if err == nil {
+			t.Fatalf("%v: failed to time out with index type %v", funcName, index)
+		} else if !strings.Contains(err.Error(), context.DeadlineExceeded.Error()) {
+			t.Fatalf("%v: failed with %v instead of %v with index type %v", funcName, err, context.DeadlineExceeded, index)
+		}
+	}
+	return cancelTested
+}
+
 // Creates a large number of series in multiple shards, which will force
 // compactions to occur.
 func testStoreCardinalityCompactions(store *Store) error {
@@ -897,7 +1280,7 @@ func testStoreCardinalityCompactions(store *Store) error {
 	}
 
 	// Estimate the series cardinality...
-	cardinality, err := store.Store.SeriesCardinality("db")
+	cardinality, err := store.Store.SeriesCardinality(context.Background(), "db")
 	if err != nil {
 		return err
 	}
@@ -908,7 +1291,7 @@ func testStoreCardinalityCompactions(store *Store) error {
 	}
 
 	// Estimate the measurement cardinality...
-	if cardinality, err = store.Store.MeasurementsCardinality("db"); err != nil {
+	if cardinality, err = store.Store.MeasurementsCardinality(context.Background(), "db"); err != nil {
 		return err
 	}
 
@@ -944,12 +1327,72 @@ func TestStore_Cardinality_Compactions(t *testing.T) {
 	}
 }
 
+func TestStore_Cardinality_Limit_On_InMem_Index(t *testing.T) {
+	t.Parallel()
+
+	if testing.Short() || os.Getenv("GORACE") != "" || os.Getenv("APPVEYOR") != "" {
+		t.Skip("Skipping test in short, race and appveyor mode.")
+	}
+
+	store := NewStore("inmem")
+	store.EngineOptions.Config.MaxSeriesPerDatabase = 100000
+	if err := store.Open(); err != nil {
+		panic(err)
+	}
+	defer store.Close()
+
+	// Generate 200,000 series to write.
+	series := genTestSeries(64, 5, 5)
+
+	// Add 1 point to each series.
+	points := make([]models.Point, 0, len(series))
+	for _, s := range series {
+		points = append(points, models.MustNewPoint(s.Measurement, s.Tags, map[string]interface{}{"value": 1.0}, time.Now()))
+	}
+
+	// Create shards to write points into.
+	numShards := 10
+	for shardID := 0; shardID < numShards; shardID++ {
+		if err := store.CreateShard("db", "rp", uint64(shardID), true); err != nil {
+			t.Fatalf("create shard: %s", err)
+		}
+	}
+
+	// Write series / points to the shards.
+	pointsPerShard := len(points) / numShards
+
+	for shardID := 0; shardID < numShards; shardID++ {
+		from := shardID * pointsPerShard
+		to := from + pointsPerShard
+
+		if err := store.Store.WriteToShard(uint64(shardID), points[from:to]); err != nil {
+			if !strings.Contains(err.Error(), "partial write: max-series-per-database limit exceeded:") {
+				t.Fatal(err)
+			}
+		}
+	}
+
+	// Get updated series cardinality from store after writing data.
+	cardinality, err := store.Store.SeriesCardinality(context.Background(), "db")
+	if err != nil {
+		t.Fatal(err)
+	}
+	expCardinality := store.EngineOptions.Config.MaxSeriesPerDatabase
+
+	// Estimated cardinality should be well within 1.5% of the actual cardinality.
+	got := math.Abs(float64(cardinality)-float64(expCardinality)) / float64(expCardinality)
+	exp := 0.015
+	if got > exp {
+		t.Errorf("got epsilon of %v for series cardinality %d (expected %d), which is larger than expected %v", got, cardinality, expCardinality, exp)
+	}
+}
+
 func TestStore_Sketches(t *testing.T) {
 	t.Parallel()
 
 	checkCardinalities := func(store *tsdb.Store, series, tseries, measurements, tmeasurements int) error {
 		// Get sketches and check cardinality...
-		sketch, tsketch, err := store.SeriesSketches("db")
+		sketch, tsketch, err := store.SeriesSketches(context.Background(), "db")
 		if err != nil {
 			return err
 		}
@@ -975,7 +1418,7 @@ func TestStore_Sketches(t *testing.T) {
 		}
 
 		// Check measurement cardinality.
-		if sketch, tsketch, err = store.MeasurementsSketches("db"); err != nil {
+		if sketch, tsketch, err = store.MeasurementsSketches(context.Background(), "db"); err != nil {
 			return err
 		}
 
@@ -1029,7 +1472,7 @@ func TestStore_Sketches(t *testing.T) {
 		}
 
 		// Delete half the the measurements data
-		mnames, err := store.MeasurementNames(nil, "db", nil)
+		mnames, err := store.MeasurementNames(context.Background(), nil, "db", nil)
 		if err != nil {
 			return err
 		}
@@ -1041,7 +1484,7 @@ func TestStore_Sketches(t *testing.T) {
 		}
 
 		// Check cardinalities. In this case, the indexes behave differently.
-		expS, expTS, expM, expTM := 160, 0, 10, 5
+		expS, expTS, expM, expTM := 160, 80, 10, 5
 		if index == inmem.IndexName {
 			expS, expTS, expM, expTM = 160, 80, 10, 5
 		}
@@ -1057,7 +1500,7 @@ func TestStore_Sketches(t *testing.T) {
 		}
 
 		// Check cardinalities. In this case, the indexes behave differently.
-		expS, expTS, expM, expTM = 160, 0, 5, 5
+		expS, expTS, expM, expTM = 80, 80, 5, 5
 		if index == inmem.IndexName {
 			expS, expTS, expM, expTM = 80, 0, 5, 0
 		}
@@ -1162,9 +1605,8 @@ func TestStore_TagValues(t *testing.T) {
 		},
 	}
 
-	var s *Store
-	setup := func(index string) []uint64 { // returns shard ids
-		s = MustOpenStore(index)
+	setup := func(index string) (*Store, []uint64) { // returns shard ids
+		s := MustOpenStore(index)
 
 		fmtStr := `cpu1%[1]d,foo=a,ignoreme=nope,host=tv%[2]d,shard=s%[3]d value=1 %[4]d
 	cpu1%[1]d,host=nofoo value=1 %[4]d
@@ -1189,14 +1631,14 @@ func TestStore_TagValues(t *testing.T) {
 			ids = append(ids, uint64(i))
 			s.MustCreateShardWithData("db0", "rp0", i, genPoints(i)...)
 		}
-		return ids
+		return s, ids
 	}
 
 	for _, example := range examples {
 		for _, index := range tsdb.RegisteredIndexes() {
-			shardIDs := setup(index)
+			s, shardIDs := setup(index)
 			t.Run(example.Name+"_"+index, func(t *testing.T) {
-				got, err := s.TagValues(nil, shardIDs, example.Expr)
+				got, err := s.TagValues(context.Background(), nil, shardIDs, example.Expr)
 				if err != nil {
 					t.Fatal(err)
 				}
@@ -1238,7 +1680,7 @@ func TestStore_Measurements_Auth(t *testing.T) {
 			},
 		}
 
-		names, err := s.MeasurementNames(authorizer, "db0", nil)
+		names, err := s.MeasurementNames(context.Background(), authorizer, "db0", nil)
 		if err != nil {
 			return err
 		}
@@ -1268,7 +1710,7 @@ func TestStore_Measurements_Auth(t *testing.T) {
 			return err
 		}
 
-		if names, err = s.MeasurementNames(authorizer, "db0", nil); err != nil {
+		if names, err = s.MeasurementNames(context.Background(), authorizer, "db0", nil); err != nil {
 			return err
 		}
 
@@ -1325,7 +1767,7 @@ func TestStore_TagKeys_Auth(t *testing.T) {
 			},
 		}
 
-		keys, err := s.TagKeys(authorizer, []uint64{0}, nil)
+		keys, err := s.TagKeys(context.Background(), authorizer, []uint64{0}, nil)
 		if err != nil {
 			return err
 		}
@@ -1360,7 +1802,7 @@ func TestStore_TagKeys_Auth(t *testing.T) {
 			return err
 		}
 
-		if keys, err = s.TagKeys(authorizer, []uint64{0}, nil); err != nil {
+		if keys, err = s.TagKeys(context.Background(), authorizer, []uint64{0}, nil); err != nil {
 			return err
 		}
 
@@ -1423,7 +1865,7 @@ func TestStore_TagValues_Auth(t *testing.T) {
 			},
 		}
 
-		values, err := s.TagValues(authorizer, []uint64{0}, &influxql.BinaryExpr{
+		values, err := s.TagValues(context.Background(), authorizer, []uint64{0}, &influxql.BinaryExpr{
 			Op:  influxql.EQ,
 			LHS: &influxql.VarRef{Val: "_tagKey"},
 			RHS: &influxql.StringLiteral{Val: "host"},
@@ -1463,7 +1905,7 @@ func TestStore_TagValues_Auth(t *testing.T) {
 			return err
 		}
 
-		values, err = s.TagValues(authorizer, []uint64{0}, &influxql.BinaryExpr{
+		values, err = s.TagValues(context.Background(), authorizer, []uint64{0}, &influxql.BinaryExpr{
 			Op:  influxql.EQ,
 			LHS: &influxql.VarRef{Val: "_tagKey"},
 			RHS: &influxql.StringLiteral{Val: "host"},
@@ -1531,6 +1973,286 @@ func createTagValues(mname string, kvs map[string][]string) tsdb.TagValues {
 	return out
 }
 
+func TestStore_MeasurementNames_ConcurrentDropShard(t *testing.T) {
+	for _, index := range tsdb.RegisteredIndexes() {
+		s := MustOpenStore(index)
+		defer s.Close()
+
+		shardN := 10
+		for i := 0; i < shardN; i++ {
+			// Create new shards with some data
+			s.MustCreateShardWithData("db0", "rp0", i,
+				`cpu,host=serverA value=1 30`,
+				`mem,region=west value=2 40`, // skip: wrong source
+				`cpu,host=serverC value=3 60`,
+			)
+		}
+
+		done := make(chan struct{})
+		errC := make(chan error, 2)
+
+		// Randomly close and open the shards.
+		go func() {
+			for {
+				select {
+				case <-done:
+					errC <- nil
+					return
+				default:
+					i := uint64(rand.Intn(int(shardN)))
+					if sh := s.Shard(i); sh == nil {
+						errC <- errors.New("shard should not be nil")
+						return
+					} else {
+						if err := sh.Close(); err != nil {
+							errC <- err
+							return
+						}
+						time.Sleep(500 * time.Microsecond)
+						if err := sh.Open(); err != nil {
+							errC <- err
+							return
+						}
+					}
+				}
+			}
+		}()
+
+		// Attempt to get tag keys from the shards.
+		go func() {
+			for {
+				select {
+				case <-done:
+					errC <- nil
+					return
+				default:
+					names, err := s.MeasurementNames(context.Background(), nil, "db0", nil)
+					if err == tsdb.ErrIndexClosing || err == tsdb.ErrEngineClosed {
+						continue // These errors are expected
+					}
+
+					if err != nil {
+						errC <- err
+						return
+					}
+
+					if got, exp := names, slices.StringsToBytes("cpu", "mem"); !reflect.DeepEqual(got, exp) {
+						errC <- fmt.Errorf("got keys %v, expected %v", got, exp)
+						return
+					}
+				}
+			}
+		}()
+
+		// Run for 500ms
+		time.Sleep(500 * time.Millisecond)
+		close(done)
+
+		// Check for errors.
+		if err := <-errC; err != nil {
+			t.Fatal(err)
+		}
+		if err := <-errC; err != nil {
+			t.Fatal(err)
+		}
+	}
+}
+
+func TestStore_TagKeys_ConcurrentDropShard(t *testing.T) {
+	for _, index := range tsdb.RegisteredIndexes() {
+		s := MustOpenStore(index)
+		defer s.Close()
+
+		shardN := 10
+		for i := 0; i < shardN; i++ {
+			// Create new shards with some data
+			s.MustCreateShardWithData("db0", "rp0", i,
+				`cpu,host=serverA value=1 30`,
+				`mem,region=west value=2 40`, // skip: wrong source
+				`cpu,host=serverC value=3 60`,
+			)
+		}
+
+		done := make(chan struct{})
+		errC := make(chan error, 2)
+
+		// Randomly close and open the shards.
+		go func() {
+			for {
+				select {
+				case <-done:
+					errC <- nil
+					return
+				default:
+					i := uint64(rand.Intn(int(shardN)))
+					if sh := s.Shard(i); sh == nil {
+						errC <- errors.New("shard should not be nil")
+						return
+					} else {
+						if err := sh.Close(); err != nil {
+							errC <- err
+							return
+						}
+						time.Sleep(500 * time.Microsecond)
+						if err := sh.Open(); err != nil {
+							errC <- err
+							return
+						}
+					}
+				}
+			}
+		}()
+
+		// Attempt to get tag keys from the shards.
+		go func() {
+			for {
+				select {
+				case <-done:
+					errC <- nil
+					return
+				default:
+					keys, err := s.TagKeys(context.Background(), nil, []uint64{0, 1, 2, 3, 4, 5, 6, 7, 8, 9}, nil)
+					if err == tsdb.ErrIndexClosing || err == tsdb.ErrEngineClosed {
+						continue // These errors are expected
+					}
+
+					if err != nil {
+						errC <- err
+						return
+					}
+
+					if got, exp := keys[0].Keys, []string{"host"}; !reflect.DeepEqual(got, exp) {
+						errC <- fmt.Errorf("got keys %v, expected %v", got, exp)
+						return
+					}
+
+					if got, exp := keys[1].Keys, []string{"region"}; !reflect.DeepEqual(got, exp) {
+						errC <- fmt.Errorf("got keys %v, expected %v", got, exp)
+						return
+					}
+				}
+			}
+		}()
+
+		// Run for 500ms
+		time.Sleep(500 * time.Millisecond)
+
+		close(done)
+
+		// Check for errors
+		if err := <-errC; err != nil {
+			t.Fatal(err)
+		}
+		if err := <-errC; err != nil {
+			t.Fatal(err)
+		}
+	}
+}
+
+func TestStore_TagValues_ConcurrentDropShard(t *testing.T) {
+	for _, index := range tsdb.RegisteredIndexes() {
+		s := MustOpenStore(index)
+		defer s.Close()
+
+		shardN := 10
+		for i := 0; i < shardN; i++ {
+			// Create new shards with some data
+			s.MustCreateShardWithData("db0", "rp0", i,
+				`cpu,host=serverA value=1 30`,
+				`mem,region=west value=2 40`, // skip: wrong source
+				`cpu,host=serverC value=3 60`,
+			)
+		}
+
+		done := make(chan struct{})
+		errC := make(chan error, 2)
+
+		// Randomly close and open the shards.
+		go func() {
+			for {
+				select {
+				case <-done:
+					errC <- nil
+					return
+				default:
+					i := uint64(rand.Intn(int(shardN)))
+					if sh := s.Shard(i); sh == nil {
+						errC <- errors.New("shard should not be nil")
+						return
+					} else {
+						if err := sh.Close(); err != nil {
+							errC <- err
+							return
+						}
+						time.Sleep(500 * time.Microsecond)
+						if err := sh.Open(); err != nil {
+							errC <- err
+							return
+						}
+					}
+				}
+			}
+		}()
+
+		// Attempt to get tag keys from the shards.
+		go func() {
+			for {
+				select {
+				case <-done:
+					errC <- nil
+					return
+				default:
+					stmt, err := influxql.ParseStatement(`SHOW TAG VALUES WITH KEY = "host"`)
+					if err != nil {
+						errC <- err
+					}
+					rewrite, err := query.RewriteStatement(stmt)
+					if err != nil {
+						errC <- err
+					}
+
+					cond := rewrite.(*influxql.ShowTagValuesStatement).Condition
+					values, err := s.TagValues(context.Background(), nil, []uint64{0, 1, 2, 3, 4, 5, 6, 7, 8, 9}, cond)
+					if err == tsdb.ErrIndexClosing || err == tsdb.ErrEngineClosed {
+						continue // These errors are expected
+					}
+
+					if err != nil {
+						errC <- err
+						return
+					}
+
+					exp := tsdb.TagValues{
+						Measurement: "cpu",
+						Values: []tsdb.KeyValue{
+							{Key: "host", Value: "serverA"},
+							{Key: "host", Value: "serverC"},
+						},
+					}
+
+					if got := values[0]; !reflect.DeepEqual(got, exp) {
+						errC <- fmt.Errorf("got keys %v, expected %v", got, exp)
+						return
+					}
+				}
+			}
+		}()
+
+		// Run for 500ms
+		time.Sleep(500 * time.Millisecond)
+
+		close(done)
+
+		// Check for errors
+		if err := <-errC; err != nil {
+			t.Fatal(err)
+		}
+		if err := <-errC; err != nil {
+			t.Fatal(err)
+		}
+	}
+}
+
 func BenchmarkStore_SeriesCardinality_100_Shards(b *testing.B) {
 	for _, index := range tsdb.RegisteredIndexes() {
 		store := NewStore(index)
@@ -1552,7 +2274,7 @@ func BenchmarkStore_SeriesCardinality_100_Shards(b *testing.B) {
 
 		b.Run(store.EngineOptions.IndexVersion, func(b *testing.B) {
 			for i := 0; i < b.N; i++ {
-				_, _ = store.SeriesCardinality("db")
+				_, _ = store.SeriesCardinality(context.Background(), "db")
 			}
 		})
 		store.Close()
@@ -1634,8 +2356,7 @@ func BenchmarkStore_TagValues(b *testing.B) {
 		{name: "s=10_m=100_v=1000", shards: 10, measurements: 100, tagValues: 1000},
 	}
 
-	var s *Store
-	setup := func(shards, measurements, tagValues int, index string, useRandom bool) []uint64 { // returns shard ids
+	setup := func(shards, measurements, tagValues int, index string, useRandom bool) (*Store, []uint64) { // returns shard ids
 		s := NewStore(index)
 		if err := s.Open(); err != nil {
 			panic(err)
@@ -1671,13 +2392,7 @@ func BenchmarkStore_TagValues(b *testing.B) {
 			shardIDs = append(shardIDs, uint64(i))
 			s.MustCreateShardWithData("db0", "rp0", i, genPoints(i, useRandom)...)
 		}
-		return shardIDs
-	}
-
-	teardown := func() {
-		if err := s.Close(); err != nil {
-			b.Fatal(err)
-		}
+		return s, shardIDs
 	}
 
 	// SHOW TAG VALUES WITH KEY IN ("host", "shard")
@@ -1716,14 +2431,20 @@ func BenchmarkStore_TagValues(b *testing.B) {
 		for useRand := 0; useRand < 2; useRand++ {
 			for c, condition := range []influxql.Expr{cond1, cond2} {
 				for _, bm := range benchmarks {
-					shardIDs := setup(bm.shards, bm.measurements, bm.tagValues, index, useRand == 1)
+					s, shardIDs := setup(bm.shards, bm.measurements, bm.tagValues, index, useRand == 1)
+					teardown := func() {
+						if err := s.Close(); err != nil {
+							b.Fatal(err)
+						}
+					}
+
 					cnd := "Unfiltered"
 					if c == 0 {
 						cnd = "Filtered"
 					}
 					b.Run("random_values="+fmt.Sprint(useRand == 1)+"_index="+index+"_"+cnd+"_"+bm.name, func(b *testing.B) {
 						for i := 0; i < b.N; i++ {
-							if tvResult, err = s.TagValues(nil, shardIDs, condition); err != nil {
+							if tvResult, err = s.TagValues(context.Background(), nil, shardIDs, condition); err != nil {
 								b.Fatal(err)
 							}
 						}
